"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package ActionHandler
#ActionHandler builds the interface and handles most of the user inputs.		
#@author Josh Blum

import os, sys, popen2
from Constants import *
from Actions import *
import pygtk
pygtk.require('2.0')
import gtk
import Graphics
import Preferences
from threading import Thread
import Messages
import ParseXML
import random

class ActionHandler:
	"""
	The action handler will setup all the major window components,
	and handle button presses and flow graph operations from the GUI.
	"""
	
	def __init__(self, file_paths): 
		"""!
		ActionHandler constructor.
		Create the main window, setup the message handler, import the preferences, 
		and connect all of the action handlers. Finally, enter the gtk main loop and block.
		@param file_paths a list of flow graph file passed from command line		
		"""				
		if PY_GTK_ICON: gtk.window_set_default_icon_from_file(PY_GTK_ICON)
		for action in ACTIONS_LIST: action.connect('activate', self._handle_actions)
		#setup the main window
		self.main_window = Graphics.MainWindow(self.handle_states)
		self.main_window.connect('delete_event', self._quit)
		self.main_window.connect('key_press_event', self._handle_key_press)
		self.get_flow_graph = self.main_window.get_flow_graph
		self.get_page = self.main_window.get_page
		#setup the messages
		Messages.register_messenger(self.main_window.add_report_line)
		Messages.register_messenger(sys.stdout.write)
		Messages.send_init()
		#flow graph settings	
		Preferences.load(self.main_window)	
		self.init_file_paths = file_paths
		self.handle_states(APPLICATION_INITIALIZE)
		#enter the mainloop		
		gtk.gdk.threads_init()
		gtk.gdk.threads_enter()
		gtk.main()
		gtk.gdk.threads_leave()		
		
	def _handle_key_press(self, widget, event):
		"""
		Handle key presses from the keyboard.		
		Translate key combos into actions.	
		Key combinations that do not include special keys, such as ctrl or Fcn*,
		Also, require that the flow graph has mouse focus when choosing to handle keys.
		@return true if the flow graph is in active use
		"""
		keyname = gtk.gdk.keyval_name(event.keyval)
		#print 'Key "%s" (%d) was pressed' % (keyname, event.keyval)
		#if event.state & gtk.gdk.CONTROL_MASK: print "Control was being held down"
		#if event.state & gtk.gdk.MOD1_MASK: print "Alt was being held down"
		#if event.state & gtk.gdk.SHIFT_MASK: print "Shift was being held down"		
		####################	save/open/new/undo/redo	###############################	
		if event.state & gtk.gdk.CONTROL_MASK and keyname == 's':
			self.handle_states(FLOW_GRAPH_SAVE)
		elif event.state & gtk.gdk.CONTROL_MASK and keyname == 'z':
			self.handle_states(FLOW_GRAPH_UNDO)
		elif event.state & gtk.gdk.CONTROL_MASK and (keyname in ('y', 'Z')):
			self.handle_states(FLOW_GRAPH_REDO)
		elif event.state & gtk.gdk.CONTROL_MASK and keyname == 'o':
			self.handle_states(FLOW_GRAPH_OPEN)
		elif event.state & gtk.gdk.CONTROL_MASK and keyname == 'n':
			self.handle_states(FLOW_GRAPH_NEW)
		####################	Delete	###############################
		elif self.get_flow_graph().get_focus_flag() and keyname == 'Delete':	#mouse focus
			self.handle_states(ELEMENT_DELETE)
		####################	Rotate	###############################
		elif self.get_flow_graph().get_focus_flag() and keyname == 'Right': #mouse focus
			self.handle_states(SIGNAL_BLOCK_ROTATE_RIGHT)
		elif self.get_flow_graph().get_focus_flag() and keyname == 'Left': #mouse focus
			self.handle_states(SIGNAL_BLOCK_ROTATE_LEFT)
		####################	Data Type	###############################
		elif self.get_flow_graph().get_focus_flag() and keyname == 'Down': #mouse focus
			self.handle_states(SIGNAL_BLOCK_INC_TYPE)
		elif self.get_flow_graph().get_focus_flag() and keyname == 'Up': #mouse focus
			self.handle_states(SIGNAL_BLOCK_DEC_TYPE)
		####################	Socket Controllers	###############################		
		elif self.get_flow_graph().get_focus_flag() and keyname in ('equal','plus', 'KP_Add'): #mouse focus
			self.handle_states(SOCKET_CONTROLLER_INC)
		elif self.get_flow_graph().get_focus_flag() and keyname in ('minus', 'KP_Subtract'): #mouse focus
			self.handle_states(SOCKET_CONTROLLER_DEC)		
		####################	Exec/Stop/Print	###############################
		elif keyname == 'F5':
			self.handle_states(FLOW_GRAPH_EXEC)
		elif keyname == 'F7':
			self.handle_states(FLOW_GRAPH_STOP)
		elif keyname == 'Print':
			self.handle_states(FLOW_GRAPH_SCREEN_CAPTURE)
		#propagate this if the fg is not in focus or nothing is selected
		return self.get_flow_graph().get_focus_flag() and self.get_flow_graph().is_selected()
	 
	def _quit(self, window, event):
		"""!
		Handle the delete event from the main window.
		Generated by pressing X to close, alt+f4, or right click+close.	
		This method in turns calls the state handler to quit.
		@return true		
		"""
		self.handle_states(APPLICATION_QUIT)
		return True	
	
	def _handle_actions(self, event):		
		"""
		Handle all of the activate signals from the gtk actions. 
		The action signals derive from clicking on a	toolbar or menu bar button. 
		Forward the action to the state handler.
		"""
		self.handle_states(event.get_name())		
		
	def handle_states(self, state=''):
		"""!
		Handle the state changes in the GUI.		
		Handle all of the state changes that arise from the action handler or other Graphics and 
		inputs in the application. The state passed to the handle_states method is a string descriping 
		the change. A series of if/elif statements handle the state by greying out action buttons, causing 
		changes in the flow graph, saving/opening files...	The handle_states method is passed to the
		contructors of many of the classes used in this application enabling them to report any state change.
		@param state a string describing the state change
		"""
		##############################################################################################
		#	Initalize/Quit
		##############################################################################################
		if state == APPLICATION_INITIALIZE:	
			for action in ACTIONS_LIST: action.set_sensitive(False) #set all actions disabled
			# enable a select few actions 
			Graphics.enable_usrp_diagnostics() 	#try to enable usrp diagnostics
			for action in (
				APPLICATION_QUIT, FLOW_GRAPH_NEW, FLOW_GRAPH_OPEN, FLOW_GRAPH_SAVE_AS, FLOW_GRAPH_CLOSE,
				ABOUT_WINDOW_DISPLAY, DATA_TYPES_WINDOW_DISPLAY, HOTKEYS_WINDOW_DISPLAY, MATH_EXPR_WINDOW_DISPLAY,
				FLOW_GRAPH_WINDOW_RESIZE, PREFS_WINDOW_DISPLAY, FLOW_GRAPH_SCREEN_CAPTURE,
			): get_action_from_name(action).set_sensitive(True)		
			if not self.init_file_paths and Preferences.restore_files(): self.init_file_paths = Preferences.files_open()
			if not self.init_file_paths: self.init_file_paths = ['']
			for file_path in self.init_file_paths: self.main_window.new_page(file_path) #load pages from file paths
			if not self.get_page():	self.main_window.new_page() #ensure that at least a blank page exists
		elif state == APPLICATION_QUIT:
			if self.main_window.close_pages():
				Preferences.save(self.main_window)
				gtk.main_quit()
				exit(0)
		##############################################################################################
		#	Selections
		##############################################################################################
		elif state == SIGNAL_BLOCK_SELECT or state == SOCKET_SELECT:
			for action in (ELEMENT_DELETE, SIGNAL_BLOCK_PARAM_MODIFY, SIGNAL_BLOCK_ROTATE_RIGHT, SIGNAL_BLOCK_ROTATE_LEFT): 
				get_action_from_name(action).set_sensitive(True)
		elif state == CONNECTION_SELECT:
			get_action_from_name(ELEMENT_DELETE).set_sensitive(True)
			for action in (SIGNAL_BLOCK_PARAM_MODIFY, SIGNAL_BLOCK_ROTATE_RIGHT, SIGNAL_BLOCK_ROTATE_LEFT): 
				get_action_from_name(action).set_sensitive(False)
		elif state == NOTHING_SELECT:
			for action in (ELEMENT_DELETE, SIGNAL_BLOCK_PARAM_MODIFY, SIGNAL_BLOCK_ROTATE_RIGHT, SIGNAL_BLOCK_ROTATE_LEFT): 
				get_action_from_name(action).set_sensitive(False)
			self.get_flow_graph().unselect()
		##############################################################################################
		#	Move/Rotate/Delete/Create
		##############################################################################################		
		elif state == SIGNAL_BLOCK_MOVE:			
			self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())
			self.get_page().set_saved(False)
		elif state == SIGNAL_BLOCK_ROTATE_LEFT:
			if self.get_flow_graph().rotate_selected(DIR_LEFT):				
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
				self.get_page().set_saved(False)		
		elif state == SIGNAL_BLOCK_ROTATE_RIGHT:
			if self.get_flow_graph().rotate_selected(DIR_RIGHT):				
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
				self.get_page().set_saved(False)		
		elif state == ELEMENT_DELETE:
			if self.get_flow_graph().delete_selected():				
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())				
				self.handle_states(NOTHING_SELECT)
				self.get_page().set_saved(False)
		elif state == CONNECTION_CREATE or state == SIGNAL_BLOCK_CREATE:			
			self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())
			self.handle_states(NOTHING_SELECT)	
			self.get_page().set_saved(False)
		elif state == SIGNAL_BLOCK_INC_TYPE:
			if self.get_flow_graph().type_controller_modify_selected(1):
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
				self.get_page().set_saved(False)
		elif state == SIGNAL_BLOCK_DEC_TYPE:
			if self.get_flow_graph().type_controller_modify_selected(-1):
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
				self.get_page().set_saved(False)	
		elif state == SOCKET_CONTROLLER_INC:
			if self.get_flow_graph().socket_controller_modify_selected(1):
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
				self.get_page().set_saved(False)
		elif state == SOCKET_CONTROLLER_DEC:
			if self.get_flow_graph().socket_controller_modify_selected(-1):
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
				self.get_page().set_saved(False)		
		##############################################################################################
		#	Window stuff
		##############################################################################################
		elif state == USRP_DIAGNOSTICS_DISPLAY:
			Graphics.USRPDiagnosticsDialog()		
		elif state == PREFS_WINDOW_DISPLAY:
			Graphics.PreferencesDialog()
			self.get_flow_graph().update()
		elif state == ABOUT_WINDOW_DISPLAY:
			Graphics.AboutDialog()
		elif state == DATA_TYPES_WINDOW_DISPLAY:
			Graphics.DataTypesDialog()
		elif state == HOTKEYS_WINDOW_DISPLAY:
			Graphics.HotKeysDialog()
		elif state == MATH_EXPR_WINDOW_DISPLAY:
			Graphics.MathExprDialog()
		elif state == FLOW_GRAPH_WINDOW_RESIZE:
			dimensions = Graphics.FlowGraphWindowSizeDialog(self.get_flow_graph().get_size_request()).run()
			if dimensions:
				self.get_flow_graph().set_size_request(*dimensions)
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())					
				self.get_page().set_saved(False)
		##############################################################################################
		#	Variable and Param Modifications
		##############################################################################################
		elif state == SIGNAL_BLOCK_PARAM_MODIFY:
			if self.get_flow_graph().param_modify_selected():
				self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
				self.get_page().set_saved(False)		
		elif state == VARIABLE_MODIFY:
			self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
			self.get_page().set_saved(False)	
			self.get_flow_graph().update()
		elif state == VARIABLE_REORDER:
			self.get_page().get_state_cache().save_new_state(self.get_flow_graph().to_nested_data())	
			self.get_page().set_saved(False)	
		##############################################################################################
		#	Undo/Redo
		##############################################################################################
		elif state == FLOW_GRAPH_UNDO:		
			nested_data = self.get_page().get_state_cache().get_prev_state()
			if nested_data	!= None:		
				self.handle_states(NOTHING_SELECT)
				self.get_flow_graph().from_nested_data(nested_data)
				self.get_page().set_saved(False)
		elif state == FLOW_GRAPH_REDO:						
			nested_data = self.get_page().get_state_cache().get_next_state()
			if nested_data	!= None:		
				self.handle_states(NOTHING_SELECT)
				self.get_flow_graph().from_nested_data(nested_data)
				self.get_page().set_saved(False)
		##############################################################################################
		#	New/Open/Save/Close
		##############################################################################################
		elif state == FLOW_GRAPH_NEW:
			self.main_window.new_page()
		elif state == FLOW_GRAPH_OPEN:
			file_path = Graphics.OpenFlowGraphFileDialog(self.get_flow_graph() and self.get_page().get_file_path() or '').run()
			if file_path != None: self.main_window.new_page(file_path, manual_open=True)
		elif state == FLOW_GRAPH_CLOSE:
			self.main_window.close_page()		
		elif state == FLOW_GRAPH_SAVE:
			if not self.get_page().get_file_path(): self.handle_states(FLOW_GRAPH_SAVE_AS)
			else:
				try:
					ParseXML.to_file(ParseXML.to_xml(self.get_flow_graph().to_nested_data()), self.get_page().get_file_path())
					self.get_page().set_saved(True)		
				except IOError: 
					Messages.send_fail_save(self.get_page().get_file_path())
					self.get_page().set_saved(False)
		elif state == FLOW_GRAPH_SAVE_AS:
			file_path = Graphics.SaveFlowGraphFileDialog(self.get_page().get_file_path()).run()
			if file_path != None:			
				self.get_page().set_file_path(file_path)
				self.handle_states(FLOW_GRAPH_SAVE)	
		elif state == FLOW_GRAPH_SCREEN_CAPTURE:
			file_path = Graphics.SaveImageFileDialog(self.get_page().get_file_path()).run()
			if file_path != None: 
				pixmap = self.get_flow_graph().pixmap
				width, height = pixmap.get_size()
				pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, 0, 8, width, height)
				pixbuf.get_from_drawable(pixmap, pixmap.get_colormap(), 0, 0, 0, 0, width, height)
				pixbuf.save(file_path, IMAGE_FILE_EXTENSION[1:])						
		##############################################################################################
		#	Run/Stop
		##############################################################################################		
		elif state == FLOW_GRAPH_EXEC:
			if not self.get_page().get_pid_file():
				if not self.get_page().get_saved() or not self.get_page().get_file_path(): 
					self.handle_states(FLOW_GRAPH_SAVE) #only save if file path missing or not saved
				if self.get_page().get_saved() and self.get_page().get_file_path(): 
					ExecFlowGraphThread(self)	#only exec if file path and saved
		elif state == FLOW_GRAPH_STOP:
			if self.get_page().get_pid_file(): 
				try: os.kill(int(open(self.get_page().get_pid_file(), 'r').read()), 9)
				except: print "could not kill pid file: %s"%self.get_page().get_pid_file()
		elif state == '': #pass and run the global actions
			pass
		else: print '!!! State "%s" not handled !!!'%state
		##############################################################################################
		#	Global Actions for all States
		##############################################################################################
		#set the exec button if the flow graph is valid and is not already running
		self.update_exec_stop()
		#saved status		
		get_action_from_name(FLOW_GRAPH_SAVE).set_sensitive(not self.get_page().get_saved())
		self.main_window.update()				

	def update_exec_stop(self):
		"""
		Update the exec and stop buttons.
		Lock and unlock the mutex for race conditions with exec flow graph threads.
		"""
		MUTEX.lock()
		get_action_from_name(FLOW_GRAPH_EXEC).set_sensitive(self.get_flow_graph().is_valid() and not self.get_page().get_pid_file())
		get_action_from_name(FLOW_GRAPH_STOP).set_sensitive(self.get_page().get_pid_file() != '')	
		MUTEX.unlock()
			
class ExecFlowGraphThread(Thread):
	"""Execute the flow graph as a new process and wait on it to finish."""
	def __init__ (self, action_handler):
		"""!
		ExecFlowGraphThread constructor.
		@param action_handler an instance of an ActionHandler
		"""
		Thread.__init__(self)
		self.update_exec_stop = action_handler.update_exec_stop
		self.flow_graph = action_handler.get_flow_graph()
		#store page and dont use main window calls in run
		self.page = action_handler.get_page()
		#random id so multiple flow graphs can run w/o files intersecting
		rand_id = random.randint(10000, 99999) 
		#set files		
		self.file_path = self.page.get_file_path()
		self.pid_file = '/tmp/grc-%d-%d.pid'%(os.getpid(), rand_id)		
		self.page.set_pid_file(self.pid_file)
		self.update_exec_stop()
		Messages.send_start_exec(self.page.get_file_path())
		self.start()
		
	def run(self):	
		"""Execute the flow graph."""
		cmd = '%s "%s" --pid_file="%s"'%(
			DEFAULT_FLOW_GRAPH_EXEC, 
			self.file_path, 
			self.pid_file
		)
		p = popen2.Popen3(cmd=cmd, capturestderr=True)
		#read all the lines of the stderr into a list
		Messages.send_end_exec(p.childerr.readlines())	
		try: os.remove(self.pid_file)
		except: print "could not remove pid file: %s"%self.pid_file
		self.page.set_pid_file('')
		self.update_exec_stop()
		
