"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Elements.Connection
#The elemental connection of input and output sockets.
#@author Josh Blum

import Element		
import Utils

class ConnectionException(Exception): pass

class Connection(Element.Element):
	"""Connection provides elemental interfaces for a connection of sockets."""	
	
	type = 'connection'
	
	def __init__(self, parent, socket1, socket2):
		"""!
		Connection constructor.		
		Initialize the Element with the given parent and a rotation and coordinate of zeros.	
		Connect a pair of sockets. The sockets must be an input and an output. The input must have no existing connections. 
		@param parent the flow graph
		@param socket1 an input/output socket
		@param socket2 another input/output socket
		@throw InvalidConnectionException cant connect		
		"""
		Element.Element.__init__(self, parent, (0, 0), 0)
		#If an input socket that is already connected is passed to the constructor,
		#the logic will use the output socket that it is connected to.
		#This way, a user could try to connect an unconnected input socket to a signal on another input.
		#Then, the program will translate this to the cooresponding output socket on that signal.
		#>>>I dont like this functionality, uncomment it to pick the output socket		
		#if is_input_socket(socket1) and socket1.is_connected(): #socket1 is input and connected
		#	socket1 = socket1.get_connections()[0].get_output_socket()	#so we use its output socket
		#if is_input_socket(socket2) and socket2.is_connected(): #same for socket2
		#	socket2 = socket2.get_connections()[0].get_output_socket()
		if Utils.is_input_socket(socket1) and Utils.is_output_socket(socket2):
			socket1.connect(self) #order is important, try input socket first (since input can throw an exception)
			socket2.connect(self)
			self.sockets = [socket1, socket2]
		elif Utils.is_input_socket(socket2) and Utils.is_output_socket(socket1):
			socket2.connect(self) #order is important, try input socket first (since input can throw an exception)
			socket1.connect(self)
			self.sockets = [socket2, socket1]
		else: raise ConnectionException, 'A connection requires a valid input and output socket!'
		
	def get_coordinate(self):
		"""!Get the 0,0 coordinate.
		Coordinates are irrelevant in connection.
		@return 0, 0
		"""
		return 0, 0
		
	def get_rotation(self):
		"""!Get the 0 degree rotation.
		Rotations are irrelevant in connection.
		@return 0
		"""
		return 0
		
	def get_input_socket(self):
		"""!
		Get the input socket.
		@return input socket
		"""
		return self.sockets[0]
		
	def get_output_socket(self):
		"""!
		Get the output socket.
		@return output socket
		"""
		return self.sockets[1]	
		
	def disconnect(self):
		"""Disconnect the sockets by calling each socket's disconnect methods and removing the sockets from the parent."""
		print "disconnecting"
		for socket in self.sockets: socket.disconnect(self)
		self.get_parent().remove_element(self)
		
							