"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Elements.Element
#The base class for graphical elements such as:
#signal blocks, input sockets, output sockets and connections.
#@author Josh Blum

from Constants import POSSIBLE_ROTATIONS,DIR_LEFT,DIR_RIGHT,CONNECTION_SELECT_SENSITIVITY

class Element:
	"""
	Element is the base class for all elements. It contains an X,Y coordinate, a list
	of rectangular areas (and lines) that the element occupies and methods to detect selection.
	"""		
		
	def __init__(self, parent, coor, rotation):
		"""!
		Make a new list of rectangular areas and lines, and set the coordinate and the rotation.
		@param parent the parent of this element
		@param coor the coordinate tuple (x,y)
		@param rotation an angle in degrees 0, 90, 180, 270
		"""	
		self.parent = parent		
		self.set_rotation(rotation)
		self.set_coordinate(coor)		
		self.clear()
		self.highlighted = False
		
	def rotate(self, direction):
		"""!
		Rotate all of the areas by 90 degrees.
		@param direction DIR_RIGHT or DIR_LEFT
		"""
		if direction == DIR_LEFT: delta_rotation = 90
		elif direction == DIR_RIGHT: delta_rotation = 270			
		self.set_rotation((self.get_rotation() + delta_rotation)%360)
		self.update()			
	
	def clear(self):
		"""Empty the lines and areas."""
		self.areas_dict = dict((rotation, list()) for rotation in POSSIBLE_ROTATIONS)
		self.lines_dict = dict((rotation, list()) for rotation in POSSIBLE_ROTATIONS)
			
	def set_coordinate(self, coor):
		"""!
		Set the reference coordinate.
		@param coor the coordinate tuple (x,y)
		"""
		self.coor = coor
	
	def get_parent(self):
		"""!
		Get the parent of this element.
		@return the parent
		"""	
		return self.parent
	
	def set_highlighted(self, highlighted):
		"""!
		Set the highlight status.
		@param highlighted true to enable highlighting
		"""
		self.highlighted = highlighted		
	
	def is_highlighted(self):
		"""!
		Get the highlight status.
		@return true if highlighted
		"""
		return self.highlighted
			
	def get_coordinate(self):
		"""!Get the coordinate.
		@return the coordinate tuple (x,y)
		"""
		return self.coor
		
	def move(self, delta_coor):
		"""!
		Move the element by adding the delta_coor to the current coordinate.
		@param delta_coor (delta_x,delta_y) tuple
		"""
		deltaX, deltaY = delta_coor
		X, Y = self.get_coordinate()	
		self.coor = (X+deltaX, Y+deltaY)	
	
	def add_area(self, rel_coor, area, rotation=None):
		"""!
		Add an area to the area list. 
		An area is actually a coordinate relative to the main coordinate with a width/height pair relative to the area coordinate. 
		A positive width is to the right of the coordinate. A positive height is above the coordinate.
		The area is associated with a rotation. If rotation is not specified, the element's current rotation is used.
		@param rel_coor (x,y) offset from this element's coordinate
		@param area (width,height) tuple
		@param rotation rotation in degrees
		"""
		if rotation == None: rotation = self.get_rotation()
		self.areas_dict[rotation].append((rel_coor, area))
		
	def add_line(self, rel_coor1, rel_coor2, rotation=None):
		"""!
		Add a line to the line list. 
		A line is defined by 2 relative coordinates.	Lines must be horizontal or vertical.
		The line is associated with a rotation. If rotation is not specified, the element's current rotation is used.
		@param rel_coor1 relative (x1,y1) tuple
		@param rel_coor2 relative (x2,y2) tuple
		@param rotation rotation in degrees
		"""
		if rotation == None: rotation = self.get_rotation()
		self.lines_dict[rotation].append((rel_coor1, rel_coor2))	
	
	def what_is_selected(self, coor):
		"""!
		Is this Element selected at given coordinate/is the coordinate encompassed by one of the areas or lines?
		@return self if one of the areas/lines encompasses coor, else None.	
		"""
		in_between = lambda p, a, b: p >= min(a, b) and p <= max(a, b)
		x,y = [a-b for a,b in zip(coor, self.get_coordinate())]
		for (x1,y1), (w,h) in self.areas_dict[self.get_rotation()]:
			if in_between(x, x1, x1+w) and in_between(y, y1, y1+h): return self
		for (x1, y1), (x2, y2) in self.lines_dict[self.get_rotation()]: #assume horizontal or vertical lines
			if x1 == x2: x1, x2 = x1-CONNECTION_SELECT_SENSITIVITY, x2+CONNECTION_SELECT_SENSITIVITY
			if y1 == y2: y1, y2 = y1-CONNECTION_SELECT_SENSITIVITY, y2+CONNECTION_SELECT_SENSITIVITY
			if in_between(x, x1, x2) and in_between(y, y1, y2): return self
		return None		
				
	def get_rotation(self):
		"""!
		Get the rotation in degrees.
		@return the rotation
		"""
		return self.rotation			
	
	def set_rotation(self, rotation):
		"""!
		Set the rotation in degrees.
		@param rotation the rotation"""
		if rotation not in POSSIBLE_ROTATIONS:
			raise Exception('"%s" is not one of the possible rotations: (%s)'%(rotation,POSSIBLE_ROTATIONS))
		self.rotation = rotation		
		
	def update(self): 
		"""Do nothing for the update. Dummy method."""
		pass

	