"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Elements.GraphicalConnection
#The graphical connection for input/output sockets.
#@author Josh Blum

import DataTypes
import Utils
import Connection
import GraphicalElement
import Colors

class GraphicalConnection(Connection.Connection, GraphicalElement.GraphicalElement):
	"""A graphical connection for sockets."""			
	
	def update(self):
		"""Add the horizontal and vertical lines that will connect the two parameters."""
		self.clear()		
		input_socket = self.get_input_socket()
		output_socket = self.get_output_socket()
		deltaAngle = abs(output_socket.get_connector_direction() - input_socket.get_connector_direction())
		(x1, y1) = output_socket.get_connector_coordinates()[1]
		self.add_line(*output_socket.get_connector_coordinates())
		(x2, y2) = input_socket.get_connector_coordinates()[1]
		self.add_line(*input_socket.get_connector_coordinates())
		if deltaAngle == 180:
			W = abs(x1 - x2)
			H = abs(y1 - y2)
			midX = (x1+x2)/2
			midY = (y1+y2)/2
			sW = x1 - x2
			if output_socket.get_connector_direction() == 0: sW = sW * -1
			sH = y1 - y2
			if output_socket.get_connector_direction() == 270: sH = sH * -1
			if ((W>H or sW<0) and Utils.is_horizontal(output_socket.get_connector_direction())) or\
				(W>=H and sH>=0 and Utils.is_vertical(output_socket.get_connector_direction())):
				self.add_line((x1,y1),(x1,midY))
				self.add_line((x1,midY),(x2,midY))
				self.add_line((x2,y2),(x2,midY))
			elif (H>=W and sW>=0 and Utils.is_horizontal(output_socket.get_connector_direction())) or\
				((H>W or sH<0) and Utils.is_vertical(output_socket.get_connector_direction())):
				self.add_line((x1,y1),(midX,y1))
				self.add_line((midX,y1),(midX,y2))
				self.add_line((x2,y2),(midX,y2))
		else:
			p1 = (x1, y2)
			p2 = (x2, y1)
			if Utils.get_angle_from_coordinates((x1,y1),p1) == (output_socket.get_connector_direction()+180)%360 or\
				Utils.get_angle_from_coordinates((x2,y2),p1) == (input_socket.get_connector_direction()+180)%360: p = p2
			else: p = p1
			self.add_line((x1,y1),p)
			self.add_line((x2,y2),p)	
		
	def draw(self, window):
		"""!
		Draw the connection.
		@param window the gtk window to draw on
		"""
		self.update()	
		GraphicalElement.GraphicalElement.draw(self, window) 	
		gc = self.gc					
		'''	draw error lines around the existing lines when data types do not match	'''
		if not self.is_valid():
			gc.foreground = Colors.ERROR_COLOR	
			for (x1, y1),(x2, y2) in self.lines_dict[self.get_rotation()]:
				if x1 == x2 and x1 > 0 and x2 > 0: #vertical
					window.draw_line(gc, x1-1, y1, x2-1, y2)
					window.draw_line(gc, x1+1, y1, x2+1, y2)
				elif y1 == y2 and y1 > 0 and y2 > 0: #horizontal
					window.draw_line(gc, x1, y1-1, x2, y2-1)
					window.draw_line(gc, x1, y1+1, x2, y2+1)
		
	def is_valid(self):
		"""!
		Is this connection valid, ie: do the input and output data types match?
		User preferences can override this validity check.
		@return true if the data types match or if check connections is off
		"""
		import Preferences	#preferences
		return not Preferences.check_connections() or\
			DataTypes.can_connect(self.sockets[0].get_data_type(), self.sockets[1].get_data_type())	
							
