"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Elements.GraphicalParam
#GTK objects for handling input and the signal block parameter class.
#@author Josh Blum

import Param
from DataTypes import *
import pygtk
pygtk.require('2.0')
import gtk
import pango
import gobject
from Constants import *
from os import path

######################################################################################################
#	gtk objects for handling input
######################################################################################################

class InputParam(gtk.HBox):
	"""The base class for an input parameter inside the input parameters dialog."""
	
	def __init__(self, data_type, _handle_changed):
		gtk.HBox.__init__(self)
		self.data_type = data_type
		self._handle_changed = _handle_changed
		self.label = gtk.Label()	#no label, markup is added by set_markup
		self.label.set_size_request(140,-1)
		self.pack_start(self.label, False)
		self.set_markup = lambda m: self.label.set_markup(m)	
		self.tp = None		
		
class EntryParam(InputParam):
	"""Provide an entry box for strings and numbers."""
	
	def __init__(self, *args):
		InputParam.__init__(self, *args)
		self.entry = input = gtk.Entry()		
		input.set_text(self.data_type.get_data())
		input.connect("changed", self._handle_changed)
		self.pack_start(input, True)
		self.get_text = input.get_text
		# tool tip fun #
		self.tp = gtk.Tooltips()
		self.tp.set_tip(self.entry, "")
		self.tp.enable()
		
class FileParam(EntryParam):
	"""Provide an entry box for filename and a button to browse for a file."""
	
	def __init__(self, *args):
		EntryParam.__init__(self, *args)		
		input = gtk.Button('...')
		input.connect('clicked', self.handle_clicked)
		self.pack_start(input, False)
	
	def handle_clicked(self, widget=None):
		""" If the button was clicked, open a file dialog in open/save format.
		Replace the text in the entry with the new filename from the file dialog.	"""
		file_path = self.data_type.parse()
		#	bad file paths will be redirected to default	#
		if not path.exists(path.dirname(file_path)): file_path = DEFAULT_FILE_PATH	
		if self.data_type.get_type() == FileOpen().get_type(): 
			file_dialog = gtk.FileChooserDialog('Open a Data File...', None,
				gtk.FILE_CHOOSER_ACTION_OPEN, ('gtk-cancel',gtk.RESPONSE_CANCEL,'gtk-open',gtk.RESPONSE_OK))
		elif self.data_type.get_type() == FileSave().get_type(): 
			file_dialog = gtk.FileChooserDialog('Save a Data File...', None,
				gtk.FILE_CHOOSER_ACTION_SAVE, ('gtk-cancel',gtk.RESPONSE_CANCEL, 'gtk-save',gtk.RESPONSE_OK))
			file_dialog.set_do_overwrite_confirmation(True)
			file_dialog.set_current_name(path.basename(file_path))	#show the current filename
		file_dialog.set_current_folder(path.dirname(file_path))	#current directory
		file_dialog.set_select_multiple(False)
		file_dialog.set_local_only(True)
		if gtk.RESPONSE_OK == file_dialog.run():	#run the dialog
			file_path = file_dialog.get_filename()	#get the file path
			self.entry.set_text(file_path)
			self._handle_changed()
		file_dialog.destroy()	#destroy the dialog						
		
class EnumParam(InputParam):
	"""Provide an entry box for Enum types with a drop down menu."""
	
	def __init__(self, *args):
		InputParam.__init__(self, *args)
		input = gtk.ComboBox(gtk.ListStore(gobject.TYPE_STRING))				
		cell = gtk.CellRendererText()
		input.pack_start(cell, True)
		input.add_attribute(cell, 'text', 0)
		for cname in self.data_type.get_cnames_list(): input.append_text(cname)
		input.set_active(int(self.data_type.get_data()))
		input.connect("changed", self._handle_changed)
		self.pack_start(input, False)
		self.get_text = lambda: str(input.get_active())	#the get text parses the selected index to a string

######################################################################################################
#	A Parameter, hold a data type and a cname
######################################################################################################

class GraphicalParam(Param.Param):
	"""The graphical parameter."""	

	def update(self):
		"""Called when an external change occurs.
		Update the graphical input by calling the change handler."""
		if self.input: self._handle_changed()	
		
	def get_input_object(self, callback=None):
		"""!
		Get the graphical gtk class to represent this parameter.	
		Create the input object with this data type and the handle changed method.
		@param callback a function of one argument(this param) to be called from the change handler
		@return gtk input object
		"""
		self.callback=callback
		if self.get_data_type().get_base_type() == Enum().get_base_type(): input = EnumParam
		elif self.get_data_type().get_base_type() == File().get_base_type(): input = FileParam
		else: input = EntryParam
		self.input = input(self.get_data_type(), self._handle_changed)
		self._handle_changed()
		return self.input
		
	def _handle_changed(self, widget=None):
		"""When the input changes, write the inputs to the data type.
		Finish by calling the exteral callback."""
		data_type = self.get_data_type()
		new_data = self.input.get_text()
		old_data = data_type.get_data()
		if old_data != new_data: data_type.set_data(new_data)
		#set the markup on the label, red for errors in corresponding data type.
		cname = self.get_cname()
		if self.variable: cname = '<span underline="low">%s</span>'%cname #alter format if the param is variable
		if not data_type.is_valid(): self.input.set_markup('<span foreground="red"><b>%s</b></span>'%cname)
		else: self.input.set_markup(cname)
		#set the tooltip 		
		if self.input.tp: self.input.tp.set_tip(self.input.entry, str(data_type)) 	
		#execute the external callback 
		if self.callback: self.callback(self)

	def get_markup(self):
		"""!
		Create a markup to display the Param as a label on the SignalBlock.	
		If the data type is an Enum type, use the cname of the Enum's current choice.	
		Otherwise, use parsed the data type and use its string representation.
		If the data type is not valid, use a red foreground color.
		@return pango markup string
		"""
		###########################################################################
		#	display logic for numbers
		###########################################################################
		def float_to_str(var):
			if var-int(var) == 0: return '%d'%int(var)					
			if var*10-int(var*10) == 0: return '%.1f'%var
			if var*100-int(var*100) == 0: return '%.2f'%var
			if var*1000-int(var*1000) == 0: return '%.3f'%var
			else: return '%.3g'%var				
		def to_str(var): 
			if type(var) == type(str()): return var
			elif type(var) == type(complex()): 
				if var.imag == var.real == 0: return '0'	#value is zero
				elif var.imag == 0: return '%s'%float_to_str(var.real)	#value is real
				elif var.real == 0: return '%sj'%float_to_str(var.imag)	#value is imaginary
				elif var.imag < 0: return '%s-%sj'%(float_to_str(var.real), float_to_str(var.imag*-1))
				else: return '%s+%sj'%(float_to_str(var.real), float_to_str(var.imag))
			elif type(var) == type(float()): return float_to_str(var)				
			elif type(var) == type(int()): return '%d'%var
			else: return var	
		###########################################################################
		if self.get_data_type().is_valid():
			data = self.get_data_type().parse()
			if self.get_data_type().get_base_type() == Enum().get_base_type(): #enum types
				dt_str = self.get_data_type().get_cname()
			elif self.get_data_type().get_base_type() == File().get_base_type():	#file types
				suggested_length = 30
				if len(data) <= suggested_length: dt_str = data
				else:	#truncate the tail if there is not enough space
					tail,head = path.split(data)
					if len(head) >= suggested_length: dt_str = head
					else: dt_str = tail[0:suggested_length-len(head)] + '...' + head
			elif self.get_data_type().get_base_type() == Vector().get_base_type(): #vector types
				for i,e in enumerate(data): data[i] = to_str(e)
				#	only keep the first X elements of the list:			
				X = (100/len(str(data))) + 1	#display more for non complex vectors					
				dt_str = '['		#	replace the removed elements with a '...' 
				for i,f in enumerate(data):
					if i < X or i+1 == len(data):	#leave one on the end 
						dt_str = dt_str + f
						if i < len(data)-1 and (i+1 != X or X+1 == len(data)): dt_str = dt_str + ', '
					elif i == X: dt_str = dt_str + ' ... '
				dt_str = dt_str + ']'
			elif self.get_data_type().get_type() == Hex().get_type(): dt_str = hex(data) #hex, base 16
			else: dt_str = to_str(data)	#other types
			return '<b>%s:</b> %s'%(self.cname, dt_str)
		else:	return '<span foreground="red"><b>%s:</b> error</span>'%self.cname
			
	def get_layout(self):
		"""!
		Create a layout based on the current markup.
		@return the pango layout
		"""
		layout = gtk.DrawingArea().create_pango_layout('')
		layout.set_markup(self.get_markup())
		desc = pango.FontDescription(PARAM_FONT)
		layout.set_font_description(desc) 
		return layout

