"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Elements.GraphicalSignalBlock
#The graphical signal block.
#@author Josh Blum

import SignalBlock
import Utils
import Colors
from GraphicalParam import GraphicalParam
import GraphicalElement
from GraphicalSocket import GraphicalInputSocket,GraphicalOutputSocket
from Constants import *
from DataTypes import Enum,Int
import pygtk
pygtk.require('2.0')
import gtk
import pango

class GraphicalSignalBlock(SignalBlock.SignalBlock, GraphicalElement.GraphicalElement):
	"""The graphical signal block."""		
	
	param_constructor = GraphicalParam
	input_socket_constructor = GraphicalInputSocket
	output_socket_constructor = GraphicalOutputSocket
	
	def update(self):
		"""Update the block, parameters, and sockets when a change occurs."""
		self.clear()
		self._adjust_sockets()		
		self._create_labels()				
		self.W = self.label_width + 2*LABEL_PADDING_WIDTH
		max_sockets = max(self.get_num_input_sockets(), self.get_num_output_sockets(), 1)
		self.H = max(self.label_height+2*LABEL_PADDING_HEIGHT, 2*SOCKET_BORDER_SEPARATION + max_sockets*SOCKET_HEIGHT + (max_sockets-1)*SOCKET_SEPARATION)		
		if Utils.is_horizontal(self.get_rotation()): self.add_area((0,0),(self.W,self.H))
		elif Utils.is_vertical(self.get_rotation()): self.add_area((0,0),(self.H,self.W))					
		for socket in self.get_sockets(): socket.update()			
	
	def _create_labels(self):
		"""Create the labels for the signal block."""
		layouts = list()
		#	create the main layout	#
		layout = gtk.DrawingArea().create_pango_layout(self.cname)
		desc = pango.FontDescription(SIGNAL_BLOCK_FONT)
		layout.set_font_description(desc) 
		layouts.append(layout)
		if not self.is_valid():	layout.set_markup('<span foreground="red"><b>'+self.cname+'</b></span>')			
		self.label_width,self.label_height = layout.get_pixel_size()	
		# handle each of the displayable params	#		
		for param in self.displayed_params:
			layout = param.get_layout()
			layouts.append(layout)
			w,h = layout.get_pixel_size()
			self.label_width = max(w, self.label_width)
			self.label_height = self.label_height + h + LABEL_SEPARATION
		width = self.label_width
		height = self.label_height
		#	setup the pixmap	#
		pixmap = gtk.gdk.Pixmap(self.get_parent().window, width, height, -1)
		gc = pixmap.new_gc()
		gc.foreground = Colors.BG_COLOR
		pixmap.draw_rectangle(gc, True, 0, 0, width, height)
		gc.foreground = Colors.TXT_COLOR
		#	draw the layouts	#		
		h_off = 0
		for i,layout in enumerate(layouts):
			w,h = layout.get_pixel_size()
			if i == 0: w_off = (width-w)/2
			else: w_off = 0
			pixmap.draw_layout(gc, w_off, h_off, layout)
			h_off = h + h_off + LABEL_SEPARATION
		#	create vertical and horizontal images	#			
		self.horizontal_label = image = pixmap.get_image(0, 0, width, height)
		if Utils.is_vertical(self.get_rotation()):
			self.vertical_label = vimage = gtk.gdk.Image(gtk.gdk.IMAGE_NORMAL, pixmap.get_visual(), height, width)
			for i in range(width):
				for j in range(height): vimage.put_pixel(j, width-i-1, image.get_pixel(i, j))
		
	def draw(self, window):
		"""!
		Draw the signal block with label and inputs/outputs.
		@param window the gtk window to draw on
		"""
		GraphicalElement.GraphicalElement.draw(self, window)		
		gc = self.gc		
		gc.foreground = Colors.TXT_COLOR
		X,Y = self.get_coordinate()		
		if Utils.is_horizontal(self.get_rotation()):
			window.draw_image(gc, self.horizontal_label, 0, 0, X+LABEL_PADDING_WIDTH, Y+(self.H-self.label_height)/2, -1, -1)
		elif Utils.is_vertical(self.get_rotation()):
			window.draw_image(gc, self.vertical_label, 0, 0, X+(self.H-self.label_height)/2, Y+LABEL_PADDING_WIDTH, -1, -1)
		for socket in self.get_sockets(): socket.draw(window)
		
	def is_valid(self):		
		"""!
		Is this block is valid/are all the params valid and all sockets connected?
		User preferences can override certain validity checks.
		@return true if valid
		"""
		import Preferences
		if Preferences.check_params():	#preferences
			for param in self.params:
				if not param.get_data_type().is_valid(): return False
		if Preferences.check_sockets():	#preferences
			for socket in self.get_sockets():
				if not socket.is_valid(): return False
		return True
		
	def modify_type_controller(self, direction):
		"""!
		If a type controller was set, increment/decrement its index by 1, use a % to stay in bounds.	
		@param direction +1 or -1
		"""
		if self.type_controller != None and self.type_controller.get_type() == Enum().get_type():
			num_choices = len(self.type_controller.get_cnames_list())
			index = int(self.type_controller.get_data())
			self.type_controller.set_data((index+direction+num_choices)%num_choices)
			self.get_parent().update()
			return True
		return False
		
	def modify_socket_controller(self, direction):
		"""!
		If a socket controller was set, increment/decrement its data type by 1, DO NOT go out of bounds.
		If there is a controller on input and outputs, show preference to the output controller. 
		@param direction +1 or -1
		@return true if operation was done, false if operation could not be completed.
		"""
		# Get the socket controller, if there is an input and output controller, use output	#
		socket_controller = None		
		if self.input_sockets_controller != None: socket_controller = self.input_sockets_controller
		elif self.output_sockets_controller != None: socket_controller = self.output_sockets_controller
		if socket_controller != None and socket_controller.is_valid() and socket_controller.get_type() == Int().get_type():
			old_data = socket_controller.get_data()
			socket_controller.set_data(socket_controller.parse() + direction)
			if socket_controller.is_valid(): #modification was successful
				self.get_parent().update()
				return True
			else: socket_controller.set_data(old_data)	#restore previous value
		return False

