"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Elements.SignalBlock
#The elemental signal block.
#@author Josh Blum

import Element
import Utils
from Socket import OutputSocket,InputSocket
from Param import Param
from DataTypes import Enum,Int,Vector

class SignalBlock(Element.Element):
	"""A SignalBlock is an Element with an index from the list of possible signal blocks and input/output sockets."""	
	
	type = 'signal block'
	param_constructor = Param
	input_socket_constructor = InputSocket
	output_socket_constructor = OutputSocket
	
	def __init__(self, parent, coor, rotation, tag, id):
		"""!
		SignalBlock constructor.
		@param parent the flow graph
		@param coordinate the (x,y) tuple
		@param rotation the rotation in degrees
		@param tag a string unique to signal blocks of this type
		@param id a string unique to this signal block
		"""
		Element.Element.__init__(self, parent, coor, rotation)
		self.tag = tag
		self.input_sockets = list()
		self.output_sockets = list()
		self.params = list()
		self.displayed_params = list()
		self.cname = tag	
		self.id = id
		self.input_sockets_controller = self.output_sockets_controller = None
		self.type_controller = None
		self.H = self.W = 0
		self.docs = None
		
	def get_docs(self):
		"""!
		Get a string with helpful hints, suggestions, and documentation on using this block.
		@return the documentation string
		"""
		return self.docs
		
	def set_docs(self, docs):
		"""!
		Set the documentation for this block.
		@param docs the documentation string
		"""
		self.docs = docs
		
	def get_sockets(self):
		"""!
		Get all sockets for this signal block.
		@return a list of sockets
		"""
		return self.input_sockets + self.output_sockets
	
	def get_input_socket(self, index):
		"""!
		Get the input socket at index.
		@param index the input socket index
		@return an input socket
		"""
		self._adjust_sockets() #a parameter change may have altered the sockets
		for socket in filter(Utils.is_input_socket, self.get_sockets()):
			if socket.get_index() == index: return socket
		return None
		
	def get_output_socket(self, index):
		"""!
		Get the output socket at index.
		@param index the output socket index
		@return an output socket
		"""
		self._adjust_sockets() #a parameter change may have altered the sockets
		for socket in filter(Utils.is_output_socket, self.get_sockets()):
			if socket.get_index() == index: return socket
		return None
		
	def get_id(self):
		"""!
		Get the id.
		@return the id string
		"""
		return self.id
		
	def get_tag(self):	
		"""!
		Get the tag.
		@return the tag string
		"""
		return self.tag
		
	def get_cname(self):
		"""!
		Get the canonical name.
		@return the cname string
		"""
		return self.cname
		
	def get_params(self):
		"""!
		Get all of the parameters for this signal block.
		@return a list of parameters
		"""
		return self.params		
	
	def get_num_input_sockets(self):
		"""!
		Get the number of sockets of input type.
		@return an integer number of sockets
		"""
		return len(self.input_sockets)
		
	def get_num_output_sockets(self):
		"""!
		Get the number of sockets of output type.
		@return an integer number of sockets
		"""
		return len(self.output_sockets)
		
	def add_input_socket(self, cname, data_type, optional=False, vlen=None):
		"""!
		Add an input socket with its cname and data type.
		@param cname the canonical name
		@param data_type the data type
		@param optional true if connections are optional
		@param vlen the width of each sample (how many numbers are in each sample?)
		"""
		self.input_sockets.append(self.input_socket_constructor(self, cname, self.get_num_input_sockets(), data_type, optional, vlen))
		
	def add_output_socket(self, cname, data_type, optional=False, vlen=None):
		"""!
		Add an output socket with its cname and data type.
		@param cname the canonical name
		@param data_type the data type
		@param optional true if connections are optional
		@param vlen the width of each sample (how many numbers are in each sample?)
		"""
		self.output_sockets.append(self.output_socket_constructor(self, cname, self.get_num_output_sockets(), data_type, optional, vlen))
		
	def add_param(self, cname, data_type, show_label=True, type=False, output_sockets_controller=False, input_sockets_controller=False, variable=False):
		"""!
		Add a parameter with its cname and data type.
		@param show_label true to display the parameter in the signal block's label
		@param type true if this param is an enum to be controlled via a keypress
		@param output_sockets_controller register this parameter to set the number of output sockets 
		@param input_sockets_controller register this parameter to set the number of input sockets
		@param variable true if the param can be changed at runtime
		"""
		param = self.param_constructor(cname, data_type, variable)
		if show_label: self.displayed_params.append(param)
		if type: self.type_controller = data_type
		if output_sockets_controller: self.output_sockets_controller = data_type
		if input_sockets_controller: self.input_sockets_controller = data_type
		self.params.append(param)		
	
	def what_is_selected(self, coor):	
		"""!
		Get the element that is selected.
		@param coor the (x,y) tuple
		@return this signal block, a socket, or None
		"""
		for socket in self.get_sockets():	
			if socket.what_is_selected(coor) != None:
				return socket.what_is_selected(coor)		
		return Element.Element.what_is_selected(self, coor)
		
	def get_connections(self):
		"""!
		Get all connections from every socket.
		@return a set of connections
		"""
		connections = list()
		for socket in self.get_sockets(): connections = connections + socket.get_connections()
		return set(connections)
	
	def _adjust_sockets(self):
		"""Use the input/output socket controllers to adjust the number of sockets."""	
		for sockets_list,controller,constructor in (
			(self.input_sockets, self.input_sockets_controller, self.input_socket_constructor), 
			(self.output_sockets, self.output_sockets_controller, self.output_socket_constructor)
		):
			if controller != None and controller.is_valid(): #must be valid before trying to use the value
				if controller.get_base_type() == Vector().get_base_type(): new_num_sockets = len(controller.parse())			
				else: new_num_sockets = int(controller.parse())
				old_num_sockets = len(sockets_list)
				if new_num_sockets > old_num_sockets:
					for i in range(new_num_sockets - old_num_sockets):
						last_socket = sockets_list[old_num_sockets + i - 1]
						new_socket = constructor(self, last_socket.get_cname(), old_num_sockets+i, last_socket.data_type, vlen=last_socket.vlen)
						sockets_list.append(new_socket)
				elif new_num_sockets < old_num_sockets and new_num_sockets > 0:#we do not want to remove all sockets
					for i in range(old_num_sockets - new_num_sockets):
						last_socket = sockets_list[old_num_sockets - i - 1]
						for connection in last_socket.get_connections(): connection.disconnect()
						sockets_list.remove(last_socket)					
					self.get_parent().update()	#removing sockets could delete a connection, update all blocks
		
##########################################################################
##	Import a Signal Block
##########################################################################
	def from_nested_data(fg, nested_data, sb_constructor):
		"""!
		Create a signal block from the nested data.
		Pass the flow graph/parent of the signal block,
		the nested data for a signal block, 
		and the contructor for the signal block.
		Return (the signal block, the runnable signal block) tuple.
		@param fg the flow graph
		@param nested_data the nested data
		@param sb_contructor the contructor used to make a new signal block
		"""
		import ParseXML,Messages,SignalBlockDefs
		find_data = ParseXML.find_data
		signal_block = find_data([nested_data], 'signal_block')
		tag = find_data(signal_block, 'tag')
		id = find_data(signal_block, 'id')
		#	These graphical parameters are optional and will default to 0 if not found	#
		x_coordinate = find_data(signal_block, 'x_coordinate')
		if x_coordinate == None: x_coordinate = 0
		y_coordinate = find_data(signal_block, 'y_coordinate')
		if y_coordinate == None: y_coordinate = 0
		rotation = find_data(signal_block, 'rotation')
		if rotation == None: rotation = 0
		#	Load Parameters	#
		params = find_data(signal_block, 'params')
		try:			
			signal_block,runnable_signal_block = SignalBlockDefs.get_signal_block(
				fg, (int(x_coordinate),int(y_coordinate)), int(rotation), tag, id, sb_constructor,
			)
			for i,param in enumerate(params):
				try:
					param = find_data([param], 'param')
					signal_block.get_params()[i].get_data_type().set_data(param)
				except: Messages.send_error_load('An invalid parameter was parsed, ignoring parameter "%d" in block "%s".'%(i, id))	
			return signal_block, runnable_signal_block		
		except SignalBlockDefs.TagNotFoundException: 
			Messages.send_error_load('The definition for "%s" was not found, ignoring block "%s".'%(tag, id))
		except Exception, e: Messages.send_error_load('Signal block "%s" could not be created:\n\t%s'%(id,e))	
		return None, None	#not found!, return None tuple			
	from_nested_data = staticmethod(from_nested_data)		
		
			