"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Elements.Socket
#The elemental input and output sockets of a signal block.
#@author Josh Blum

import Element
import Utils,DataTypes
from Connection import ConnectionException

class Socket(Element.Element):
	"""
	A Socket is an input/output element of a signal block. 
	It has an index, input/output setting, and a data type.
	"""	
	
	type = 'socket'
	
	def __init__(self, parent, cname, index, data_type, optional=False, vlen=None):
		"""!
		Socket constructor. 
		@param parent the signal block
		@param coordinate the (x,y) tuple
		@param rotation the rotation in degrees
		@param optional true if connections are optional
		@param vlen the width of each sample (how many numbers are in each sample?)
		"""
		Element.Element.__init__(self, parent, (0,0), 0) #use zeros since the parent contains this information
		self.connections = list()	
		self.w = self.h = 0
		self.index = index	
		self.data_type = data_type
		self.cname = cname		
		self.optional = optional
		self.connector_coordinates = dict()	
		self.vlen = vlen
				
	def move(self, delta_coor):
		"""!
		Move the parent rather than self.
		@param delta_corr the (delta_x, delta_y) tuple
		"""
		self.get_parent().move(delta_coor)
		
	def rotate(self, direction):
		"""!
		Rotate the parent rather than self.
		@param direction degrees to rotate
		"""
		self.get_parent().rotate(direction)
		
	def get_coordinate(self):
		"""!
		Get the parent's coordinate rather than self.
		@return the parents coordinate
		"""
		return self.get_parent().get_coordinate()
		
	def set_highlighted(self, highlight):
		"""!
		Set the parent highlight rather than self.
		@param highlight true to enable highlighting
		"""
		self.get_parent().set_highlighted(highlight)
		
	def is_highlighted(self):
		"""!
		Get the parent's is highlight rather than self.
		@return the parent's highlighting status
		"""
		return self.get_parent().is_highlighted()	
		
	def is_valid(self):	
		"""!
		Is this socket valid?
		@return true if connected or the optional flag is true
		"""
		return self.optional or self.is_connected()
	
	def get_rotation(self):
		"""!
		Get the parent's rotation rather than self.
		@return the parent's rotation
		"""
		return self.get_parent().get_rotation()
		
	def get_data_type(self):
		"""!
		Get the data type. If the parsed vlen > 1, then return the vectorized data type.
		@return the data type
		"""
		if self.vlen != None and self.vlen.is_valid() and self.vlen.parse() > 1: 
			return DataTypes.vectorize(self.data_type)
		return self.data_type
			
	def get_index(self):
		"""!
		Get the index of this socket.
		@return the index (integer)	
		"""
		return self.index	
			
	def get_cname(self):
		"""!
		Get the canonical name.
		@return the cname (string)
		"""
		return self.cname
		
	def is_connected(self):
		"""!
		Have one or more connections?
		@return true if at least one connection exists
		"""
		return len(self.connections) > 0 
			
	def connect(self, connection):
		"""!
		Add a connection to this socket.
		@param connection the connection
		@throw TooManyConnectionsException input socket with an existing connection
		"""
		if Utils.is_input_socket(self) and self.is_connected(): 
			raise ConnectionException, 'An input socket may only have one connection!'
		self.connections.append(connection)
		self.update()
			
	def disconnect(self, connection):
		"""!
		Remove a Connection from this socket.
		@param connection the connection
		"""
		self.connections.remove(connection)		
		self.update()	
		
	def get_connections(self):
		"""!
		Get all connections for this socket.
		@return a list of connections
		"""
		return self.connections			

class InputSocket(Socket):
	""" The socket for input	"""
	type = 'input socket'

class OutputSocket(Socket):
	""" The socket for output	"""
	type = 'output socket'
		
		