#! /usr/bin/env python
"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package ExecFlowGraph 
#Use an xml input file to build and run a gnu radio flow graph without graphics.
#@author Josh Blum

import ParseXML
import Variables
from Elements import SignalBlock
from gnuradio import gr
import os, sys
from Constants import FLOW_GRAPH_FILE_EXTENSION,MUTEX
from optparse import OptionParser

##############################################################################################
#	Flow Graph Builder
##############################################################################################	
class FlowGraphBuilder(gr.top_block):
	"""
	Parse the input file to build the gnuradio flow graph.
	Register the variables, handle signal block callbacks.
	"""
		
	def __init__(self, file_path):		
		gr.top_block.__init__(self, file_path)	
		#internal data structures		
		self.callbacks = list()
		self.callbacks_locked = list()
		self.callbacks_cond = list()
		self.var_keys = list()		
		self.started = False
		#build the flow graph
		self._parse_nested_data(ParseXML.from_xml(ParseXML.from_file(file_path)))	
		
	def _start(self):
		"""Start the flow graph."""
		self.start()
		self.started = True
		
	def _stop(self):
		"""Stop the flow graph."""
		self.stop()
		self.started = False
			
	def add_window(self, *args):
		"""Empty method for adding a window in the GUI flow graph."""
		pass
		
	def get_panel(self):
		"""Empty method for getting the wx panel in the GUI flow graph.""" 
		print """
Graphical Sinks are not supported in ExecFlowGraph.py:
 use ExecFlowGraphGUI.py [-d] or remove the Graphical Sink.
Exiting!"""
		exit(-1)
		
	def add_callback(self, function, *data_type_params):
		"""
		Register a callback function with its associated data.
		@param function the callback function
		@param data_type_params a list of data types
		"""
		self.callbacks.append((function, data_type_params))
		
	def add_callback_locked(self, function, *data_type_params):
		"""
		Register a "locked" callback function with its associated data.
		These callbacks will be called inside of a lock/unlock sequence.
		@param function the callback function
		@param data_type_params a list of data types
		"""
		self.callbacks_locked.append((function, data_type_params))
		
	def add_callback_cond(self, function, *data_type_params):
		"""
		Register a "conditional" callback function with its associated data.
		These callbacks will be called inside of a lock/unlock sequence
		only if at least one locked callback has been registered.
		@param function the callback function
		@param data_type_params a list of data types
		"""
		self.callbacks_cond.append((function, data_type_params))
		
	def _parse_callback(self, function, *data_type_params):
		"""
		Parse a single callback. Call function on the data types.
		@param function the callback function
		@param data_type_params a list of data types
		"""
		try: function(*map(lambda param: param.parse(), data_type_params))
		except Exception, e: print "***\n\nerror parsing a callback -> ignoring\n%s...\n\n***"%e
		
	def parse_callbacks(self):
		"""
		For each call back, parse all of the data and 
		call the registered callback function on that data.
		"""
		MUTEX.lock()
		if self.started: 			
			if self.callbacks:	#parse regular callbacks
				for function, data_type_params in self.callbacks:
					self._parse_callback(function, *data_type_params)
			if self.callbacks_locked:	#parse locked callbacks	and conditional callbacks				
				self.lock() 
				for function, data_type_params in self.callbacks_locked + self.callbacks_cond:
					self._parse_callback(function, *data_type_params)
				self.unlock()
		MUTEX.unlock()
		
	def _parse_nested_data(self, nested_data):
		"""!
		Parse nested data from a flow graph file and build the graphics. 
		@param nested_data the nested flow graph data
		"""
		find_data = ParseXML.find_data
		flow_graph = find_data([nested_data], 'flow_graph')	
		vars = find_data(flow_graph, 'vars')
		signal_blocks = find_data(flow_graph, 'signal_blocks')
		connections = find_data(flow_graph, 'connections')
		self.var_keys = Variables.from_nested_data(vars)
		signal_blocks_dict = dict()				
		for signal_block in signal_blocks:			
			signal_block,runnable_signal_block = SignalBlock.from_nested_data(None, signal_block, SignalBlock)			
			data_type_params = list()
			data_type_params.append(self) #put the flow graph in the front of the list
			for param in signal_block.get_params(): #load the list of arguments to the runnable block		
				data_type_params.append(param.get_data_type())			
			signal_blocks_dict[signal_block.get_id()] = runnable_signal_block(*data_type_params)
		for connection in connections:
			connection = find_data([connection], 'connection')
			input_signal_block_id = find_data(connection, 'input_signal_block_id')
			input_socket_index = int(find_data(connection, 'input_socket_index'))
			output_signal_block_id = find_data(connection, 'output_signal_block_id')
			output_socket_index = int(find_data(connection, 'output_socket_index'))	
			input_signal_block = signal_blocks_dict[input_signal_block_id]
			output_signal_block = signal_blocks_dict[output_signal_block_id]	
			self.connect( #use this flow graph's connect method
				(output_signal_block, output_socket_index),
				(input_signal_block, input_socket_index))
				
##############################################################################################
#	Option Parser for running a flow graph
##############################################################################################				
class FlowGraphOptionParser(OptionParser):
	"""
	This flow graph option parser provides a basic help, 
	and ensures that a flow graph is passed on the command line.
	"""
	
	def __init__(self):
		"""Create the base option parser and add usage and basic options."""			
		usage = 'usage: %prog [options] flow_graph'+FLOW_GRAPH_FILE_EXTENSION
		OptionParser.__init__(self, usage=usage)		
		self.add_option("-p", "--pid_file", action="store", type="string", dest="pid_file", help="record process id")	
		
	def parse_args(self):
		"""
		Call the base parse_args, handle the basic options,
		and ensure a flow graph was passed in args. Return the (options, args).
		"""
		(options, args) = OptionParser.parse_args(self)
		if options.pid_file:
			try: open(options.pid_file, 'w').write('%d\n' % os.getpid())
			except: 
				print '\nCould not create pid file: "%s".\nExiting!'%options.pid_file
				exit(-1)
		if not len(args): 
			self.print_help()
			sys.exit(-1)
		return options, args

##############################################################################################
#	Main
##############################################################################################					
if __name__ == '__main__':	
	parser = FlowGraphOptionParser()
	(options, args) = parser.parse_args()	
	fg = FlowGraphBuilder(args[0])
	fg.start()
	raw_input('Flow Graph Running...\nPress Enter to Exit: ')
	fg.stop()	
	sys.exit(0)
			
