#! /usr/bin/env python
"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package ExecFlowGraphGUI 
#Use an xml input file to build and run a gnu radio flow graph with wx GUI elements.
#@author Josh Blum

import Variables
from ExecFlowGraph import FlowGraphBuilder,FlowGraphOptionParser
import wx
import os
from Constants import *
import math

##############################################################################################
#	Variable Control
##############################################################################################
class VariableControl(wx.BoxSizer):
	"""House a Slider and a Text Box for variable control."""
	
	def __init__(self, parent, key, parse_callbacks):
		"""!
		VariableControl contructor.
		Create the slider, text box, and label given the variable key.
		parse_callbacks is a method to update the registered signal block callback methods.
		@param parent the wx parent window
		@param key the variable key
		@param parse_callbacksa the callback function for variables
		"""
		self.key = key
		self.parse_callbacks = parse_callbacks
		value, min, max, step = Variables.get_values(key)
		wx.BoxSizer.__init__(self, wx.VERTICAL)
		label_text_sizer = wx.BoxSizer(wx.HORIZONTAL) #label and text box container 
		label = wx.StaticText(parent, -1, key+' -> ')
		self.text_box = text_box = wx.TextCtrl(parent, -1, value, style=wx.TE_PROCESS_ENTER)
		text_box.Bind(wx.EVT_TEXT_ENTER, self._handle_enter)	#bind this special enter hotkey event
		for obj in (label, text_box):	#fill the container with label and text entry box
			label_text_sizer.Add(obj, 0, wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL)
		self.Add(label_text_sizer, 0, wx.ALIGN_CENTER)
		###	Determine the slider width based on the number of steps	###		
		self.num_steps = num_steps = (float(max)-float(min))/float(step)			
		if num_steps == DEFAULT_SLIDER_STEPS: slider_width = DEFAULT_SLIDER_WIDTH
		elif num_steps > DEFAULT_SLIDER_STEPS:	# make the slider wider than default
			slider_width = MAX_SLIDER_WIDTH - (MAX_SLIDER_WIDTH-DEFAULT_SLIDER_WIDTH)*DEFAULT_SLIDER_STEPS/num_steps
		elif num_steps < DEFAULT_SLIDER_STEPS: # make the slider slimer then default
			slider_width = MIN_SLIDER_WIDTH + (DEFAULT_SLIDER_WIDTH-MIN_SLIDER_WIDTH)*num_steps/DEFAULT_SLIDER_STEPS
		self.slider_width = slider_width = int(slider_width)	#make sure width is an integer number of pixels
		### make the slider	###				
		self.slider = slider = wx.Slider(parent, -1, size=wx.Size(slider_width, SLIDER_HEIGHT), style=wx.SL_HORIZONTAL)				
		try: slider.SetRange(0, num_steps)
		except Exception, e:
			sys.stderr.write('Slider, "%s", has too many steps!\n'%key)
			sys.exit(-1)
		slider.Bind(wx.EVT_SCROLL, self._handle_scroll)	#bind the scrolling event		
		self.Add(slider, 0, wx.ALIGN_CENTER)
		self._set_slider_value(value)#sets the slider's value
	
	def get_slider_width(self):
		"""!
		Get the width of the slider.
		@return the slider width in pixels
		"""
		return self.slider_width
		
	def _get_slider_value(self):
		"""!
		Read the value from the slider and,  
		translate the slider value into a real numerical value.
		@return the numeric representation of the slider
		"""
		slider_value = self.slider.GetValue()		
		value, min, max, step = Variables.get_values(self.key)		
		return str(slider_value*(float(max)-float(min))/self.num_steps + float(min))
		
	def _set_slider_value(self, real_value):
		"""!
		Translate the real numerical value into a slider value and,  
		write the value to the slider.
		@param real_value the numeric value the slider should represent
		"""	
		value, min, max, step = Variables.get_values(self.key)		
		slider_value = (float(real_value)-float(min))*self.num_steps/(float(max)-float(min))
		self.slider.SetValue(slider_value)
	
	def _handle_scroll(self, event=None):
		"""A scroll event is detected. Read the slider. Try to register the new entry:
		If the registration passes, update the text box, and parse the callbacks.
		If the registration fails, restore the slider to the previous registered values. """		
		key = self.key		
		new_value = str(self._get_slider_value())	#the new value must be a string!
		try:
			if Variables.reregister(key, new_value): self.parse_callbacks()
		except Exception, e: print 'Error in handle scroll: "%s".'%e 
		self._set_slider_value(Variables.get_value(key))
		self.text_box.SetValue(Variables.get_value(key))
		
	def _handle_enter(self, event=None):
		"""An enter key was pressed. Read the text box. Try to register the new entry: 
		If the registration passes, move the slider, and parse the callbacks.	
		If the registration fails, restore the text box to the previous registered value."""	
		key = self.key
		new_value = str(self.text_box.GetValue())	#the new value must be a string!	
		try:
			if Variables.reregister(key, new_value): self.parse_callbacks()
		except Exception, e: print 'Error in handle enter: "%s".'%e 
		self._set_slider_value(Variables.get_value(key))
		self.text_box.SetValue(Variables.get_value(key))

##############################################################################################
#	Flow Graph Frame
##############################################################################################	
class FlowGraphFrame(wx.Frame, FlowGraphBuilder):
	"""
	A FlowGraphFrame is a wx.Frame and a FlowGraphBuilder. 
	This flow graph frame parses a saved flow graph file,
	houses all the sliders and graphical sinks, 
	and starts the flow graph.
	"""
	
	def __init__(self, file_path):
		"""!
		FlowGraphFrame contructor.
		@param file_path the file path to the flow graph
		"""
		file_path = os.path.abspath(file_path)
		wx.Frame.__init__(self, None , -1, MAIN_WINDOW_PREFIX+' - Executing: '+file_path)
		if WX_APP_ICON: self.SetIcon(wx.Icon(WX_APP_ICON, wx.BITMAP_TYPE_ANY))
		self.SetSizeHints(200, 100)
		# create holder for sliders and graphs	#
		self.sliders_dict = dict()
		self.sink_windows = dict()
		graphs_box = wx.GridSizer(1, 1, 0, 0)
		self.panel = panel = wx.ScrolledWindow(self, -1)
		panel.SetSizer(graphs_box)
		#	create the main box for all components	#
		main_box = wx.BoxSizer(wx.VERTICAL)
		self.sliders_box = wx.BoxSizer(wx.VERTICAL)
		main_box.Add(self.sliders_box, 0, wx.ALIGN_CENTER)
		main_box.Add(panel, 0, wx.EXPAND)		
		self.Bind(wx.EVT_CLOSE, self._quit)
		#	fill in all the sliders and graphs	#		
		FlowGraphBuilder.__init__(self, file_path)		
		self._create_sliders()
		graphs_box.SetCols(int(math.sqrt(len(self.sink_windows))))
		for key in sorted(self.sink_windows.keys()): graphs_box.Add(self.sink_windows[key], 0, wx.EXPAND)		
		self.SetSizerAndFit(main_box)
		self.Show()	
		self._start() #start the flow graph	
			
	def _quit(self, event):
		"""Exit the application."""
		self._stop()	#stop the flow graph
		self.Destroy() #destroy the wx frame
		
	def add_window(self, window, type='', title=''):
		"""!
		Register the graphical sink window with a unique key.
		The key is based on the type and title.
		@param window the new gui window
		@param type an integer used to categorize the window
		@param title the title of the window
		"""
		key = str(type) + str(title)
		while self.sink_windows.has_key(key): key = key + '_'	
		self.sink_windows[key] = window		
		
	def get_panel(self):
		"""!
		Get the panel that will parent the graphical blocks.
		@return the wx panel
		""" 
		return self.panel
		
	def _create_sliders(self):
		"""
		Using the list of variable keys imported by the builder, 
		create sliders for those variables with ranges.
		"""	
		len = 0		
		for key in self.var_keys:
			value, min, max, step = Variables.get_values(key)
			if Variables.is_ranged(key):
				variable_control = VariableControl(self, key, self.parse_callbacks)
				width = variable_control.get_slider_width()		
				###	Create a new row of sliders if the current row is full	###
				if len == 0 or len + width > MAX_SLIDERS_WIDTH:
					len = 0
					slider_box = wx.BoxSizer(wx.HORIZONTAL)
					self.sliders_box.Add(slider_box, 0, wx.ALIGN_CENTER)
				slider_box.Add(variable_control, 0, wx.ALIGN_CENTER)				
				len = len + width	
		
##############################################################################################
#	Flow Graph App
##############################################################################################
class FlowGraphApp(wx.App):
	"""The FlowGraphApp is the wx.App containing the FlowGraphFrame."""
	
	def __init__(self, flow_graph_file_path):
		"""!
		FlowGraphApp constructor.
		@param flow_graph_file_path the file path to the flow graph
		"""
		self.flow_graph_file_path = flow_graph_file_path
		wx.App.__init__(self, 0)	
			
	def OnInit(self):
		"""!
		Create the top level flow graph frame.
		If the creation of the frame fails, exit with error.
		@return true if the flow graph frame is created
		"""
		try:
			self.SetTopWindow(FlowGraphFrame(self.flow_graph_file_path))#first argument is the flow graph
			return True
		except Exception:
			import traceback
			traceback.print_exc()
			sys.exit(1)
			
##############################################################################################
#	Main
##############################################################################################	
if __name__ == '__main__':	
	parser = FlowGraphOptionParser()
	parser.add_option("-d", "--disable-wx", action="store_true", dest="no_wx", default=False, help="disable the wx gui")
	(options, args) = parser.parse_args()
	app = FlowGraphApp(args[0])
	if options.no_wx: raw_input('Flow Graph Running...\nPress Enter to Exit: ')	#do not start wx
	else: app.MainLoop()	#start the wxApp MainLoop	
	sys.exit(0)
		
		
