#! /usr/bin/env python
"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package ExecFlowGraphXMLRPC 
#Use a xml input file to build and run a gnu radio flow graph without graphics.
#Create an XMLRPC sever to control the running flow graph.
#@n @n XMLRPC access functions:
#@li var_has_range(key): returns true if the variable represented by key has a range @n
#@li var_exists(key): returns true if the variable represented by key exists @n
#@li get_var_key_list(): returns a list of keys representing all registered variables @n
#@li get_var(key): returns the value of the variable represented by key @n
#@li get_var_range(key): returns the (min,max) of the variable represented by key @n
#@li set_var(key, new_value): sets the new value of the variable represented by key @n
#@li start_fg(): starts the flow graph @n
#@li stop_fg(): stops the flow graph @n
#
#@author Josh Blum

from SimpleXMLRPCServer import SimpleXMLRPCServer
from xmlrpclib import Error
from ExecFlowGraph import FlowGraphBuilder,FlowGraphOptionParser
import Variables		

DEFAULT_ADDR = 'localhost'
DEFAULT_PORT = 3030

##############################################################################################
#	Flow Graph Access
##############################################################################################	
class FlowGraphAccess(FlowGraphBuilder):			
	"""Implement a flow graph builder, and provide access functions for the xmlrpc server."""
		
	def get_var_range(self, key):
		"""!
		Get the min and the max of a variable (strings).
		@return the min,max tuple 
		"""		
		value, min, max, step = Variables.get_values(key)	
		return str(min), str(max)		
		
	def set_var(self, key, new_value):
		"""!
		Register the new value and call parse callbacks.
		Value can be any type (will be cast to a string).
		@param key the variable key
		@param new_value the new value for the variable		
		@return the new value as a string or the old value if failed
		"""		
		try: 
			if Variables.reregister(key, str(new_value)): self.parse_callbacks()
		except: pass
		return Variables.get_value(key)
			
##############################################################################################
#	Main
##############################################################################################	
if __name__ == '__main__':	
	parser = FlowGraphOptionParser()
	parser.add_option("-a", "--addr", action="store", type="string", dest="addr", default=DEFAULT_ADDR, help='bind address for XMLRPC server')
	parser.add_option("-P", "--port", action="store", type="int", dest="port", default=DEFAULT_PORT, help='tcp port for XMLRPC server')
	(options, args) = parser.parse_args()
	fg = FlowGraphAccess(args[0])	#create the flow graph	
	server = SimpleXMLRPCServer((options.addr, options.port))	#create the xmlrpc server
	# register access methods #
	server.register_function(lambda key: bool(Variables.is_ranged(key)), "var_has_range")
	server.register_function(lambda key: bool(Variables.has_key(key)), "var_exists")
	server.register_function(lambda: fg.var_keys, "get_var_key_list")
	server.register_function(lambda key: str(Variables.get_value(key)), "get_var")
	server.register_function(fg.get_var_range, "get_var_range")
	server.register_function(fg.set_var, "set_var")
	server.register_function(fg._start, "start_fg")
	server.register_function(fg._stop, "stop_fg")		
	fg._start()	#start the flow graph 
	print 'Starting XMLRPC server on address: "%s" and port: "%d"'%(options.addr, options.port)
	server.serve_forever()	#start the xmlrpc server 	
	
	