"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Graphics.Dialogs
#Misc dialogs.
#@author Josh Blum

import Colors
import pygtk
pygtk.require('2.0')
import gtk
from DataTypes import *
from Constants import *
from Elements import GraphicalParam
import Preferences

class TextDisplay(gtk.TextView):
	"""A non editable gtk text view."""
	
	def __init__(self, text=''):
		"""!
		TextDisplay constructor.
		@param text the text to display (string)
		"""
		text_buffer = gtk.TextBuffer()
		text_buffer.set_text(text)	
		self.set_text = text_buffer.set_text		
		self.insert = lambda line: text_buffer.insert(text_buffer.get_end_iter(), line)
		gtk.TextView.__init__(self, text_buffer)
		self.set_editable(False)
		self.set_cursor_visible(False)
		self.set_wrap_mode(gtk.WRAP_WORD)

######################################################################################################
class PreferencesDialog(gtk.Dialog):
	"""A dialog box to display the preferences."""
	
	def __init__(self):
		"""PreferencesDialog constructor."""
		gtk.Dialog.__init__(self, buttons=('gtk-close', gtk.RESPONSE_CLOSE))
		self.set_title("Preferences")	
		self.set_size_request(MIN_DIALOG_WIDTH, MIN_DIALOG_HEIGHT)
		notebook = gtk.Notebook()
		for title,notes,params in Preferences.PREFS_LIST: 			
			if title:				
				vbox = gtk.VBox()
				vbox.pack_start(gtk.Label(), False)	#blank label for spacing
				for param, key in params: vbox.pack_start(param.get_input_object(), False)		
				if notes: vbox.pack_start(TextDisplay(notes), False, padding=5)
				notebook.append_page(vbox, gtk.Label(title))		
		self.vbox.pack_start(notebook, True)
		self.show_all()
		self.run()
		self.destroy()

######################################################################################################
class FlowGraphWindowSizeDialog(gtk.Dialog):
	"""A dialog box to set the window size, width and heigh in pixels."""
	
	def __init__(self, dimensions):
		"""!
		FlowGraphWindowSizeDialog constructor.
		@param dimensions the (width,height) tuple from the flow graph
		"""
		gtk.Dialog.__init__(self, buttons=('gtk-close', gtk.RESPONSE_CLOSE))
		self.set_title("Set the Window Size")		
		#	the helpful dimension constraints label	#
		window_size_label = gtk.Label()
		window_size_label.set_markup("""<i>
Minimum window (width/height) in pixels is (%d/%d).
Maximum window (width/height) in pixels is (%d/%d).
</i>"""%(MIN_WINDOW_WIDTH, MIN_WINDOW_HEIGHT, MAX_WINDOW_WIDTH, MAX_WINDOW_HEIGHT))
		self.vbox.pack_end(window_size_label, False)					
		self.original_dimensions = width,height = dimensions	
		self.width = Int(width, min=MIN_WINDOW_WIDTH, max=MAX_WINDOW_WIDTH)
		self.height = Int(height, min=MIN_WINDOW_HEIGHT, max=MAX_WINDOW_HEIGHT)
		self.vbox.pack_start(GraphicalParam('width (pixels)', self.width).get_input_object(), False)
		self.vbox.pack_start(GraphicalParam('height (pixels)', self.height).get_input_object(), False)	
		self.show_all()
		
	def run(self):
		"""!
		Get the new dimensions.
		@return the new dimensions (width,height) or None if invalid.
		"""
		self.dimensions = None			
		gtk.Dialog.run(self)
		if self.width.is_valid() and self.height.is_valid():
			self.dimensions = (self.width.parse(), self.height.parse())
		self.destroy()
		if self.original_dimensions == self.dimensions: return None	# do not return dimensions if no change
		return self.dimensions		

######################################################################################################
def MessageDialogHelper(type, buttons, title=None, markup=None):
	"""!
	Create a modal message dialog and run it.
	@param type the type of message: gtk.MESSAGE_INFO, gtk.MESSAGE_WARNING, gtk.MESSAGE_QUESTION or gtk.MESSAGE_ERROR
	@param buttons the predefined set of buttons to use: gtk.BUTTONS_NONE, gtk.BUTTONS_OK, gtk.BUTTONS_CLOSE, gtk.BUTTONS_CANCEL, gtk.BUTTONS_YES_NO, gtk.BUTTONS_OK_CANCEL
	@param tittle the title of the window (string)
	@param markup the message text with pango markup
	@return the gtk response from run()
	"""
	message_dialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, type, buttons)
	if title != None: message_dialog.set_title(title)
	if markup != None: message_dialog.set_markup(markup)
	response = message_dialog.run()
	message_dialog.destroy()
	return response
		
######################################################################################################
class AboutDialog(gtk.AboutDialog):
	"""A cute little about dialog."""
	
	def __init__(self):
		"""AboutDialog constructor."""
		gtk.AboutDialog.__init__(self)
		self.set_version(VERSION)
		self.set_name(MAIN_WINDOW_PREFIX)	
		self.set_license(__doc__)
		self.set_copyright('Copyright 2007 Free Software Foundation, Inc.')
		self.set_website('http://gnuradio.org/trac/wiki/GNURadioCompanion')
		self.set_comments("""\
Thank you to all those from the mailing list who tested GNU Radio Companion and offered advice.
--
Special Thanks:
A. Brinton Cooper -> starting the project
CER Technology Fellowship Grant -> initial funding
William R. Kenan Jr. Fund -> usrp & computers
Patrick Strasser -> the GRC icon
Achilleas Anastasopoulos -> trellis support
--""")
		self.run()
		self.destroy()
					
######################################################################################################			
class DataTypesDialog(gtk.Dialog):
	"""Display each data type with its associated color."""
	
	def __init__(self):
		"""DataTypesDialog constructor."""
		gtk.Dialog.__init__(self, buttons=('gtk-close', gtk.RESPONSE_CLOSE))
		self.set_title('Data Types')
		self.set_size_request(170, -1)
		markup = ''
		for color_spec in (
			'Regular Types:',
			('Complex', Colors.COMPLEX_COLOR_SPEC),
			('Float', Colors.FLOAT_COLOR_SPEC),
			('Int', Colors.INT_COLOR_SPEC),
			('Short', Colors.SHORT_COLOR_SPEC),										
			('Byte', Colors.BYTE_COLOR_SPEC),
			'Vector Types:',
			('Complex Vector', Colors.COMPLEX_VECTOR_COLOR_SPEC),
			('Float Vector', Colors.FLOAT_VECTOR_COLOR_SPEC),
			('Int Vector', Colors.INT_VECTOR_COLOR_SPEC),
			('Short Vector', Colors.SHORT_VECTOR_COLOR_SPEC),										
			('Byte Vector', Colors.BYTE_VECTOR_COLOR_SPEC),
		):
			width = 20 #width in monospace characters
			if type(color_spec) == type(str()): markup = '%s\n\n<b>%s</b>'%(markup, color_spec.center(width, ' '))
			else:	
				tag,spec = color_spec					
				markup = '%s\n<span background="%s">%s</span>'%(markup, spec, tag.center(width, ' '))
		label = gtk.Label()
		label.set_markup('<tt>%s</tt>\n'%markup[1:])	#strip 1st newline, append newline
		self.vbox.pack_start(label, False)
		self.show_all()
		self.run()
		self.destroy()			
			
######################################################################################################			
class HotKeysDialog(gtk.Dialog):
	"""Display each action with the associated hotkey."""
	
	def __init__(self):
		"""HotKeysDialog constructor."""
		gtk.Dialog.__init__(self, buttons=('gtk-close', gtk.RESPONSE_CLOSE))
		self.set_title('Hot Keys')
		markup = ''
		for action, hotkey in (
			('New Flow Graph', 'Ctrl + n'),
			('Open Flow Graph', 'Ctrl + o'),
			('Save Flow Graph', 'Ctrl + s'),
			('Undo Change', 'Ctrl + z'),
			('Redo Change', 'Ctrl + y'),
			('Redo Change', 'Ctrl + Z'),
			('Delete Block', 'Delete'),
			('Rotate Block', 'Right'),
			('Rotate Block', 'Left'),
			('Modify Data Type', 'Up'),
			('Modify Data Type', 'Down'),
			('Add a Socket', '+'),
			('Remove a Socket', '-'),
			('Close Dialog', 'Esc'),
			('Flow Graph Run', 'F5'),
			('Flow Graph Stop', 'F7'),
			('Screen Shot', 'PrintScreen'),
		): markup = '%s\n<b>%s:</b>%s'%(markup, action, hotkey.rjust(25-len(action),' '))
		label = gtk.Label()
		label.set_markup('<tt>%s</tt>\n'%markup)	#append newline
		self.vbox.pack_start(label, False)
		self.show_all()
		self.run()
		self.destroy()		
		
######################################################################################################			
class MathExprDialog(gtk.Dialog):
	"""A dialog to test math expressions for the parser."""
	
	##the help message
	HELP_MSG = """\
<b>Operators - Complex   <i>(complex arguments)</i></b> 
	+ - * / ^	
<b>Trigonometric Functions   <i>(complex arguments, radians)</i></b> 
	sin(z) cos(z) tan(z) asin(z) acos(z) atan(z) 
	sinh(z) cosh(z) tanh(z) asinh(z) acosh(z) atanh(z)	
<b>Complex Components   <i>(complex arguments)</i></b>
	mag(z*) real(z*) imag(z*) conj(z*) arg(z)
<b>Misc Functions   <i>(complex arguments)</i></b> 
	abs(z*) pow(b, p) sqrt(z) log10(z) log(z, [b=e]) ln(z) exp(z) 
<b>Misc Functions   <i>(floating point arguments)</i></b> 
	floor(f) ceil(f) radians(deg) degrees(rad) mod(f1, f2) atan2(y, x) sgn(f) min(f*) max(f*)
<b>Filter Taps Generators   <i>(floating point arguments, rates and frequencies in Hz)</i></b> 
	low_pass(gain, samp_rate, cutoff_freq, width, [window=hamming], [beta=6.76])
	high_pass(gain, samp_rate, cutoff_freq, width, [window=hamming], [beta=6.76])	
	band_pass(gain, samp_rate, low_cutoff_freq, high_cutoff_freq, width, [window=hamming], [beta=6.76])
	complex_band_pass(gain, samp_rate, low_cutoff_freq, high_cutoff_freq, width, [window=hamming], [beta=6.76])
	band_reject(gain, samp_rate, low_cutoff_freq, high_cutoff_freq, width, [window=hamming], [beta=6.76])	
	gaussian(gain, spb, bt, int ntaps)
	hilbert(int ntaps, [window=rectangular], [beta=6.76])
	root_raised_cosine(gain, samp_rate, symbol_rate, alpha, int ntaps) 
	window(window, int ntaps, beta)	
<b>Window Types for Filters</b>
	hamming hann blackman rectangular kaiser
<b>Other Constants</b>
	pi e j	
<b>Using Vectors</b>
	() (num1,) (num2, num2, ...)	
<b>Test Your Expressions Below:</b>\
"""

	def __init__(self):
		"""MathExprDialog constrcutor. Create a new gtk Dialog with a close button, an extry box, and an output text box."""		
		gtk.Dialog.__init__(self, buttons=('gtk-close', gtk.RESPONSE_CLOSE))
		#	Create the title and label	#
		self.set_title('Mathematical Expressions')
		label = gtk.Label()
		label.set_markup(self.HELP_MSG)
		self.vbox.pack_start(label, False)		
		self.set_size_request(800, 600)
		#	create a text box for parser output	#
		self.text_box = TextDisplay()
		self.text_box.set_text('')			
		# create the entry box and give it the change handler #
		self.param = GraphicalParam('Expression', RawExpr(''))
		self.input_obj = self.param.get_input_object(self._handle_changed)
		self.vbox.pack_start(self.input_obj, False)			
		# add the scrolled window for the text box #
		scrolled_window = gtk.ScrolledWindow()
		scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		scrolled_window.add_with_viewport(self.text_box)
		self.vbox.pack_start(scrolled_window, True)	
		self.show_all()
		self.run()
		self.destroy()
		
	def _handle_changed(self, param=None):
		"""Handle changed in the param's entry box by updating the text box."""
		self.text_box.set_text(str(self.param.get_data_type()))	
		