"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Graphics.FlowGraph
#A flow graph structure for storing signal blocks and their connections.
#@author Josh Blum

from Constants import *
from Actions import *
from Colors import BACKGROUND_COLOR, TXT_COLOR
from Elements import GraphicalSignalBlock,GraphicalConnection
from Elements import Utils
from Elements.Connection import ConnectionException
import Variables
import pygtk
pygtk.require('2.0')
import gtk
import time,socket #for tagging saved files
import SignalBlockDefs
from SignalBlockParamsDialog import SignalBlockParamsDialog
import random
import Messages
import Preferences
import ParseXML
import Messages

class FlowGraph(gtk.DrawingArea):
	"""
	FlowGraph is the data structure to store graphical signal blocks, 
	graphical inputs and outputs, 
	and the connections between inputs and outputs.
	"""	
		
	def __init__(self, handle_states, variable_modification_window):
		"""!
		FlowGraph contructor.
		Create a list for signal blocks and connections. Connect mouse handlers.
		@param handle_states the callback function
		@param variable_modification_window var mod window also for callbacks
		"""	
		#setup
		self.elements = list()
		self.remove_element = lambda e: self.elements.remove(e)
		self.gc = None
		self.handle_states = handle_states
		self.variable_modification_window = variable_modification_window
		gtk.DrawingArea.__init__(self)
		self.connect('expose-event', self._handle_window_expose)
		self.connect('motion-notify-event', self._handle_mouse_motion)
		self.connect('button-press-event', self._handle_mouse_button_press)  
		self.connect('button-release-event', self._handle_mouse_button_release)		
		self.set_events(
			gtk.gdk.BUTTON_PRESS_MASK | \
			gtk.gdk.POINTER_MOTION_MASK | \
			gtk.gdk.BUTTON_RELEASE_MASK | \
			gtk.gdk.LEAVE_NOTIFY_MASK | \
			gtk.gdk.ENTER_NOTIFY_MASK
		)	
		#	setup the focus flag	#		
		self.focus_flag = False
		self.get_focus_flag = lambda: self.focus_flag		
		self.connect("leave-notify-event", self._handle_focus_event, False)
		self.connect("enter-notify-event", self._handle_focus_event, True)		
		#	important vars dealing with mouse event tracking	#	
		self.has_moved = False
		self.mouse_pressed = False
		self.coordinate = (0,0)
		self.selected_element = None
		self.is_selected = lambda: self.selected_element != None
		self.count = 0
		self.pixmap = None
		
###########################################################################
#	Flow Graph Access Methods
###########################################################################	
	
	def _handle_focus_event(self, widget, event, focus_flag):
		"""Record the focus state of the flow graph window."""
		self.focus_flag = focus_flag
				
	def add_signal_block(self, tag):	
		"""!
		Add a signal block of the given tag to this flow graph.
		@param tag the signal block tag
		"""
		index = 0
		while True:
			id = tag+str(index)
			index = index + 1	
			if not [
				None for element in self.elements if Utils.is_signal_block(element) and id == element.get_id()
			]:		#make sure that the id is not used by another signal block
				rot = 0
				vAdj = self.get_parent().get_vadjustment().get_value()
				hAdj = self.get_parent().get_hadjustment().get_value()
				x = random.randint(100,400)+int(hAdj)	
				y = random.randint(100,400)+int(vAdj)	
				self.elements.append(
					SignalBlockDefs.get_signal_block(self, (x, y), rot, tag, id, GraphicalSignalBlock)[0]
				)
				self.handle_states(SIGNAL_BLOCK_CREATE)	
				self.update()
				return		
	
	def type_controller_modify_selected(self, direction):
		"""
		!Change the registered type controller for the selected signal block.
		@param direction +1 or -1
		@return true for success
		"""
		if Utils.is_socket(self.selected_element): self.selected_element = self.selected_element.get_parent()
		if Utils.is_signal_block(self.selected_element): return self.selected_element.modify_type_controller(direction)
		return False
		
	def socket_controller_modify_selected(self, direction):
		"""!
		Change socket controller for the selected signal block.
		@param direction +1 or -1
		@return true for success
		"""
		if Utils.is_socket(self.selected_element): self.selected_element = self.selected_element.get_parent()
		if Utils.is_signal_block(self.selected_element): return self.selected_element.modify_socket_controller(direction)
		return False

	def param_modify_selected(self):
		"""!
		Create and show a param modification dialog for the selected element (socket and signal block only).
		@return true if parameters were changed
		"""
		if Utils.is_socket(self.selected_element): self.selected_element = self.selected_element.get_parent()
		if Utils.is_signal_block(self.selected_element):
			signal_block_params_dialog = SignalBlockParamsDialog(self.selected_element)
			changed = signal_block_params_dialog.run()	
			self.update()	
			return changed #changes were made?
		return False	
		
	def connect_sockets(self, s1, s2):
		"""!
		Connect input and output sockets. Ignore state change if creation of connection fails.
		@param s1 a socket
		@param s2 another socket
		"""
		try:		
			connection = GraphicalConnection(self, s1, s2)
			self.elements.append(connection)
			self.selected_element = None
			self.handle_states(CONNECTION_CREATE)	
			self.update()	
		except ConnectionException: 
			Messages.send_fail_connection()	
			self.handle_states(NOTHING_SELECT)	
	
	def move_selected(self, delta_coordinate):
		"""!
		Move the element and by the change in coordinates.
		@param delta_coordinate the change in coordinates
		"""
		if self.selected_element != None:					
			self.selected_element.move(delta_coordinate)	
			self.has_moved = True	
			self.draw()
			
	def rotate_selected(self, direction):
		"""!
		Rotate the selected element by 90 degrees. Only rotate SignalBlocks and Sockets.
		@param direction DIR_LEFT or DIR_RIGHT
		@return 	true if rotated, otherwise false. 
		"""
		if self.selected_element != None and (Utils.is_signal_block(self.selected_element) or Utils.is_socket(self.selected_element)):					
			self.selected_element.rotate(direction)	
			self.draw()
			return True
		return False	
	
	def delete_selected(self):
		"""!
		If an element is selected, remove it and its attached parts from the element list.
		@return true if the element was deleted, otherwise false.
		"""
		if self.selected_element != None:
			if Utils.is_socket(self.selected_element):	# found a socket, set to parent signal block
				self.selected_element = self.selected_element.get_parent()						
			if Utils.is_signal_block(self.selected_element):	# delete a signal block
				connections = self.selected_element.get_connections()
				for connection in connections: connection.disconnect()
				self.remove_element(self.selected_element)
			elif Utils.is_connection(self.selected_element):	# delete a connection
				self.selected_element.disconnect()					
			self.selected_element = None				
			self.update()	
			return True
		return False
		
	def unselect(self):
		"""If an element is selected, un-highlight it and set selected to None."""
		if self.selected_element != None: 
			self.selected_element.set_highlighted(False)			
			self.selected_element = None	
			self.update()	
					
	def what_is_selected(self, coor):		
		"""!
		What is selected?
		At the given coordinate, return the first element found to be selected	
		- iterate though the elements backwards since top elements are at the end of the list	
		- if an element is selected, place it at the end of the list so that is is drawn last, 
			and hence on top.
		@param coor the coordinate of the mouse click
		@return the selected element or None
		"""		 
		#	check the elements	#			
		for element in reversed(self.elements):
			if element.what_is_selected(coor) != None:
				self.elements.remove(element)
				self.elements.append(element)
				return element.what_is_selected(coor)		
		return None
		
	def draw(self, drawable=None):		
		"""Draw the background and then all of the Elements in this FlowGraph on the pixmap, 
		then draw the pixmap to the drawable window of this FlowGraph."""
		if self.gc != None:				
			#	draw the background	#		
			W,H = self.get_size_request()
			self.gc.foreground = BACKGROUND_COLOR
			self.pixmap.draw_rectangle(self.gc, True, 0, 0, W, H)	
			if Preferences.show_grid():				
				grid_size = Preferences.get_grid_size()				
				points = list()
				for i in range(W/grid_size):
					for j in range(H/grid_size): 
						points.append((i*grid_size,j*grid_size))					
				self.gc.foreground = TXT_COLOR					
				self.pixmap.draw_points(self.gc, points)	
			#	draw the foreground	#
			for element in filter(Utils.is_signal_block, self.elements) + filter(Utils.is_connection, self.elements): 
				element.draw(self.pixmap)	# draw signal blocks first, then connections on the top			
			if self.mouse_pressed and self.selected_element != None:
				self.selected_element.draw(self.pixmap)
			self.window.draw_drawable(self.gc, self.pixmap, 0, 0, 0, 0, -1, -1)	 
			
	def is_valid(self):
		"""!
		Is the flow graph valid, ie: are all elements valid?
		@return true if flow graph is valid
		"""		
		for element in self.elements: 
			if not element.is_valid(): return False
		return True
		#return all([element.is_valid() for element in self.elements])	#python 2.5 and higher
		
	def update(self):
		"""Call update on all elements."""	
		for element in self.elements: element.update() 	
		self.draw()
		
	##########################################################################
	##	Handlers
	##########################################################################	
		
	def _handle_mouse_button_press(self, widget, event):
		""" A mouse button is pressed. Record the state of the mouse, find the selected element,
		set the Element highlighted, handle the state change, and redraw the FlowGraph."""		
		#	unselect anything in the vars mod window	#
		self.variable_modification_window.unselect_all()
		if event.button == 1:
			if self.selected_element != None: self.selected_element.set_highlighted(False)
			self.count = 0
			self.mouse_pressed = True
			self.coordinate = (event.x, event.y)
			old_selection = self.selected_element
			self.selected_element = self.what_is_selected((event.x, event.y))
			#	handle the state change with the new selection	#
			if Utils.is_connection(self.selected_element): self.handle_states(CONNECTION_SELECT)	
			elif Utils.is_socket(self.selected_element): self.handle_states(SOCKET_SELECT)	
			elif Utils.is_signal_block(self.selected_element): self.handle_states(SIGNAL_BLOCK_SELECT)	
			elif self.selected_element == None: self.handle_states(NOTHING_SELECT)						
			#	this selection and the last were Sockets, try to connect them	#
			if Utils.is_socket(old_selection) and Utils.is_socket(self.selected_element) and\
				old_selection is not self.selected_element: #cannot be the same socket
				self.connect_sockets(old_selection, self.selected_element)						
			if self.selected_element != None: self.selected_element.set_highlighted(True)									
			#	double click detected, bring up params dialog if possible	#
			if event.type == gtk.gdk._2BUTTON_PRESS and Utils.is_signal_block(self.selected_element):
				self.mouse_pressed = False
				self.handle_states(SIGNAL_BLOCK_PARAM_MODIFY)				
			self.draw()
		return True
		
	def _handle_mouse_button_release(self, widget, event):
		"""A mouse button is released, record the state."""
		if event.button == 1:
			self.mouse_pressed = False
			if self.has_moved:
				if Preferences.snap_to_grid():
					grid_size = Preferences.get_grid_size()
					X,Y = self.selected_element.get_coordinate()
					deltaX = X%grid_size
					if deltaX < grid_size/2: deltaX = -1 * deltaX
					else: deltaX = grid_size - deltaX
					deltaY = Y%grid_size
					if deltaY < grid_size/2: deltaY = -1 * deltaY
					else: deltaY = grid_size - deltaY
					self.move_selected((deltaX,deltaY))
				self.handle_states(SIGNAL_BLOCK_MOVE)
				self.has_moved = False
		return True
		
	def _handle_mouse_motion(self, widget, event):		
		"""The mouse has moved. If mouse_pressed is true, react to the motions:	
		-	if an Element is highlighted, this will move the Element by redrawing it at the new position."""
		fgW,fgH = self.get_size_request()
		self.count = (1 + self.count)%MOTION_DETECT_REDRAWING_SENSITIVITY
		#	to perform a movement, the mouse must be pressed, an element selected, count of zero.	#
		if self.mouse_pressed and\
			self.count == 0 and\
			self.selected_element != None:
			#	The event coordinates must be within 10 pixels away from the bounds of the flow graph.	#
			x = event.x
			if x <= BORDER_PROXIMITY_SENSITIVITY:
				x = BORDER_PROXIMITY_SENSITIVITY
			if	x >= fgW - BORDER_PROXIMITY_SENSITIVITY:
				x = fgW - BORDER_PROXIMITY_SENSITIVITY					
			y = event.y
			if y <= BORDER_PROXIMITY_SENSITIVITY:
				y = BORDER_PROXIMITY_SENSITIVITY
			if	y >= fgH - BORDER_PROXIMITY_SENSITIVITY:	
				y = fgH - BORDER_PROXIMITY_SENSITIVITY	
			#	get the change in coordinates 	#		
			X,Y = self.coordinate
			deltaX = int(x - X)
			deltaY = int(y - Y)											
			vAdj = self.get_parent().get_vadjustment().get_value()
			hAdj = self.get_parent().get_hadjustment().get_value()
			width = self.get_parent().get_hadjustment().page_size
			height = self.get_parent().get_vadjustment().page_size
			# scroll horizontal if we moved near the border	#
			if x-hAdj > width-SCROLL_PROXIMITY_SENSITIVITY and\
				hAdj+SCROLL_DISTANCE < fgW - width:
				self.get_parent().get_hadjustment().set_value(hAdj+SCROLL_DISTANCE)
			elif x-hAdj < SCROLL_PROXIMITY_SENSITIVITY:
				self.get_parent().get_hadjustment().set_value(hAdj-SCROLL_DISTANCE)
			# scroll vertical if we moved near the border	#
			if y-vAdj > height-SCROLL_PROXIMITY_SENSITIVITY and\
				vAdj+SCROLL_DISTANCE < fgH - height:
				self.get_parent().get_vadjustment().set_value(vAdj+SCROLL_DISTANCE)
			elif y-vAdj < SCROLL_PROXIMITY_SENSITIVITY:
				self.get_parent().get_vadjustment().set_value(vAdj-SCROLL_DISTANCE)
			# move the selected element and record the new coordinate	#			
			self.move_selected((deltaX, deltaY))	
			self.coordinate = (x, y)	
					
	def _handle_window_expose(self, widget, event):	
		"""Called when the window initially appears or is resized: create a new pixmap, draw the flow graph."""
		self.gc = self.window.new_gc()
		width, height = self.get_size_request()
		if self.pixmap == None or (width, height) != self.pixmap.get_size():		
			self.pixmap = gtk.gdk.Pixmap(self.window, width, height, -1)
		self.draw()

##########################################################################
##
##			Export the Flow Graph
##
##########################################################################						
	def to_nested_data(self):
		"""!
		Dump all the values in this flow graph into a nested data format.
		@return nested data representing a flow graph
		"""
		vars_list = list()
		signal_blocks_list = list()
		connections_list = list()
		W,H = self.get_size_request()
		nested_data = ('flow_graph', [
			('timestamp', str(time.time())),
			('hostname', socket.gethostname()),
			('version', VERSION),
			('valid', str(self.is_valid())),
			('window_width', str(W)),
			('window_height', str(H)),
			('vars', vars_list),
			('signal_blocks', signal_blocks_list),
			('connections', connections_list)
		])  	
	##########################################################################
	##	Export the Variables
	##########################################################################	
		for key in self.variable_modification_window.to_key_list():
			row = (key,) + Variables.get_values(key)
			vars_list.append(('var', [
				('key', row[0]), 
				('value', row[1]),
				('min', row[2]),
				('max', row[3]),
				('step', row[4]),
			]))
	##########################################################################
	##	Export the Signal Blocks
	##########################################################################
		for element in filter(Utils.is_signal_block, self.elements):
			params_list = list()
			signal_blocks_list.append(('signal_block', [
				('tag', element.get_tag()), 
				('id', element.get_id()),
				('x_coordinate', str(element.get_coordinate()[0])),
				('y_coordinate', str(element.get_coordinate()[1])),
				('rotation', str(element.get_rotation())),
				('params', params_list)
			]))
			for param in element.get_params():
				params_list.append(('param', str(param.get_data_type().get_data()))	)	
	##########################################################################
	##	Export the Connections
	##########################################################################
		for element in filter(Utils.is_connection, self.elements):
			connections_list.append(('connection', [
				('input_signal_block_id', str(element.get_input_socket().get_parent().get_id())), 
				('input_socket_index', str(element.get_input_socket().get_index())),
				('output_signal_block_id', str(element.get_output_socket().get_parent().get_id())), 
				('output_socket_index', str(element.get_output_socket().get_index()))
			]))
		#print 'To',nested_data
		return nested_data

##########################################################################
##	
##			Import the Flow Graph
##
##########################################################################		
	def from_nested_data(self, nested_data):
		"""!
		Set all the values in this flow graph using the nested data.
		@param nested_data nested data representing a flow graph
		"""
		#print 'From',nested_data
		#TODO: use a non-destructive method to clear the elements list
		self.elements = list()	#clear the elements
		find_data = ParseXML.find_data
		flow_graph = find_data([nested_data], 'flow_graph')
		#	window width and height are optional	#
		window_width = find_data(flow_graph, 'window_width')
		if window_width == None: window_width = MAX_WINDOW_WIDTH
		window_height = find_data(flow_graph, 'window_height')
		if window_height == None: window_height = MAX_WINDOW_HEIGHT
		self.set_size_request(int(window_width),int(window_height))
		vars = find_data(flow_graph, 'vars')
		signal_blocks = find_data(flow_graph, 'signal_blocks')
		connections = find_data(flow_graph, 'connections')
	##########################################################################
	##	Import the Variables
	##########################################################################
		keys = Variables.from_nested_data(vars)
		self.variable_modification_window.from_key_list(keys)
	##########################################################################
	##	Import the Signal Blocks
	##########################################################################
		for signal_block in signal_blocks:
			signal_block = GraphicalSignalBlock.from_nested_data(self, signal_block, GraphicalSignalBlock)[0]	#index 0 for just signal block	
			if signal_block: self.elements.append(signal_block)			
	##########################################################################
	##	Import the Connections
	##########################################################################
		for connection in connections:
			connection = find_data([connection], 'connection')
			input_signal_block_id = find_data(connection, 'input_signal_block_id')
			input_socket_index = int(find_data(connection, 'input_socket_index'))
			output_signal_block_id = find_data(connection, 'output_signal_block_id')
			output_socket_index = int(find_data(connection, 'output_socket_index'))
			input_socket = output_socket = None										
			for element in filter(Utils.is_signal_block, self.elements):
				if element.get_id() == input_signal_block_id: input_socket = element.get_input_socket(input_socket_index)
				if element.get_id() == output_signal_block_id: output_socket = element.get_output_socket(output_socket_index)
			try: self.elements.append(GraphicalConnection(self, input_socket, output_socket))
			except ConnectionException: 
				Messages.send_error_load('Could not connect "%s" input[%d] and "%s" output[%d].'%(
						input_signal_block_id, 
						input_socket_index, 
						output_signal_block_id, 
						output_socket_index,
					)
				)	
		self.selected_element = None		
		self.update()
		# done importing the flow graph	#

