"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Graphics.VariableModificationWindow
#The variable modification window allows the user to enter variables (keys, values, ranges).
#@author Josh Blum

from Constants import *
from Actions import VARIABLE_MODIFY,VARIABLE_REORDER,NOTHING_SELECT
import pygtk
pygtk.require('2.0')
import gtk
import gobject
import Variables

class VariableModificationWindow(gtk.VBox):
	"""A scrollable window to add/remove variables in the variable registry."""
	
	def __init__(self, handle_states):
		"""!
		VariableModificationWindow constructor.Create a scrolled window, and add buttons and lists.
		@param handle_states callback function to handle the variable changes
		"""
		gtk.VBox.__init__(self)	
		self.handle_states = handle_states			
		#title label
		label = gtk.Label()
		label.set_markup('<b>Variables</b>')
		self.pack_start(label, False)		
		#create a srolled window to hold the tree view
		scrolled_window = gtk.ScrolledWindow()
		scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		scrolled_window.set_size_request(
			VARIABLE_MODIFICATION_WINDOW_WIDTH, 
			VARIABLE_MODIFICATION_WINDOW_HEIGHT
		)
		self.pack_start(scrolled_window)		
		#liststore holds the graphical list
		self.liststore = gtk.ListStore(
			gobject.TYPE_STRING, 	#key
			gobject.TYPE_STRING, 	#value
			gobject.TYPE_STRING, 	#min
			gobject.TYPE_STRING,		#max
			gobject.TYPE_STRING,		#step
		)		
		self.treeview = gtk.TreeView(self.liststore)	
		self.treeview.set_enable_search(False) #disable pop up search box
		self.renderers = list()
		for i,title in enumerate(['Variable', 'Default', 'Min', 'Max', 'Step']):	#create each data column
			#renderer for column i		
			renderer = gtk.CellRendererText()
			renderer.set_property('editable', True)
			self.renderers.append(renderer)
			#connect signal handlers
			renderer.connect('edited', self._handle_renderer_edited)
			renderer.connect('editing-started', self._handle_start_editing)
			renderer.connect('editing-canceled', self._handle_cancel_editing)
			#tree view for column i
			tree_view_column = gtk.TreeViewColumn(title, renderer, text=i)
			tree_view_column.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE) #allow the column to shrink
			self.treeview.append_column(tree_view_column)
		self.treeview.set_reorderable(True)	
		selection = self.treeview.get_selection()
		selection.set_mode('single')
		selection.connect('changed', self._handle_selection_change)
		scrolled_window.add_with_viewport(self.treeview)
		#buttons add and remove
		buttons_hbox = gtk.HBox()
		self.pack_start(buttons_hbox, False)	
		add_button = gtk.Button(None, 'gtk-add')		
		add_button.connect('clicked', self._handle_add_button)
		buttons_hbox.pack_start(add_button, False)
		self.remove_button = gtk.Button(None, 'gtk-remove')
		self.remove_button.connect('clicked', self._handle_remove_button)
		buttons_hbox.pack_start(self.remove_button, False)
		#connect signal handlers
		self.treeview.connect("cursor-changed", self._handle_cursor_changed)	
		self.treeview.connect("drag_begin", self._handle_drag_begin)	
		self.treeview.connect("drag_end", self._handle_drag_end)
		#initialize
		self._handle_cancel_editing()
		self._handle_selection_change()	
		
	#############################################################################################	
	#	Access Methods
	#############################################################################################	
		
	def to_key_list(self):
		"""!
		Get a list of the keys in the current order. 
		@return a list of ordered variable keys
		"""
		key_list = list()
		for row in self.liststore: key_list.append(row[0])
		return key_list
	
	def from_key_list(self, key_list):
		"""!
		Clear the rows and refill the list with keys from the key list.
		@param key_list a list of ordered keys
		"""		
		self.liststore.clear()
		for key in key_list: self.liststore.append((key,)+Variables.get_values(key))
							
	def unselect_all(self):		
		"""Stop any editing and unselect every row in the tree view."""
		if self.editable: self.editable.remove_widget()
		self.treeview.get_selection().unselect_all()	

	#############################################################################################	
	#	Row Manipulation Handlers
	#############################################################################################	
		
	def _handle_drag_begin(self, drag_context, data):
		"""A drag event has begun, make a backup of the key list."""
		self.backup_key_list = self.to_key_list()
		
	def _handle_drag_end(self, drag_context, data):
		"""A drag event has ended. 
		If the backup key list does not match the current key list, 
		handle the variable state change."""
		if self.backup_key_list != self.to_key_list(): self.handle_states(VARIABLE_REORDER)	
		
	def _handle_cursor_changed(self, treeview):
		"""Called when the cursor is placed on a row or in a cell.
		Unselect any selected element in the flow graph."""
		self.handle_states(NOTHING_SELECT)
				
	def _handle_selection_change(self, selection=None):
		"""If a selection changes, set the edit and remove buttons sensitive."""
		selection = self.treeview.get_selection()
		model, iter = selection.get_selected()
		if iter != None: self.remove_button.set_sensitive(True)
		else: self.remove_button.set_sensitive(False)
			
	#############################################################################################	
	#	Cell Editing Handlers
	#############################################################################################				
			
	def _handle_start_editing(self, cellrenderer, editable, path):
		"""Editing of a cell has begun, record the editable object."""
		self.editable = editable
		
	def _handle_cancel_editing(self, cellrenderer=None):
		"""Editing of a cell has been canceled."""
		self.editable = None					
			
	def _handle_renderer_edited(self, renderer, row_index, new_text):
		"""!
		A cell was edited, determine what was edited and how to handle it.
		@param renderer the renderer
		@param row_index the index of the row that changes
		@param new_text the new text in the cell
		"""
		self._handle_cancel_editing()
		column_index = self.renderers.index(renderer)
		if new_text == self.liststore[row_index][column_index]: return #no change
		old_list = self.liststore[row_index]
		new_list = list()
		for i,var in enumerate(old_list):	#fill the new list
			if i == column_index: new_list.append(new_text)
			#	blank the step to regenerate it, only if min/max was changed to a non-blank	#
			elif new_text != '' and i == 4 and column_index in (2,3): new_list.append('') 
			else: new_list.append(var)	
		Variables.unregister(old_list[0]) #use the key to remove the entry		
		try: Variables.register(*new_list)
		except Exception, e:	#register failed, restore values and end
			Variables.register(*old_list)
			print e
			return
		#register succeeded, reload the values in the window	
		values = list(Variables.get_values(new_list[0]))
		for i,value in enumerate([new_list[0]]+values):		
			self.liststore[row_index][i] = value	
		self.handle_states(VARIABLE_MODIFY)	
			
	#############################################################################################	
	#	Button Handlers
	#############################################################################################				
				
	def _handle_remove_button(self, widget):
		"""Handle the remove button by removing the selected element from the list and from the variables registry."""
		selection = self.treeview.get_selection()
		model, iter = selection.get_selected()
		if iter != None: 
			Variables.unregister(model.get_value(iter, 0))
			model.remove(iter)			
			self.handle_states(VARIABLE_MODIFY)
			
	def _handle_add_button(self, widget):	
		"""Handle the add button by adding a variable with key and value. 
		The key will be unique in the form 'new_var' + index."""
		new_key_index = 0
		while True:
			new_key_name = 'new_var'+str(new_key_index)
			if Variables.has_key(new_key_name):
				new_key_index = new_key_index + 1
			else:
				var_info = (new_key_name, '0', '', '', '')
				Variables.register(*var_info)
				self.liststore.append((new_key_name,)+Variables.get_values(new_key_name))
				self.handle_states(VARIABLE_MODIFY)
				return	#leave the while loop
				
				
				