"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package MathExprParser
#Evaluate mathematical expressions and provide utilties to cast numbers.
#@author Josh Blum

import math,cmath
from gnuradio import gr

_ORDER_OF_OPS = ('^-', '^', '*', '/', '-', '+')

_WHITESPACE = (' ', '\t', '')

_CONSTANTS = {	#constants must be lower case
	'pi' : cmath.pi,
	'e' : cmath.e,
	'j' : 1j,
	
	'hamming' : gr.firdes.WIN_HAMMING,	#filter windows
	'hann' : gr.firdes.WIN_HANN,
	'blackman' : gr.firdes.WIN_BLACKMAN,
	'rectangular' : gr.firdes.WIN_RECTANGULAR,
	'kaiser' : gr.firdes.WIN_KAISER,
}
	
#########################################################
#	Number parsers
#########################################################

def verify_complex(num):
	"""!
	Verify that the number is a complex number.
	@param num the potential complex number
	@throw ValueError not a number
	@return a complex number
	"""
	try: c_num = complex(num)	#ensure that the num can be complex
	except: raise ValueError, '"%s" is not a number.'%(num,)
	return c_num

def verify_float(num):
	"""!
	Verify the number as a float.
	The number can have no imaginary part.
	@param num the potential float number
	@throw ValueError imaginary part is non zero.
	@return a float number
	"""
	c_num = verify_complex(num)	#ensure that the num can be complex
	if c_num.imag != 0: raise ValueError, 'Expected a float but found "%s".'%(c_num,)
	return float(c_num.real)

def verify_int(num):
	"""!
	Verify the number as an integer.
	The number can have no imaginary part.
	The number can have no decimal part.
	@param num the potential integer number
	@throw ValueError imaginary or decimal part is non zero.
	@return an integer number
	"""
	f_num = verify_float(num)
	if f_num - int(f_num) != 0: raise ValueError, 'Expected an integer but found "%s".'%(f_num,)
	return int(f_num)
	
def verify_number(num):
	"""!
	Try to cast the number into its most primative form: int, float, or complex. 
	@throw ValueError num is not a number.
	@return a number of int, float, or complex type 
	"""
	try: return verify_int(num)	#try int
	except:
		try: return verify_float(num)	#try float
		except: return verify_complex(num)	#complex or fail

#########################################################
#	Complex mathematical functions 
#########################################################
	
def _complex_abs_v(*args):
	"""!
	Compute the abs of a complex number or vector (recursively).
	@param args list of numbers
	@return the abs of args
	"""
	sum = 0
	for arg in args: 
		if _is_list(arg): sum = sum + _complex_abs_v(*arg)
		else: sum = sum + \
			verify_complex(arg).real*verify_complex(arg).real + \
			verify_complex(arg).imag*verify_complex(arg).imag
	return verify_number(cmath.sqrt(sum))			

def _complex_real_v(*args):
	"""!
	Get the real part of a number or vector (recursively).
	@param args list of numbers
	@return the real of args
	"""
	result = list()
	for arg in args: 
		if _is_list(arg): result.append(_complex_real_v(*arg))
		else: result.append(verify_number(verify_complex(arg).real))
	return result

def _complex_imag_v(*args):
	"""!
	Get the imaginary part of a number or vector (recursively).
	@param args list of numbers
	@return the imaginary of args
	"""
	result = list()
	for arg in args:  
		if _is_list(arg): result.append(_complex_imag_v(*arg))
		else: result.append(verify_number(verify_complex(arg).imag))
	return result

def _complex_conj_v(*args):
	"""!
	Get the conjugate part of a number or vector (recursively).
	@param args list of numbers
	@return the conjugate of args
	"""
	result = list()
	for arg in args:  
		if _is_list(arg): result.append(_complex_conj_v(*arg))
		else: result.append(verify_number(verify_complex(arg).conjugate()))
	return result

def _complex_arg(arg):
	"""!
	Get the angle of the complex number.
	@param arg a complex number
	@return the angle in radians	
	"""
	return verify_number(math.atan2(verify_complex(arg).imag, verify_complex(arg).real))

def _handle_filter(filter, *filter_args):
	"""!
	Pass the filter arguments to the filter function. 
	Use a look up method to format the arguments to floats and ints before calling the filter function.
	@throw Exception invalid filter arguments.
	@return a list of taps.	
	"""
	filter_args = list(filter_args)	#make filter args a list
	filter_props = {	# map a filter to filter argument properties:
		# (allowed lengths of args), (args that are floats), (args that are ints)
		gr.firdes.band_pass : ((5, 6, 7), (0, 1, 2, 3, 4, 6), ()),
		gr.firdes.band_reject : ((5, 6, 7), (0, 1, 2, 3, 4, 6), ()),
		gr.firdes.complex_band_pass : ((5, 6, 7), (0, 1, 2, 3, 4, 6), ()),
		
		gr.firdes.high_pass : ((4, 5, 6), (0, 1, 2, 3, 5), ()),
		gr.firdes.low_pass : ((4, 5, 6), (0, 1, 2, 3, 5), ()),
		
		gr.firdes.gaussian : ((4,), (0, 1, 2), (3,)),
		gr.firdes.hilbert : ((1, 2, 3), (2,), (0,)),
		gr.firdes.root_raised_cosine : ((5,), (0, 1, 2, 3), (4,)),
		gr.firdes.window : ((3,), (2,), (1,)),
	}
	arg_lens, floats, ints = filter_props[filter]
	# handle the filter arguments #
	if len(filter_args) not in arg_lens: raise SyntaxError, 'Invalid number of arguments for "%s".'%filter	
	for i,arg in enumerate(filter_args):
		if i in floats: filter_args[i] = verify_float(arg)
		if i in ints: filter_args[i] = verify_int(arg)	
	taps = list(filter(*filter_args))	#the taps must be a list
	return taps 

_FUNCTIONS = {
	'pow' : lambda b, p: verify_number(b**p),	#complex power
	'sqrt' : cmath.sqrt,

	'abs' : _complex_abs_v,
	'mag' : _complex_abs_v,
	'real' : _complex_real_v,
	'imag' : _complex_imag_v,
	'conj' : _complex_conj_v,
	'arg' : _complex_arg,
	
	'sin' : lambda z: verify_number(cmath.sin(z)),	
	'cos' : lambda z: verify_number(cmath.cos(z)),
	'tan' : lambda z: verify_number(cmath.tan(z)),
	
	'asin' : lambda z: verify_number(cmath.asin(z)),
	'acos' : lambda z: verify_number(cmath.acos(z)),
	'atan' : lambda z: verify_number(cmath.atan(z)),
	
	'sinh' : lambda z: verify_number(cmath.sinh(z)),	#hyperbolics
	'cosh' : lambda z: verify_number(cmath.cosh(z)),
	'tanh' : lambda z: verify_number(cmath.tanh(z)),
	
	'asinh' : lambda z: verify_number(cmath.asinh(z)),
	'acosh' : lambda z: verify_number(cmath.acosh(z)),
	'atanh' : lambda z: verify_number(cmath.atanh(z)),
	
	'log10' : lambda z: verify_number(cmath.log10(z)),
	'log' : lambda *z: verify_number(cmath.log(*z)),	#optional 2nd argument as base
	'ln' : lambda z: cmath.log(z), #only allow natural log
	'exp' : lambda z: verify_number(cmath.exp(z)),	
	
	'band_pass' : lambda *args: _handle_filter(gr.firdes.band_pass, *args),	#taps generators
	'band_reject' : lambda *args: _handle_filter(gr.firdes.band_reject, *args),
	'complex_band_pass' : lambda *args: _handle_filter(gr.firdes.complex_band_pass, *args),
	'gaussian' : lambda *args: _handle_filter(gr.firdes.gaussian, *args),
	'high_pass' : lambda *args: _handle_filter(gr.firdes.high_pass, *args),
	'hilbert' : lambda *args: _handle_filter(gr.firdes.hilbert, *args),
	'low_pass' : lambda *args: _handle_filter(gr.firdes.low_pass, *args),
	'root_raised_cosine' : lambda *args: _handle_filter(gr.firdes.root_raised_cosine, *args),
	'window' : lambda *args: _handle_filter(gr.firdes.window, *args),
	
	'floor' : lambda f: math.floor(verify_float(f)),
	'ceil' : lambda f: math.ceil(verify_float(f)),
	'radians' : lambda f: math.radians(verify_float(f)),
	'degrees' : lambda f: math.degrees(verify_float(f)),
	'mod' : lambda f1, f2: math.fmod(verify_float(f1), verify_float(f2)),
	'atan2' : lambda y, x: math.atan2(verify_float(y), verify_float(x)),
	'sgn' : lambda f: abs(verify_float(f))/verify_float(f),
	'min' : lambda *f: verify_number(min(*map(verify_float, f))),
	'max' : lambda *f: verify_number(max(*map(verify_float, f))),
}

#########################################################
#	Boolean tests for special characters and symbols
#########################################################

def _is_list(symbol): return type(symbol) is list

def _is_function(symbol): return symbol in _FUNCTIONS.keys()

def _is_constant(symbol): return symbol in _CONSTANTS.keys()

def _is_number(symbol):	
	try: verify_number(symbol)
	except: return False
	return True
	
def _is_terminating_char(char_str): 
	return \
	_is_operator(char_str) or \
	_is_bracket(char_str) or \
	_is_whitespace(char_str) or \
	_is_comma(char_str)

def _is_comma(char_str): return char_str == ','

def _is_whitespace(char_str): return char_str in _WHITESPACE
	
def _is_operator(char_str): return char_str in _ORDER_OF_OPS
	
def _is_bracket(char_str): return _is_open_bracket(char_str) or _is_close_bracket(char_str)
	
def _is_open_bracket(char_str): return char_str in ('(', '[', '{')
	
def _is_close_bracket(char_str): return char_str in (')', ']', '}')
	
def _get_oposing_bracket(bracket): 
	return {
		'(' : ')',
		')' : '(',
		'[' : ']',
		']' : '[',
		'{' : '}',
		'}' : '{',
		'' : ''
	}[bracket]
	
#########################################################
#	Symbolic operations
#########################################################		
	
def _handle_operation(op, arg1, arg2):
	"""!
	Handle basic operations specified in the order of operations (recursively).
	All calculations are complex, some vector operations are supported.
	@param op the operator
	@param arg1 first argument to the operator	
	@param arg2 second argument to the operator
	@throw NameError unknown operations	
	@throw ZeroDivisionError divide by zero
	@throw ArithmeticError impossible operations
	@return the nested data representing the result (numbers and lists)
	"""
	if op not in _ORDER_OF_OPS: raise NameError, 'Unknown operator: "%s".'%(op)
	arg1_is_list = _is_list(arg1)
	arg2_is_list = _is_list(arg2)
	# addition and subtraction of vectors #
	if arg1_is_list and arg2_is_list and \
		len(arg1) == len(arg2) and op in ('+', '-'):
		result = list()
		for i,arg1_elem in enumerate(arg1):
			arg2_elem = arg2[i]
			result.append(_handle_operation(op, arg1_elem, arg2_elem))
		return result
	# scalar times vector #
	if not arg1_is_list and arg2_is_list and op == '*':
		return map(lambda elem: _handle_operation(op, arg1, elem), arg2)
	# vector times/divided scalar #
	if arg1_is_list and not arg2_is_list and op in ('*', '/'):
		return map(lambda elem: _handle_operation(op, elem, arg2), arg1)
	# non vectored operations #
	if not arg1_is_list and not arg2_is_list and \
		_is_number(arg1) and _is_number(arg2):	#non vector operations
		if op == '^-':
			return verify_number(arg1**(-1*arg2))
		elif op == '^':
			return verify_number(arg1**arg2)
		elif op == '*':
			return verify_number(arg1 * arg2)
		elif op == '/':
			try: return verify_number(verify_complex(arg1) / verify_complex(arg2))	#make sure to use non-integer division
			except ZeroDivisionError: raise ZeroDivisionError, 'Cannot divide "%s" by 0.'%(arg1,)
		elif op == '+':
			return verify_number(arg1 + arg2)
		elif op == '-':
			return verify_number(arg1 - arg2)
	raise ArithmeticError, 'Operation of "%s" cannot be performed on "%s", "%s".'%(op, arg1, arg2)
	
#########################################################
#	Evaluate the expression string 
#########################################################	

def _separate_expression(expr_str):
	"""!
	Separate the string into a list of elements.
	Use only the lower-cased version of the expression string.
	Terminating characters will dictate the separations.
	Terminating characters will be included in the list.
	Whitespace characters will be removed from the list.
	@param expr_str the expression string	
	@return a list of strings representing each element
	"""
	elements = list()
	element = ''
	for char_str in expr_str.lower():
		if _is_terminating_char(char_str):			
			elements.append(element)
			elements.append(char_str)
			element = ''
		else: element = element + char_str		
	elements.append(element)	#append the remaining characters
	#	remove whitespace	#
	return filter(lambda elem: not _is_whitespace(elem), elements)

def _rectify_operators(elements):
	"""!
	Rectify operators.	
	Fix the functionality of the minus sign (-)	and the plus sign (+) in the floating point format,
	which would have been separated by separate_expression.
	If the minus sign is used as a negative replace it with -1* or ^-. 
	@param elements a list of strings 
	@return the list of elements with minus and plus used only as operators
	"""
	rectified_elements = list()
	for i,element in enumerate(elements):
		if i == 0 and element == '-':	#first element was a minus
			rectified_elements.extend(['-1', '*'])
		elif i != 0 and element == '-' and i+1 != len(elements):	#minus found!
			if _is_operator(elements[i-1]) or \
			_is_open_bracket(elements[i-1]) or \
			_is_comma(elements[i-1]):	#negative implied
				if elements[i-1] != '^': rectified_elements.extend(['-1', '*'])
				else: #^ proceeds - in order of ops, remove the ^ and append ^- as a special operator
					rectified_elements.pop()	#remove last element
					rectified_elements.append('^-')
			elif elements[i-1][-1] == 'e' and _is_number(elements[i-1][0:-1]):	#found a -floating point format 
				f = rectified_elements.pop()	#remove last element
				elements[i+1] = "%s-%s"%(f, elements[i+1])
			else: rectified_elements.append(element)	#regular minus
		elif i != 0 and element == '+' and i+1 != len(elements) and \
			elements[i-1][-1] == 'e' and _is_number(elements[i-1][0:-1]):	#found a +floating point format 
			f = rectified_elements.pop()	#remove last element
			elements[i+1] = "%s+%s"%(f, elements[i+1])	
		else: rectified_elements.append(element)	#non minus
	return rectified_elements

def _nest_elements(elements, open_bracket=''):
	"""!
	Put the elements into a nested data format (recursively).
	Read through the elements, recursing at each open bracket.
	Put elements inside bracket pairs into a list.	
	@param a list of strings (rectified)
	@param open_bracket the last open bracket
	@throw SyntaxError bracket pairs do not match
	@return the nested elements.	
	"""
	nested_elements = list()
	while elements:	#continue until elements is empty
		element = elements.pop(0)
		if _is_open_bracket(element):
			nested_elements.append(_nest_elements(elements, element))
		elif _is_close_bracket(element):	
			if element != _get_oposing_bracket(open_bracket):	#matching open/close brackets?
				raise SyntaxError, 'Expected "%s", but found "%s".'%(_get_oposing_bracket(open_bracket), element)
			return nested_elements
		else: nested_elements.append(element)
	if open_bracket:	#any unclosed bracket pairs? 
		raise SyntaxError, 'Expected "%s", but found nothing.'%(_get_oposing_bracket(open_bracket),)
	return nested_elements
	
def _simplify_nested_elements(nested_elements):
	"""!
	Read through the nested elements and simplify expressions (recursively).
	Work according to the order of operations.
	Recurse at each list, substitute constans, parse numbers, handle functions.
	Make comma separated values into lists to represent vectors.
	@param nested_elements nested elements (strings and lists)
	@throw SyntaxError bad syntax
	@throw NameError unknown symbol
	@return the final value of the expression.
	"""
	for i,element in enumerate(nested_elements):	
		# simplify the nested lists #
		if _is_list(element): nested_elements[i] = _simplify_nested_elements(element)
		# substitute in values for constants #
		elif _is_constant(element): nested_elements[i] = _CONSTANTS[element]
		# substitute in values for numbers #
		elif _is_number(element): nested_elements[i] = verify_number(element)			
	def comma_separate(ans):
		if not _is_list(ans): ans = [ans]	#must be a list
		ans_with_commas = list()
		for an in ans: ans_with_commas.extend([an, ','])
		return ans_with_commas[0:-1]	#remove last comma	
	# handle functions #
	for function in _FUNCTIONS.keys():
		while function in nested_elements:
			i = nested_elements.index(function)				
			if i+1 == len(nested_elements): raise SyntaxError, 'Function "%s" has no arguments.'%(function)
			args = nested_elements[i+1]
			if not _is_list(args): args = [args]
			try: ans = _FUNCTIONS[function](*args)
			except Exception, e: 
				raise SyntaxError, 'Function "%s" with arguments "%s" failed.\n\t%s'%(function, nested_elements[i+1], e)
			nested_elements = nested_elements[0:i] + comma_separate(ans) + nested_elements[i+2:]		
	# simplify operations #
	for operator in _ORDER_OF_OPS:
		while operator in nested_elements:
			i = nested_elements.index(operator)					
			arg1 = arg2 = None
			if i > 0: arg1 = nested_elements[i-1]
			if i+1 < len(nested_elements): arg2 = nested_elements[i+1]
			#raise error if arg1 or arg2 is None
			if arg1 == None or arg2 == None: raise SyntaxError, 'Operator "%s" is missing argument.'%(operator)
			ans = _handle_operation(operator, arg1, arg2)
			nested_elements = nested_elements[0:i-1] + comma_separate(ans) + nested_elements[i+2:]		
	# convert comma separated elements into a list #
	vector = list()	
	last_element = None
	for element in nested_elements:
		if not _is_comma(element): 
			if not _is_list(element) and not _is_number(element) and element not in _CONSTANTS.values(): 
				raise NameError, 'Unknown symbol "%s".'%(element)
			if last_element != None and not _is_comma(last_element): #elements not separated by commas
				raise SyntaxError, 'Expected comma, but found "%s".'%(element)
			vector.append(element)	#good syntax, record the element
		elif _is_comma(element) and _is_comma(last_element):	#2 commas in a row
			raise SyntaxError, 'Commas must be separated by non-commas.'
		elif _is_comma(element) and last_element == None:	#a comma was the first element
			raise SyntaxError, 'A non-comma must precede a comma.'
		last_element = element
	if len(vector) == 1 and not _is_comma(last_element): return vector[0]	#return single number
	return vector	#otherwise return vector

def eval_expr(expr_str):
	"""!
	Evaluate a mathematical expression string with numbers, and operators.
	@throw Exception failue	
	@return nested data (numbers and lists)
	"""	
	separated_elements = _separate_expression(expr_str)
	#print "separated", separated_elements
	rectified_elements = _rectify_operators(separated_elements)
	#print "rectified", rectified_elements
	nested_elements = _nest_elements(rectified_elements)
	#print "nested", nested_elements
	simplified = _simplify_nested_elements(nested_elements)
	return simplified
	
if __name__ == '__main__':
	"""Evaluate expressions passed by arg1."""
	import sys
	if len(sys.argv) > 1: print eval_expr(sys.argv[1])
	else: print "The mathematical expression parser will evaluate a string (in quotes) passed via the command line."
	