"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package ParseXML
#Parse xml files to nested data and vice-versa.
#@author Josh Blum

import xml.dom.minidom
import xml.dom.ext

def find_data(nested_data, my_tag):
	"""!
	Find a tag one level deep.	
	Helper function for dealing with nested data.	
	@param nested_data the data to search
	@param my_tag the tag to search for
	@return the data associated with the tag, None if not found
	"""
	for tag,data in nested_data:
		if tag == my_tag: return data
	return None

def from_xml(doc):
	"""!
	Create nested data from an xml doc using the from xml helper.
	@param doc the xml doc
	@return the nested data
	"""
	return __from_xml(doc.firstChild)
	
def __from_xml(doc):
	"""!
	Recursivly parse the xml doc into nested data format.	
	@param doc the xml doc
	@return the nested data
	"""
	#print doc.nodeType,doc.localName,doc.nodeValue,doc.childNodes
	tag_name = doc.localName
	if len(doc.childNodes) == 0:
		return (tag_name,'')
	elif len(doc.childNodes) == 1 and doc.firstChild.nodeValue != None:
		return (tag_name, str(doc.firstChild.nodeValue))#parse the unicode to str (very important!)
	nested_data = list()	
	for elem in doc.childNodes:		
		if elem.localName != None:
			nested_data.append(__from_xml(elem))	
	return tag_name,nested_data	
	
def to_xml(nested_data):
	"""!
	Create an xml doc and use the to xml helper method to load it.
	@param nested_data the nested data
	@return the xml doc
	"""
	doc = xml.dom.minidom.Document()
	doc.appendChild(__to_xml(nested_data, doc))
	return doc
	
def __to_xml(nested_data, doc):
	"""!
	Recursivly parse the nested data into xml doc format.
	@param nested_data the nested data
	@param the xml doc
	@return the xml doc
	"""	
	tag_name,value = nested_data
	child = doc.createElementNS(None, tag_name)
	if type(value) == type(list()):		
		for elem in value:
			child.appendChild(__to_xml(elem, doc))
		doc.appendChild(child)		
		return child
	elif value != '':
		child.appendChild(doc.createTextNode(value))
	return child

def to_file(doc, file_path):
	"""!
	Save the xml doc to the file path.
	@param doc the xml doc
	@param file_path the destination xml-file
	""" 
	xml.dom.ext.PrettyPrint(doc, open(file_path,'w'))
	
def from_file(file_path):
	"""!
	Load an xml doc from the file path.
	@param file_path the source xml-file
	@return the xml doc
	"""
	return xml.dom.minidom.parse(open(file_path, 'r'))

if __name__ == '__main__':	
	"""Use the main method to test parse xml's functions."""
	pass
	