"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Preferences
#Holds global preferences stored as GraphicalParams.
#@author Josh Blum

from DataTypes import Bool,Enum,Int,String
from Elements import GraphicalParam, Param
import time,socket #for tagging saved files
from Constants import *
import ParseXML
import Messages

FILE_DELIMITER = '\n'

PARAM_CHECKING_PREF = GraphicalParam('Parameter Checking', Bool(true='Verify Parameters', false='Ignore Invalid Params', default=True))
CONNECTION_CHECKING_PREF = GraphicalParam('Connection Checking', Bool(true='Match Data Types', false='Ignore Data Types', default=True))
SOCKET_CHECKING_PREF = GraphicalParam('Socket Checking', Bool(true='Require Connections', false='Allow Open Sockets', default=True))
FILES_OPEN_PREF = Param('Files Open', String())
RESTORE_FILES_PREF = GraphicalParam('Restore Files', Bool(true='Restore Previous Files', false='Restore Nothing', default=True))
SNAP_TO_GRID_PREF = GraphicalParam('Snap to Grid', Bool(true='On', false='Off', default=False))
GRID_SIZE_PREF = GraphicalParam('Grid Size (pixels)', Enum([
	('10 pixels', 10), 
	('20 pixels', 20), 
	('30 pixels', 30),
	('40 pixels', 40),
	('50 pixels', 50),
	('60 pixels', 60),
	('70 pixels', 70),
	('80 pixels', 80),
	('90 pixels', 90),
	('100 pixels', 100),], 2))
SHOW_GRID_PREF = GraphicalParam('Grid Points', Bool(true='Show Grid', false='Hide Grid', default=False))
REPORTS_WIN_SHOW_PREF = GraphicalParam('Reports Window', Bool(true='Show Reports Window', false='Hide Reports Window', default=True))
MAIN_WINDOW_WIDTH_PREF = Param('Main Window Width', Int(DEFAULT_MAIN_WINDOW_WIDTH))
MAIN_WINDOW_HEIGHT_PREF = Param('Main Window Height', Int(DEFAULT_MAIN_WINDOW_HEIGHT))

###########################################################################
#	List of Preferences
###########################################################################

PREFS_LIST = [	#(title, notes, list of preferences)
	("Verification", '''\
Verification ensures that:
	signal block parameters pass validation,
	data types on either side of a connection match,
	and input/output sockets connect.
	
You may disable any one of these validation checks. \
However, the resulting flow graph may fail execution. \
''', [	
			(PARAM_CHECKING_PREF, 'param_check'),
			(CONNECTION_CHECKING_PREF, 'connection_check'),
			(SOCKET_CHECKING_PREF, 'socket_check'),
		]
	),
	("Grid Options", '''\
Snap to Grid forces the upper right corner of the signal block to align with a grid point. \
''', [
			(SNAP_TO_GRID_PREF, 'snap_to_grid'),
			(GRID_SIZE_PREF, 'grid_size'),
			(SHOW_GRID_PREF, 'show_grid'),
		]
	),
	("Misc Options", '', [
			(REPORTS_WIN_SHOW_PREF, 'show_reports_window'),
			(RESTORE_FILES_PREF, 'restore_files'),
		]
	),	
	('', '', [	#put hidden prefs here
			(FILES_OPEN_PREF, 'files_open'),
			(MAIN_WINDOW_WIDTH_PREF, 'window_width'),
			(MAIN_WINDOW_HEIGHT_PREF, 'window_height'),
		]
	),	
]
	
PREFS_DICT = dict((key, param) for param, key in reduce(lambda l1, l2: l1 + l2, zip(*PREFS_LIST)[2]))

###########################################################################
#	Preference Access Methods
###########################################################################

def _get_value(param):
	"""!
	Get the value of the given parameter.
	If the param is valid, return the parsed value of the param, otherwise return a blank string.	
	@param param the parameter
	@return the value of the parameter
	"""
	if param.get_data_type().is_valid(): return param.get_data_type().parse()
	else: return ''
	
def _to_nested():
	"""!
	Convert the param's data to nested format.
	@return the nested format
	"""
	prefs = [('pref', [('key', key), ('value', param.get_data_type().get_data())]) for key,param in PREFS_DICT.iteritems()]
	nested_data = ('preferences', [
			('timestamp', str(time.time())),
			('hostname', socket.gethostname()),
			('version', VERSION),											
			('prefs', prefs),
		]
	)	
	return nested_data
	
def _from_nested(nested_data):
	"""!
	Parse the nested data to retrieve each preference.
	@param nested_data the nested data
	"""
	find_data = ParseXML.find_data
	preferences = find_data([nested_data], 'preferences')
	prefs = find_data(preferences, 'prefs')
	for pref in prefs:
		pref = find_data([pref], 'pref')
		key = find_data(pref, 'key')
		value = find_data(pref, 'value')
		try: PREFS_DICT[key].get_data_type().set_data(value)
		except: pass
		
def load(window=None):
	"""!
	Load the preferences from the preferences file.
	@param window optional flow graph window
	"""
	try: 
		_from_nested(ParseXML.from_xml(ParseXML.from_file(PREFERENCES_FILE_PATH)))
		if window: window.resize(_get_value(MAIN_WINDOW_WIDTH_PREF), _get_value(MAIN_WINDOW_HEIGHT_PREF))
	except: Messages.send_fail_load_preferences()		

def save(window=None):
	"""!
	Save the preferences to the preferences file.
	@param window optional flow graph window
	"""
	if window:
		width,height = window.get_size()
		MAIN_WINDOW_WIDTH_PREF.get_data_type().set_data(width)
		MAIN_WINDOW_HEIGHT_PREF.get_data_type().set_data(height)
	try: ParseXML.to_file(ParseXML.to_xml(_to_nested()), PREFERENCES_FILE_PATH)		
	except IOError: Messages.send_fail_save_preferences()
		
###########################################################################
#	Special methods for specific program functionalities
###########################################################################
	
def restore_files():
	return _get_value(RESTORE_FILES_PREF)	
	
def files_open():
	return _get_value(FILES_OPEN_PREF).split(FILE_DELIMITER)
	
def save_files_open(files):
	FILES_OPEN_PREF.get_data_type().set_data(FILE_DELIMITER.join(files))
	
def check_connections():
	return _get_value(CONNECTION_CHECKING_PREF)
	
def check_sockets():
	return _get_value(SOCKET_CHECKING_PREF)

def check_params():
	return _get_value(PARAM_CHECKING_PREF)
	
def show_reports_window():
	return _get_value(REPORTS_WIN_SHOW_PREF)
	
def get_grid_size():
	return _get_value(GRID_SIZE_PREF)	
	
def snap_to_grid():
	return _get_value(SNAP_TO_GRID_PREF)
	
def show_grid():
	return _get_value(SHOW_GRID_PREF)
		
	