"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Conversions
#Convert between data types.
#@author Josh Blum

from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import all_choices,all_vector_choices,default_samp_rate,MAX_NUM_SOCKETS

def ComplexComponents(sb):		
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Complex(), vlen=vlen)
	sb.add_output_socket('out', Float(), vlen=vlen)
	sb.add_param('Output Type', Enum([
		('Real', gr.complex_to_real),
		('Imaginary', gr.complex_to_imag),
		('Magnitude', gr.complex_to_mag),
		('Magnitude Squared', gr.complex_to_mag_squared),
		('Phase Angle', gr.complex_to_arg),
	]), type=True)	
	sb.add_param('Vector Length', vlen)
	return sb, lambda fg, type, vlen: type.parse()(vlen.parse())
	
def ComplexConjugate(sb):
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Complex())
	return sb, lambda fg: gr.conjugate_cc()
	
def FloatToComplex(sb):
	sb.add_input_socket('in', Float())
	sb.add_input_socket('jin', Float(), optional=True)
	sb.add_output_socket('out', Complex())
	sb.set_docs('''The imaginary input socket (jin) can be unconnected.''')
	return sb, lambda fg: gr.float_to_complex()
	
def ComplexToFloat(sb):	
	vlen = Int(1, min=1)
	sb.add_output_socket('out', Float(), vlen=vlen)
	sb.add_output_socket('jout', Float(), vlen=vlen, optional=True)
	sb.add_input_socket('in', Complex())		
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''The imaginary output socket (jout) can be unconnected.''')
	return sb, lambda fg, vlen: gr.complex_to_float(vlen.parse())
	
def ShortToFloat(sb):
	sb.add_input_socket('in', Short())
	sb.add_output_socket('out', Float())
	return sb, lambda fg: gr.short_to_float()
	
def FloatToShort(sb):
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Short())
	return sb, lambda fg: gr.float_to_short()
	
def CharToFloat(sb):
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Float())
	return sb, lambda fg: gr.char_to_float()
	
def FloatToChar(sb):
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Byte())
	return sb, lambda fg: gr.float_to_char()
	
def UCharToFloat(sb):
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Float())
	return sb, lambda fg: gr.uchar_to_float()
	
def FloatToUChar(sb):
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Byte())
	return sb, lambda fg: gr.float_to_uchar()

def IShortToComplex(sb):
	fcn = gr.interleaved_short_to_complex	
	sb.add_input_socket('in', Short())
	sb.add_output_socket('out', Complex())
	return sb, lambda fg: fcn()
	
def ComplexToIShort(sb):
	fcn = gr.complex_to_interleaved_short
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Short())
	return sb, lambda fg: fcn()
	
def UnpackedToPacked(sb):
	type = Enum([
		('Byte', (gr.unpacked_to_packed_bb, Byte())),
		('Short', (gr.unpacked_to_packed_ss, Short())),
		('Int', (gr.unpacked_to_packed_ii, Int())),
	])	
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Bits per Chunk', Int(1, min=0))
	sb.add_param('Endianness', Enum([('MSB', gr.GR_MSB_FIRST), ('LSB', gr.GR_LSB_FIRST)]))
	sb.set_docs('''Unpacked to Packed''')
	return sb, lambda fg, type, bits_per_chunk, endianness: type.parse()[0](bits_per_chunk.parse(), endianness.parse())
	
def PackedToUnpacked(sb):
	type = Enum([
		('Byte', (gr.packed_to_unpacked_bb, Byte())),
		('Short', (gr.packed_to_unpacked_ss, Short())),
		('Int', (gr.packed_to_unpacked_ii, Int())),
	])	
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Bits per Chunk', Int(1, min=0))
	sb.add_param('Endianness', Enum([('MSB', gr.GR_MSB_FIRST), ('LSB', gr.GR_LSB_FIRST)]))	
	sb.set_docs('''Packed to Unpacked''')
	return sb, lambda fg, type, bits_per_chunk, endianness: type.parse()[0](bits_per_chunk.parse(), endianness.parse())

def UnpackKBits(sb):
	fcn = gr.unpack_k_bits_bb  
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Byte())
	sb.add_param('K', Int(10, min=0))
	return sb, lambda fg, k: fcn(k.parse())
	
def BinarySlicer(sb):
	fcn = gr.binary_slicer_fb
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Byte())
	return sb, lambda fg: fcn()
	
def ChunksToSymbols(sb):
					#name, fcn, input, output, symbol table, dimensions
	type = Enum([
		('Byte->Complex', (gr.chunks_to_symbols_bc, Byte(), Complex(), ComplexVector())),
		('Byte->Float', (gr.chunks_to_symbols_bf, Byte(), Float(), FloatVector())),
		('Int->Complex', (gr.chunks_to_symbols_ic, Int(), Complex(), ComplexVector())),
		('Int->Float', (gr.chunks_to_symbols_if, Int(), Float(), FloatVector())),
		('Short->Complex', (gr.chunks_to_symbols_sc, Short(), Complex(), ComplexVector())),
		('Short->Float', (gr.chunks_to_symbols_sf, Short(), Float(), FloatVector())),
	])
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=2))
	sb.add_param('Type', type, type=True)	
	sb.add_param('Symbol Table', Variable(type, index=3))
	sb.add_param('Dimensions', Int(1, min=1))
	return sb, lambda fg, type, sym_table, dim: type.parse()[0](sym_table.parse(), dim.parse())
	
def Interleave(sb):
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Inputs', Int(3, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''1 <= Num Inputs <= %d'''%MAX_NUM_SOCKETS)	
	return sb, lambda fg, type, num_inputs, vlen: gr.interleave(type.parse().get_num_bytes()*vlen.parse())
	
def Deinterleave(sb):
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Outputs', Int(3, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, output_sockets_controller=True)	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''1 <= Num Outputs <= %d'''%MAX_NUM_SOCKETS)	
	return sb, lambda fg, type, num_outputs, vlen: gr.deinterleave(type.parse().get_num_bytes()*vlen.parse())
	
def StreamsToStream(sb):
	fcn = gr.streams_to_stream
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Streams', Int(3, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''\
Interleave N streams into a single stream.	
1 <= Num Streams <= %d'''%MAX_NUM_SOCKETS)	
	return sb, lambda fg, type, num_streams, vlen: fcn(type.parse().get_num_bytes()*vlen.parse(), num_streams.parse())
	
def StreamToStreams(sb):
	fcn = gr.stream_to_streams
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Streams', Int(3, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, output_sockets_controller=True)	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''\
Deinterleave a stream into N streams.
1 <= Num Streams <= %d'''%MAX_NUM_SOCKETS)	
	return sb, lambda fg, type, num_streams, vlen: fcn(type.parse().get_num_bytes()*vlen.parse(), num_streams.parse())
	
def StreamsToVector(sb):
	fcn = gr.streams_to_vector
	type = Enum(all_vector_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type, index=1), vlen=vlen)
	sb.add_output_socket('vout', Variable(type, index=0))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Streams', Int(3, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''\
Convert N streams into a vector-stream of length N.
1 <= Num Streams <= %d'''%MAX_NUM_SOCKETS)	
	return sb, lambda fg, type, num_streams, vlen: fcn(type.parse()[1].get_num_bytes()*vlen.parse(), num_streams.parse())
	
def VectorToStreams(sb):
	fcn = gr.vector_to_streams
	type = Enum(all_vector_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('vin', Variable(type, index=0))
	sb.add_output_socket('out', Variable(type, index=1), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Streams', Int(3, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, output_sockets_controller=True)	
	sb.add_param('Vector Length', vlen)	
	sb.set_docs('''\
Convert a vector-stream of length N into N streams.
1 <= Num Streams <= %d'''%MAX_NUM_SOCKETS)	
	return sb, lambda fg, type, num_streams, vlen: fcn(type.parse()[1].get_num_bytes()*vlen.parse(), num_streams.parse())
	
def StreamToVector(sb):
	fcn = gr.stream_to_vector
	type = Enum(all_vector_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type, index=1), vlen=vlen)
	sb.add_output_socket('vout', Variable(type, index=0))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Items / Block', Int(1, min=1))	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Separate a single stream into a vector-stream of size N.''')	
	return sb, lambda fg, type, items_per_block, vlen: fcn(type.parse()[1].get_num_bytes()*vlen.parse(), items_per_block.parse())
	
def VectorToStream(sb):
	fcn = gr.vector_to_stream
	type = Enum(all_vector_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('vin', Variable(type, index=0))
	sb.add_output_socket('out', Variable(type, index=1), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Items / Block', Int(1, min=1))	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Merge a vector-stream of size N into a single stream.''')	
	return sb, lambda fg, type, items_per_block, vlen: fcn(type.parse()[1].get_num_bytes()*vlen.parse(), items_per_block.parse())
	
def Map(sb):
	fcn = gr.map_bb
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Byte())
	sb.add_param('Vector', IntVector(''))
	sb.set_docs('''output[i] = map[input[i]]''')	
	return sb, lambda fg, vector: fcn(vector.parse())
	
def VCO(sb):
	fcn = gr.vco_f
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Float())
	sb.add_param('Sampling Rate', Float(default_samp_rate))
	sb.add_param('Sensitivity', Float(1))
	sb.add_param('Amplitude', Float(1))
	sb.set_docs('''\
VCO - Voltage controlled oscillator.

Parameters:
    	sampling_rate 	sampling rate (Hz)
    	sensitivity 	units are radians/sec/volt
    	amplitude 	output amplitude	
''')	
	return sb, lambda fg, samp_rate, sense, amp: fcn(samp_rate.parse(), sense.parse(), amp.parse())
	