"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Filters
#Taps generators are from gr.firdes. Most FIR filters are implemented and a few misc ones.
#@author Josh Blum

from DataTypes import *
from gnuradio import gr,gru,blks2
from SignalBlockConstants import default_samp_rate,all_choices

###########################################################################
#	Generic filters with taps parameters
###########################################################################
def FIRFilter(sb):
	filters = Enum([
		('Complex->Complex (Complex Taps)', (gr.fir_filter_ccc, gr.interp_fir_filter_ccc, Complex(), Complex(), ComplexVector())),	
		('Complex->Complex (Float Taps)', (gr.fir_filter_ccf, gr.interp_fir_filter_ccf, Complex(), Complex(), FloatVector())),	
		('Float->Complex (Complex Taps)', (gr.fir_filter_fcc, gr.interp_fir_filter_fcc, Float(), Complex(), ComplexVector())),	
		('Float->Float (Float Taps)', (gr.fir_filter_fff, gr.interp_fir_filter_fff, Float(), Float(), FloatVector())),	
		('Float->Short (Float Taps)', (gr.fir_filter_fsf, gr.interp_fir_filter_fsf, Float(), Short(), FloatVector())),	
		('Short->Complex (Complex Taps)', (gr.fir_filter_scc, gr.interp_fir_filter_scc, Short(), Complex(), ComplexVector())),		
	])
	sb.add_input_socket('in', Variable(filters, index=2))
	sb.add_output_socket('out', Variable(filters, index=3))
	sb.add_param('Filter Type', filters, False, type=True)
	sb.add_param('Type', Enum([('Decimating', 0), ('Interpolating', 1)]))	#this parses into an index for the filters Enum
	sb.add_param('Decimation', Int(1, min=1))	
	sb.add_param('Taps', Variable(filters, index=4, min=1), variable=True)	
	sb.set_docs('''\
The decimation parameter becomes interpolation if the filter type is set to interpolation. ''')
	def make(fg, filter_type, type, decim, taps):
		block = filter_type.parse()[type.parse()](decim.parse(), taps.parse())
		fg.add_callback(block.set_taps, taps)
		return block
	return sb,make
	
def FreqXlatingFIRFilter(sb):	
	filters = Enum([
		('Complex', (gr.freq_xlating_fir_filter_ccc, gr.freq_xlating_fir_filter_ccf, Complex())),	
		('Float', (gr.freq_xlating_fir_filter_fcc, gr.freq_xlating_fir_filter_fcf, Float())),
		('Short', (gr.freq_xlating_fir_filter_scc, gr.freq_xlating_fir_filter_scf, Short())),			
	])
	taps = Enum([
		('Complex', (0, ComplexVector())),
		('Float', (1, FloatVector())),
	])
	sb.add_input_socket('in', Variable(filters, index=2))
	sb.add_output_socket('out', Complex())
	sb.add_param('Input Type', filters, False, type=True)
	sb.add_param('Taps Type', taps)	#this parses into an index for the filters Enum
	sb.add_param('Samp Rate', Float(default_samp_rate))
	sb.add_param('Center Freq', Float(), variable=True)
	sb.add_param('Decimation', Int(1, min=1))	
	sb.add_param('Taps', Variable(taps, index=1, min=1), variable=True)
	def make(fg, filter_type, taps_type, samp_rate, center_freq, decim, taps):
		block = filter_type.parse()[taps_type.parse()[0]](decim.parse(), taps.parse(), center_freq.parse(), samp_rate.parse())
		fg.add_callback(block.set_taps, taps)
		fg.add_callback(block.set_center_freq, center_freq)
		return block
	return sb,make
	
def FFTFilter(sb):
	filters = Enum([
		('Complex', (gr.fft_filter_ccc, Complex(), ComplexVector())),	
		('Float', (gr.fft_filter_fff, Float(), FloatVector())),			
	])	
	sb.add_input_socket('in', Variable(filters, index=1))
	sb.add_output_socket('out', Variable(filters, index=1))
	sb.add_param('Type', filters, False, type=True)
	sb.add_param('Decimation', Int(1, min=1))	
	sb.add_param('Taps', Variable(filters, index=2, min=1), variable=True)
	sb.set_docs('''The type selection controls the data type of the input, output, and taps.''')
	def make(fg, type, decim, taps):
		block = type.parse()[0](decim.parse(), taps.parse())
		fg.add_callback(block.set_taps, taps)
		return block
	return sb,make
	
def RationalResampler(sb):
	filters = Enum([
		('Complex->Complex (Complex Taps)', (gr.rational_resampler_base_ccc, Complex(), Complex(), ComplexVector())),
		('Complex->Complex (Float Taps)', (gr.rational_resampler_base_ccf, Complex(), Complex(), FloatVector())),
		('Float->Complex (Complex Taps)', (gr.rational_resampler_base_fcc, Float(), Complex(), ComplexVector())),
		('Float->Float (Float Taps)', (gr.rational_resampler_base_fff, Float(), Float(), FloatVector())),
		('Float->Short (Float Taps)', (gr.rational_resampler_base_fsf, Float(), Short(), FloatVector())),
		('Short->Complex (Complex Taps)', (gr.rational_resampler_base_scc, Short(), Complex(), ComplexVector())),
	])	
	sb.add_input_socket('in', Variable(filters, index=1))
	sb.add_output_socket('out', Variable(filters, index=2))
	sb.add_param('Filter Type', filters, False, type=True)
	sb.add_param('Interpolation', Int(1, min=1))	
	sb.add_param('Decimation', Int(1, min=1))	
	sb.add_param('Taps', Variable(filters, index=3, min=1), variable=True)
	def make(fg, filter_type, interp, decim, taps):
		block = filter_type.parse()[0](interp.parse(), decim.parse(), taps.parse())
		fg.add_callback(block.set_taps, taps)
		return block
	return sb,make

def IIRFiler(sb):
	fcn = gr.iir_filter_ffd
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Float())
	sb.add_param('FF Taps', FloatVector())
	sb.add_param('FB Taps', FloatVector())
	return sb, lambda fg, fftaps, fbtaps: fcn(fftaps.parse(), fbtaps.parse())

def FilterDelay(sb):
	fcn = gr.filter_delay_fc
	sb.add_input_socket('in1', Float())
	sb.add_input_socket('in2', Float(), optional=True)
	sb.add_output_socket('out', Complex())
	sb.add_param('Taps', FloatVector())
	sb.set_docs('''\
The block takes one or two float stream and outputs a complex stream. \
If only one float stream is input, \
the real output is a delayed version of this input and the imaginary output is the filtered output. \
If two floats are connected to the input, \
then the real output is the delayed version of the first input, \
the imaginary output is the filtered output. \
The delay in the real path accounts for the group delay introduced by the filter in the imaginary path.\
''')
	return sb, lambda fg, taps: fcn(taps.parse())
	
def ChannelModel(sb):
	fcn = blks2.channel_model
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Complex())
	sb.add_param('Noise Voltage', Float(0.0), variable=True)
	sb.add_param('Freq Offset', Float(0.0), variable=True)
	sb.add_param('Epsilon', Float(1.0))
	sb.add_param('Taps', ComplexVector('1.0,0.0'), variable=True)
	sb.set_docs('''\
Creates a channel model that includes:
	- AWGN noise power in terms of noise voltage
	- A frequency offest in the channel in ratio
	- A timing offset ratio to model clock difference (epsilon)
	- Multipath taps
''')
	def make(fg, noise_voltage, frequency_offset, epsilon, taps):
		block = fcn(noise_voltage.parse(), frequency_offset.parse(), epsilon.parse(), taps.parse())
		fg.add_callback(block.set_noise_voltage, noise_voltage)
		fg.add_callback(block.set_frequency_offset, frequency_offset)
		fg.add_callback(block.set_taps, taps)
		return block
	return sb, make

###########################################################################
#	Special filters using taps generators
###########################################################################
def make_filter(fg, filter_type, decimation, taps_maker, taps_args):
	filter = filter_type.parse()[0]
	decimation = decimation.parse()	
	taps = taps_maker(*map(lambda data:data.parse(), taps_args))		
	block = filter(decimation, taps)
	fg.add_callback(lambda *args: block.set_taps(taps_maker(*args)), *taps_args)
	return block	

window_choices = [
	('Hamming', gr.firdes.WIN_HAMMING),
	('Hann', gr.firdes.WIN_HANN),
	('Blackman', gr.firdes.WIN_BLACKMAN),
	('Rectangular', gr.firdes.WIN_RECTANGULAR),
	('Kaiser', gr.firdes.WIN_KAISER),
]
filter_choices = [
	('FFT: Float->Float', (gr.fft_filter_fff, Float(), Float())),
	('FIR: Complex->Complex', (gr.fir_filter_ccf, Complex(), Complex())),
	('FIR: Float->Float', (gr.fir_filter_fff, Float(), Float())),
	('FIR: Float->Short', (gr.fir_filter_fsf, Float(), Short())),
	('Interp FIR: Complex->Complex', (gr.interp_fir_filter_ccf, Complex(), Complex())),
	('Interp FIR: Float->Float', (gr.interp_fir_filter_fff, Float(), Float())),
	('Interp FIR: Float->Short', (gr.interp_fir_filter_fsf, Float(), Short())),
]
			
def LowPassFilter(sb):
	taps_maker = gr.firdes.low_pass
	pass_filter_helper(sb, taps_maker)
	def make(fg, filter_type, gain, samp_freq, cutoff_freq, transition_width, decimation, window, beta):
		taps_args = (gain, samp_freq, cutoff_freq, transition_width, window, beta)		
		return make_filter(fg, filter_type, decimation, taps_maker, taps_args)
	return sb, make

def HighPassFilter(sb):
	taps_maker = gr.firdes.high_pass
	pass_filter_helper(sb, taps_maker)
	def make(fg, filter_type, gain, samp_freq, cutoff_freq, transition_width, decimation, window, beta):
		taps_args = (gain, samp_freq, cutoff_freq, transition_width, window, beta)		
		return make_filter(fg, filter_type, decimation, taps_maker, taps_args)
	return sb, make

def BandPassFilter(sb):
	taps_maker = gr.firdes.band_pass
	pass_filter_helper(sb, taps_maker)
	def make(fg, filter_type, gain, samp_freq, low_cutoff_freq, high_cutoff_freq, transition_width, decimation, window, beta):
		taps_args = (gain, samp_freq, low_cutoff_freq, high_cutoff_freq, transition_width, window, beta)		
		return make_filter(fg, filter_type, decimation, taps_maker, taps_args)
	return sb, make

def BandRejectFilter(sb):
	taps_maker = gr.firdes.band_reject
	pass_filter_helper(sb, taps_maker)
	def make(fg, filter_type, gain, samp_freq, low_cutoff_freq, high_cutoff_freq, transition_width, decimation, window, beta):
		taps_args = (gain, samp_freq, low_cutoff_freq, high_cutoff_freq, transition_width, window, beta)		
		return make_filter(fg, filter_type, decimation, taps_maker, taps_args)
	return sb, make

def pass_filter_helper(sb, taps_maker):
	filters = Enum(filter_choices, 2)
	windows = Enum(window_choices, 1)
	sb.add_input_socket('in', Variable(filters, index=1))
	sb.add_output_socket('out', Variable(filters, index=2))
	sb.add_param('Filter Type', filters, type=True)
	sb.add_param('Gain', Float(1), variable=True)
	sb.add_param('Sampling Rate', Float(default_samp_rate), variable=True)
	if taps_maker in (gr.firdes.low_pass, gr.firdes.high_pass): 
		sb.add_param('Cutoff Freq', Float(1000), variable=True)
	elif taps_maker in (gr.firdes.band_pass, gr.firdes.band_reject):
		sb.add_param('Low Cutoff Freq', Float(500), variable=True)
		sb.add_param('High Cutoff Freq', Float(1500), variable=True)
	sb.add_param('Transition Width', Float(100), variable=True)		
	sb.add_param('Decimation', Int(1, min=1))
	sb.add_param('Window Type', windows)
	sb.add_param('Beta', Float(6.76), variable=True)
	sb.set_docs('''\
The decimation parameter becomes interpolation if the filter type is set to interpolation.''')

def Window(sb):
	taps_maker = gr.firdes.window
	filters = Enum(filter_choices, 2)
	windows = Enum(window_choices, 1)
	sb.add_input_socket('in', Variable(filters, index=1))
	sb.add_output_socket('out', Variable(filters, index=2))	
	sb.add_param('Filter Type', filters, type=True)
	sb.add_param('Decimation', Int(1, min=1))
	sb.add_param('Window Type', windows)
	sb.add_param('Num Taps', Int(20, min=0), variable=True)
	sb.add_param('Beta', Float(6.76), variable=True)
	def make(fg, filter_type, decimation, window, ntaps, beta):
		taps_args = (window, ntaps, beta)
		return make_filter(fg, filter_type, decimation, taps_maker, taps_args)
	return sb, make

def RootRaisedCosine(sb):
	taps_maker = gr.firdes.root_raised_cosine
	filters = Enum(filter_choices, 2)
	sb.add_input_socket('in', Variable(filters, index=1))
	sb.add_output_socket('out', Variable(filters, index=2))
	sb.add_param('Filter Type', filters, type=True)		
	sb.add_param('Decimation', Int(1, min=1))
	sb.add_param('Gain', Float(1), variable=True)
	sb.add_param('Sampling Rate', Float(default_samp_rate), variable=True)
	sb.add_param('Symbol Rate', Float(100), variable=True)
	sb.add_param('Alpha', Float(6.76), variable=True)
	sb.add_param('Num Taps', Int(20, min=0), variable=True)		
	def make(fg, filter_type, decimation, gain, samp_freq, symbol_rate, alpha, ntaps):		
		taps_args = (gain, samp_freq, symbol_rate, alpha, ntaps)
		return make_filter(fg, filter_type, decimation, taps_maker, taps_args)
	return sb, make
	
###########################################################################
#	Other filters
###########################################################################
def Hilbert(sb):
	fcn = gr.hilbert_fc
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Complex())
	sb.add_param('Num Taps', Int(10, min=0))
	return sb, lambda fg, num_taps: fcn(num_taps.parse())
	
def Goertzel(sb):   
	fcn = gr.goertzel_fc  
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Complex())
	sb.add_param('Rate', Int())
	sb.add_param('Length', Int())
	sb.add_param('Frequency', Float())   
	return sb, lambda fg, rate, len, freq: fcn(rate.parse(),len.parse(),freq.parse()) 
	
def PowerSquelch(sb):
	fcn = gr.simple_squelch_cc
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Complex())
	sb.add_param('Threshold (dB)', Float(30), variable=True)
	sb.add_param('Alpha', Float(.5), variable=True)
	sb.set_docs('''Power levels below the threshold are not passed through.''')
	def make(fg, threshold, alpha):
		block = fcn(threshold.parse(), alpha.parse())
		fg.add_callback(block.set_threshold, threshold)
		fg.add_callback(block.set_alpha, alpha)
		return block
	return sb, make
	
def SinglePoleIIRFilter(sb):
	vlen = Int(1, min=1)
	type = Enum([
		('Float', (gr.single_pole_iir_filter_ff, Float())),
		('Complex', (gr.single_pole_iir_filter_cc, Complex())),
	])
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Alpha', Float(1), variable=True)
	sb.add_param('Vector Length', vlen)
	def make(fg, type, alpha, vlen):
		block = type.parse()[0](alpha.parse(), vlen.parse())
		fg.add_callback(block.set_taps, alpha)
		return block
	return sb, make
	
def AutomaticGainControl(sb):
	type = Enum([('Complex', (gr.agc_cc, Complex())), ('Float', (gr.agc_ff, Float())),], 1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Rate', Float(1e-4))
	sb.add_param('Reference', Float(1))
	sb.add_param('Gain', Float(1))
	sb.add_param('Max Gain', Float(0))	
	return sb, lambda fg, type, rate, ref, gain, max_gain: \
		type.parse()[0](rate.parse(), ref.parse(), gain.parse(), max_gain.parse())		
		
def AutomaticGainControl2(sb):
	type = Enum([('Complex', (gr.agc2_cc, Complex())), ('Float', (gr.agc2_ff, Float())),], 1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Attack Rate', Float(1e-1))
	sb.add_param('Decay Rate', Float(1e-2))
	sb.add_param('Reference', Float(1))
	sb.add_param('Gain', Float(1))
	sb.add_param('Max Gain', Float(0))	
	return sb, lambda fg, type, attack_rate, decay_rate, ref, gain, max_gain: \
		type.parse()[0](attack_rate.parse(), decay_rate.parse(), ref.parse(), gain.parse(), max_gain.parse())
		
def FeedForwardAGC(sb):
	fcn = gr.feedforward_agc_cc
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Complex())
	sb.add_param('Num Samples', Int(100, min=1))
	sb.add_param('Reference', Float(0))
	sb.set_docs('''Non-causal AGC which computes required gain based on max absolute value over nsamples.''')
	return sb, lambda fg, nsamps, ref: fcn(nsamps.parse(), ref.parse())	
	
def CMAFilter(sb):
	fcn = gr.cma_equalizer_cc
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Complex())
	sb.add_param('Num Taps', Int(10, min=0))
	sb.add_param('Modulus', Float())
	sb.add_param('Mu', Float())
	sb.set_docs('''Constant Modulus Adaptive Filter.''')
	return sb, lambda fg, num_taps, mod, mu: fcn(num_taps.parse(), mod.parse(), mu.parse())
	
def ClockRecovery(sb):
	type = Enum([
		('Complex', (gr.clock_recovery_mm_cc, Complex())),
		('Float', (gr.clock_recovery_mm_ff, Float())),
	])
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Omega', Float(), variable=True)
	sb.add_param('Gain Omega', Float(), variable=True)
	sb.add_param('Mu', Float(), variable=True)
	sb.add_param('Gain Mu', Float(), variable=True)
	sb.add_param('Omega Relative Limit', Float())
	sb.set_docs('''Mueller & Mueller clock recovery.''')
	def make(fg, type, omega, gain_omega, mu, gain_mu, omega_rel_limit): 
		block = type.parse()[0](omega.parse(), gain_omega.parse(), mu.parse(), gain_mu.parse(), omega_rel_limit.parse())
		fg.add_callback(block.set_gain_mu, gain_mu)
		fg.add_callback(block.set_gain_omega, gain_omega)
		fg.add_callback(block.set_mu, mu)
		fg.add_callback(block.set_omega, omega)
		return block
	return sb,make	
	
def FFT(sb):
	taps_maker = gr.firdes.window
	type = Enum([
		('Complex', (gr.fft_vcc, ComplexVector())),
		('Float', (gr.fft_vfc, FloatVector())),
	])
	sb.add_input_socket('vin', Variable(type, index=1))
	sb.add_output_socket('vout', ComplexVector())
	sb.add_param('Input Type', type, False, type=True)
	sb.add_param('FFT Size', Int(512, min=1))
	sb.add_param('Window', Enum(window_choices, 1))	
	sb.add_param('Beta', Float(6.76))
	return sb, lambda fg, type, size, window, beta: \
		type.parse()[0](size.parse(), True, taps_maker(window.parse(), size.parse(), beta.parse()))

def IFFT(sb):
	fcn = gr.fft_vcc
	taps_maker = gr.firdes.window
	sb.add_input_socket('vin', ComplexVector())
	sb.add_output_socket('vout', ComplexVector())
	sb.add_param('FFT Size', Int(512, min=1))
	sb.add_param('Window', Enum(window_choices, 1), type=True)	
	sb.add_param('Beta', Float(6.76))
	return sb, lambda fg, size, window, beta: \
		fcn(size.parse(), False, taps_maker(window.parse(), size.parse(), beta.parse()))
															
def Upsample(sb):
	taps_maker = gr.firdes.window
	type = Enum([
		('Complex', (gr.rational_resampler_base_ccf, Complex())),
		('Float', (gr.rational_resampler_base_fff, Float())),
	], 1)
	sb.add_input_socket('in', Variable(type, index=1))	
	sb.add_output_socket('out', Variable(type, index=1))	
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Interpolation (L)', Int(1, min=1))	
	sb.set_docs('''\
Add L-1 zeros between each sample in f(k). 
>>> Implements a gr.rational_resampler_base_**f(decim=1,  taps=[1.0])''')
	return sb, lambda fg, type, interp: type.parse()[0](interp.parse(), 1, [1.0])

def Downsample(sb):
	fcn = gr.keep_one_in_n
	type = Enum(all_choices, 1)
	sb.add_input_socket('in', Variable(type))
	sb.add_output_socket('out', Variable(type))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Decimation (M)', Int(1, min=1), variable=True)
	sb.set_docs('''\
Reduce the sample rate by keeping every Mth sample.
>>> Implements the gr.keep_one_in_n block.''')
	def make(fg, type, decimation):
		type = type.parse()
		block = fcn(type.get_num_bytes(), decimation.parse())
		fg.add_callback(block.set_n, decimation)
		return block
	return sb, make

def FractionalResampler(sb):
	lcm = gru.lcm 
	filters = Enum([
		('Complex', (blks2.rational_resampler_ccc, Complex())),
		('Float', (blks2.rational_resampler_fff, Float())),
	], 1)
	sb.add_input_socket('in', Variable(filters, index=1))
	sb.add_output_socket('out', Variable(filters, index=1))
	sb.add_param('Type', filters, False, type=True)
	sb.add_param('Input Rate', Int(100e3))
	sb.add_param('Output Rate', Int(100e3))
	sb.add_param('Fractional Bandwidth', Float(0.4, min=0.0, max=0.5))
	sb.set_docs('''\
Rational Resampler with a low pass FIR filter. \
The fractional bandwidth must be a number between 0 and 0.5. \
The fractional bandwidth controls the transition width of the low pass filter. ''')
	def make(fg, filter_type, inrate, outrate, fractional_bw):
		in_rate = inrate.parse()
		out_rate = outrate.parse()
		interp = int(lcm(in_rate, out_rate)/in_rate)
		decim = int(lcm(in_rate, out_rate)/out_rate)
		return filter_type.parse()[0](interp, decim, taps=None, fractional_bw=fractional_bw.parse())
	return sb, make
	
def FractionalInterpolator(sb):
	filters = Enum([
		('Complex', (gr.fractional_interpolator_cc, Complex())),
		('Float', (gr.fractional_interpolator_ff, Float())),
	], 1)
	sb.add_input_socket('in', Variable(filters, index=1))
	sb.add_output_socket('out', Variable(filters, index=1))
	sb.add_param('Type', filters, False, type=True)
	sb.add_param('Phase Shift', Float())
	sb.add_param('Interp Ratio', Float())	
	def make(fg, filter_type, phase_shit, interp_ratio):
		return filter_type.parse()[0](phase_shift.parse(), interp_ratio.parse())
	return sb, make
	
