"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.GraphicalSinks
#Various graphical sinks.
#@author Josh Blum

from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import default_samp_rate

number_display_pritority = 0
fft_display_priority = 1
scope_display_priority = 2
constellation_display_pritority = 3
waterfall_display_priority = 4

def FFTSink(sb):
	from gnuradio.wxgui import fftsink2
	type = Enum([
		('Complex', (fftsink2.fft_sink_c, Complex())),
		('Float', (fftsink2.fft_sink_f, Float())),
	], 1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Title', String('FFT'))	
	sb.add_param('Sampling Rate', Int(default_samp_rate))
	sb.add_param('Y per div', Int(20), variable=True)
	sb.add_param('Reference Level', Int(20), variable=True)
	sb.add_param('FFT Size', Int(512))	
	sb.add_param('FFT Rate', Int(15))	
	sb.add_param('Options', Enum([('Off', None), ('Average Values', 1), ('Peak Hold', 2)]))
	sb.set_docs('''The fft sink has a throttle automatically attatched to it at runtime to save the CPU.''')	
	def make(fg, type, title, samp_rate, y_per_div, ref_lvl, fft_size, fft_rate, options):
		options = options.parse()
		peak_hold = (options == 2)
		average = (options == 1)
		block = type.parse()[0](fg.get_panel(), title=title.parse(), y_per_div=y_per_div.parse(), 
			ref_level=ref_lvl.parse(), fft_size=fft_size.parse(), sample_rate=samp_rate.parse(),
			fft_rate=fft_rate.parse())
		block.set_peak_hold(peak_hold)	#set the peak hold option outside the contructor 
		block.set_average(average)			#set the average option outside the contructor
		fg.add_window(block.win, fft_display_priority, title.parse())
		fg.add_callback(block.set_ref_level, ref_lvl)
		fg.add_callback(block.set_y_per_div, y_per_div)
		throttle = gr.throttle(type.parse()[1].get_num_bytes(), samp_rate.parse())
		fg.connect(throttle, block)
		return throttle
	return sb, make
	
def WaterfallSink(sb):
	from gnuradio.wxgui import waterfallsink2
	type = Enum([
		('Complex', (waterfallsink2.waterfall_sink_c, Complex())),
		('Float', (waterfallsink2.waterfall_sink_f, Float())),
	], 1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Title', String('Waterfall'))
	sb.add_param('Sampling Rate', Int(default_samp_rate))
	sb.add_param('Baseband Freq', Int(0), variable=True)
	sb.add_param('Y per div', Int(20))
	sb.add_param('Reference Level', Int(20))
	sb.add_param('FFT Size', Int(512))	
	sb.add_param('FFT Rate', Int(15))	
	sb.add_param('Options', Enum([('Off', None), ('Average Values', 1)]))
	sb.set_docs('''The waterfall sink has a throttle automatically attatched to it at runtime to save the CPU.''')	
	def make(fg, type, title, samp_rate, baseband_freq, y_per_div, ref_lvl, fft_size, fft_rate, options):
		options = options.parse()
		average = (options == 1)
		block = type.parse()[0](fg.get_panel(), title=title.parse(), y_per_div=y_per_div.parse(), 
			ref_level=ref_lvl.parse(), fft_size=fft_size.parse(), sample_rate=samp_rate.parse(),
			baseband_freq=baseband_freq.parse(), fft_rate=fft_rate.parse())		
		block.set_average(average)			#set the average option outside the contructor
		fg.add_window(block.win, waterfall_display_priority, title.parse())
		fg.add_callback(block.set_baseband_freq, baseband_freq)
		throttle = gr.throttle(type.parse()[1].get_num_bytes(), samp_rate.parse())
		fg.connect(throttle, block)
		return throttle
	return sb, make
	
def ScopeSink(sb):
	from gnuradio.wxgui import scopesink2
	type = Enum([
		('Complex', (scopesink2.scope_sink_c, Complex())),
		('Float', (scopesink2.scope_sink_f, Float())),
	], 1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Title', String('Scope'))
	sb.add_param('Sampling Rate', Int(default_samp_rate))	
	sb.add_param('Frame Decimation', Int(1))
	sb.add_param('Vertical Scale', Float(0))	
	sb.add_param('Time Scale', Float(.001))
	sb.add_param('Marker', Enum([('Plus', 0), ('Dot', 1), ('Line', 2)], 2))
	sb.add_param('Num Inputs', Int(1, min=1, max=4), show_label=False, input_sockets_controller=True)
	sb.set_docs('''\
The scope sink has a throttle automatically attatched to it at runtime to save the CPU. \
Complex inputs to the scope show up as 2 channels (real and imaginary). \
A Vertical Scale of 0, sets the scope to auto-scale.''')	
	def make(fg, type, title, samp_rate, frame_decim, v_scale, t_scale, marker, num_inputs):
		v_scale = v_scale.parse()
		if v_scale == 0: v_scale = None	# v_scale = None means auto-scale
		block = type.parse()[0](fg.get_panel(), title=title.parse(), 
			sample_rate=samp_rate.parse(), frame_decim=frame_decim.parse(), 
			v_scale=v_scale, t_scale=t_scale.parse(), num_inputs=num_inputs.parse())
		{#set the marker		
			0: block.win.set_format_plus,
			1: block.win.set_format_dot,
			2: block.win.set_format_line,
		}[marker.parse()]()	
		fg.add_window(block.win, scope_display_priority, title.parse())
		#setup the throttles
		scope_throttle = gr.hier_block2(
			'scope throttle', 
			gr.io_signature(num_inputs.parse(), num_inputs.parse(), type.parse()[1].get_num_bytes()), 
			gr.io_signature(num_inputs.parse(), num_inputs.parse(), type.parse()[1].get_num_bytes()),
		)	
		for i in range(num_inputs.parse()): #connect the inputs
			scope_throttle.connect(
				(scope_throttle, i), 
				gr.throttle(type.parse()[1].get_num_bytes(), samp_rate.parse()),
				(block, i),
			)
		return scope_throttle
	return sb, make
	
def ConstellationSink(sb):
	from gnuradio.wxgui import scopesink2
	fcn = scopesink2.constellation_sink	
	sb.add_input_socket('in', Complex())
	sb.add_param('Title', String('Constellation'))
	sb.add_param('Sampling Rate', Int(default_samp_rate))	
	sb.add_param('Frame Decimation', Int(1))
	sb.add_param('Marker', Enum([('Plus', 0), ('Dot', 1), ('Line', 2)]))
	sb.set_docs('''\
The constellation sink has a throttle automatically attatched to it at runtime to save the CPU. ''')	
	def make(fg, title, samp_rate, frame_decim, marker):
		block = fcn(fg.get_panel(), title=title.parse(), 
			sample_rate=samp_rate.parse(), frame_decim=frame_decim.parse())
		{#set the marker		
			0: block.win.set_format_plus,
			1: block.win.set_format_dot,
			2: block.win.set_format_line,
		}[marker.parse()]()			
		fg.add_window(block.win, constellation_display_pritority, title.parse())	
		throttle = gr.throttle(Complex().get_num_bytes(), samp_rate.parse())
		fg.connect(throttle, block)
		return throttle
	return sb, make

def NumericalSink(sb):
	from gnuradio.wxgui import numbersink2
	type = Enum([
		('Complex', (numbersink2.number_sink_c, Complex())),
		('Float', (numbersink2.number_sink_f, Float())),
	], 1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Title', String('Number'))
	sb.add_param('Unit', String('Units'))
	sb.add_param('Sampling Rate', Int(default_samp_rate))
	sb.add_param('Base Value', Float(0.0), variable=True)
	sb.add_param('Min Value', Float(-100))
	sb.add_param('Max Value', Float(100))
	sb.add_param('Factor', Float(1.0))
	sb.add_param('Decimal Places', Int(6), variable=True)
	sb.add_param('Reference Level', Int(50), variable=True)
	sb.add_param('Number Rate', Int(15))	
	sb.add_param('Options', Enum([('Off', None), ('Average Values', 1), ('Peak Hold', 2)]))
	sb.add_param('Gauge', Bool(true='Show', false='Hide', default=True))
	sb.set_docs('''\
The number sink has a throttle automatically attatched to it at runtime to save the CPU. ''')	
	def make(fg, type, label, unit, samp_rate, base, min, max, factor, decimals, ref_lvl, num_rate, options, gauge):		
		options = options.parse()
		peak_hold = (options == 2)
		average = (options == 1)
		block = type.parse()[0](fg.get_panel(), unit=unit.parse(), 
			base_value=base.parse(), minval=min.parse(), maxval=max.parse(), factor=factor.parse(),
			decimal_places=decimals.parse(), ref_level=ref_lvl.parse(), sample_rate=samp_rate.parse(), #number_size=512,
			number_rate=num_rate.parse(),	label=label.parse())
		block.set_show_gauge(gauge.parse())
		block.set_peak_hold(peak_hold)	#set the peak hold option outside the contructor 
		block.set_average(average)			#set the average option outside the contructor
		fg.add_window(block.win, number_display_pritority, label.parse()+unit.parse())	
		fg.add_callback(block.set_base_value, base)
		fg.add_callback(block.set_ref_level, ref_lvl)
		fg.add_callback(block.set_decimal_places, decimals)
		throttle = gr.throttle(type.parse()[1].get_num_bytes(), samp_rate.parse())
		fg.connect(throttle, block)
		return throttle
	return sb, make
		
	