"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Misc
#These blocks were not categorized. Try to keep the number of misc blocks small.
#@author Josh Blum

import gnuradio.gr.gr_threading as threading
from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import default_samp_rate,all_choices,MAX_NUM_SOCKETS
from ext import selector

def Throttle(sb):
	fcn = gr.throttle
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Sampling Rate', Float(default_samp_rate))	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Connect a throttle to control the actual rate of the data.''')	
	return sb, lambda fg, type, samp_rate, vlen: fcn(type.parse().get_num_bytes()*vlen.parse(), samp_rate.parse())
	
def Head(sb):
	fcn = gr.head
	type = Enum(all_choices, 1)	
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Items', Int(1000, min=0))	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Copies the first num items to the output then stops.''')	
	return sb, lambda fg, type, num_items, vlen: fcn(type.parse().get_num_bytes()*vlen.parse(), num_items.parse())
	
def SkipHead(sb):
	fcn = gr.skiphead
	type = Enum(all_choices, 1)	
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Items', Int(1000, min=0))	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Skips the first num items, and then copies input to output.''')	
	return sb, lambda fg, type, num_items, vlen: fcn(type.parse().get_num_bytes()*vlen.parse(), num_items.parse())
	
def RMS(sb):
	type = Enum([
		('Complex', (gr.rms_cf, Complex())),
		('Float', (gr.rms_ff, Float())),
	], 1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Float())
	sb.add_param('Input Type', type, False, type=True)
	sb.add_param('Alpha', Float(0.0001), variable=True)
	def make(fg, type, alpha):
		block = type.parse()[0](alpha.parse())
		fg.add_callback(block.set_alpha, alpha)
		return block	
	return sb, make
	
def About(sb):
	sb.add_param('Title', String(''))
	sb.add_param('Author', String(''))
	return sb, lambda *args: None
	
def Note(sb):
	sb.add_param('Note', String(''))
	return sb, lambda *args: None	

def Selector(sb):
	fcn = selector.selector
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Input Index', Int(0), variable=True)
	sb.add_param('Output Index', Int(0), variable=True)	
	sb.add_param('Num Inputs', Int(1, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.add_param('Num Outputs', Int(1, min=1, max=MAX_NUM_SOCKETS), 
		show_label=False, output_sockets_controller=True)
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Forward data from the input index to the output index.''')	
	def make(fg, type, input_index, output_index, num_inputs, num_outputs, vlen):
		item_size = type.parse().get_num_bytes()*vlen.parse()
		block = fcn(
			item_size, 
			num_inputs.parse(), 
			num_outputs.parse(), 
			input_index.parse(), 
			output_index.parse(),
		)
		fg.add_callback_locked(block.set_input_index, input_index)
		fg.add_callback_locked(block.set_output_index, output_index)
		return block
	return sb, make

def Valve(sb):
	fcn = selector.selector
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Open', Int(0), variable=True)	
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''When open is 0, the valve will forward data.''')	
	def make(fg, type, open, vlen):
		item_size = type.parse().get_num_bytes()*vlen.parse()
		block = fcn(item_size, 1, 1, 0, open.parse())
		fg.add_callback_locked(block.set_output_index, open)
		return block
	return sb, make
	
def Copy(sb):
	fcn = gr.kludge_copy
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Output[i] = Input[i]''')	
	def make(fg, type, vlen):
		item_size = type.parse().get_num_bytes()*vlen.parse()
		return fcn(item_size)
	return sb, make
	
def InputTerminator(sb):
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''Terminate a dangling input. The sample rate is zero.''')	
	def make(fg, type, vlen):
		item_size = type.parse().get_num_bytes()*vlen.parse()
		null_source = gr.null_source(item_size)
		head_zero = gr.head(item_size, 0)
		fg.connect(null_source, head_zero)
		return head_zero
	return sb, make
	
