"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Modulations
#Various modulation schemes.
#@author Josh Blum

from DataTypes import *
from gnuradio import gr,blks2

def FrequencyMod(sb):
	fcn = gr.frequency_modulator_fc
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Complex())	
	sb.add_param('Sensitivity', Float(.01))	
	return sb, lambda fg, sensitivity: fcn(sensitivity.parse())
	
def PhaseMod(sb):
	fcn = gr.phase_modulator_fc
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Complex())	
	sb.add_param('Sensitivity', Float(.01))	
	return sb, lambda fg, sensitivity: fcn(sensitivity.parse())
	
def QuadratureDemod(sb):
	fcn = gr.quadrature_demod_cf
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Float())	
	sb.add_param('Gain', Float(1))	
	return sb, lambda fg, gain: fcn(gain.parse())
	
###########################################################################
#	Narrow band and Wide band FM
###########################################################################		
	
def WFMReceive(sb):
	fcn = blks2.wfm_rcv
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Float())	
	sb.add_param('Quad Rate', Float(100e3))
	sb.add_param('Audio Decimation', Int(1, min=1))		
	sb.set_docs('''\
Wide Band FM Receiver.
The audio rate is the quad rate/audio decimation.''')
	return sb, lambda fg, quad_rate, audio_dec: fcn(quad_rate.parse(), audio_dec.parse())
	
def WFMTransmit(sb):
	fcn = blks2.wfm_tx
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Complex())	
	sb.add_param('Quad Rate', Int(100e3))
	sb.add_param('Audio Decimation', Int(1, min=1))	
	sb.add_param('Tau', Float(75e-6))
	sb.add_param('Max Deviation', Float(75e3))		
	sb.set_docs('''\
Wide Band FM Transmitter.
The audio rate is the quad rate/audio decimation.''')
	def make(fg, quad_rate, audio_decimation, tau, max_dev):
		quad_rate = quad_rate.parse()
		audio_decimation = audio_decimation.parse()
		#	make sure that the quad rate is an integer multiple of the audio rate	#
		audio_rate = quad_rate/audio_decimation
		quad_rate = audio_decimation * (quad_rate/audio_decimation)
		return fcn(audio_rate, quad_rate, tau.parse(), max_dev.parse())
	return sb, make
	
def NBFMReceive(sb):
	fcn = blks2.nbfm_rx
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Float())	
	sb.add_param('Quad Rate', Int(100e3))
	sb.add_param('Audio Decimation', Int(1, min=1))	
	sb.add_param('Tau', Float(75e-6))
	sb.add_param('Max Deviation', Float(5e3))		
	sb.set_docs('''\
Narrow Band FM Receiver.
The audio rate is the quad rate/audio decimation.''')
	def make(fg, quad_rate, audio_decimation, tau, max_dev):
		quad_rate = quad_rate.parse()
		audio_decimation = audio_decimation.parse()
		#	make sure that the quad rate is an integer multiple of the audio rate	#
		audio_rate = quad_rate/audio_decimation
		quad_rate = audio_decimation * (quad_rate/audio_decimation)
		return fcn(audio_rate, quad_rate, tau.parse(), max_dev.parse())
	return sb, make	
	
def NBFMTransmit(sb):
	fcn = blks2.nbfm_tx
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Complex())	
	sb.add_param('Quad Rate', Int(100e3))
	sb.add_param('Audio Decimation', Int(1, min=1))	
	sb.add_param('Tau', Float(75e-6))
	sb.add_param('Max Deviation', Float(5e3))		
	sb.set_docs('''\
Narrow Band FM Transmitter.
The audio rate is the quad rate/audio decimation.''')
	def make(fg, quad_rate, audio_decimation, tau, max_dev):
		quad_rate = quad_rate.parse()
		audio_decimation = audio_decimation.parse()
		#	make sure that the quad rate is an integer multiple of the audio rate	#
		audio_rate = quad_rate/audio_decimation
		quad_rate = audio_decimation * (quad_rate/audio_decimation)
		return fcn(audio_rate, quad_rate, tau.parse(), max_dev.parse())
	return sb, make	
		
###########################################################################
#	AM and FM demodulation
###########################################################################			
		
def AMDemod(sb):
	fcn = blks2.am_demod_cf
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Float())	
	sb.add_param('Channel Rate', Int(100e3))
	sb.add_param('Audio Decimation', Int(1, min=1))	
	sb.add_param('Audio Passband (Hz)', Float(5e3))
	sb.add_param('Audio Stopband (Hz)', Float(5.5e3))
	return sb, lambda fg, *args: fcn(*map(lambda a: a.parse(), args))
		
def FMDemod(sb):
	fcn = blks2.fm_demod_cf
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Float())	
	sb.add_param('Channel Rate', Int(100e3))
	sb.add_param('FM Deviation (Hz)', Int(5e3))
	sb.add_param('Audio Decimation', Int(1, min=1))	
	sb.add_param('Audio Passband (Hz)', Float(3e3))
	sb.add_param('Audio Stopband (Hz)', Float(4e3))
	sb.add_param('Gain', Float(1))
	sb.add_param('Tau', Float(75e-6))
	return sb, lambda fg, *args: fcn(*map(lambda a: a.parse(), args))
	
###########################################################################
#	Phase shift keying
###########################################################################	
	
def PSKMod(sb):
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Complex())	
	sb.add_param('Type', Enum([
		('DBPSK', blks2.dbpsk_mod), 
		('DQPSK', blks2.dqpsk_mod),
		('D8PSK', blks2.d8psk_mod),
	]), type=True)
	sb.add_param('Samples/Symbol', Int(2, min=2))
	sb.add_param('Excess BW', Float(0.35))
	sb.add_param('Use Gray Code', Bool(true='Yes', false='No', default=True))
	return sb, lambda fg, type, *args: type.parse()(*map(lambda a: a.parse(), args))
	
def PSKDemod(sb):
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Byte())	
	sb.add_param('Type', Enum([
		('DBPSK', blks2.dbpsk_demod), 
		('DQPSK', blks2.dqpsk_demod),
		('D8PSK', blks2.d8psk_demod),
	]), type=True)
	sb.add_param('Samples/Symbol', Int(2, min=2))
	sb.add_param('Excess BW', Float(0.35))
	sb.add_param('Costas Alpha', Float(0.03))
	sb.add_param('Gain Mu', Float(0.05))
	sb.add_param('Mu', Float(0.005))
	sb.add_param('Omega Relative Limit', Float(0.05))
	sb.add_param('Use Gray Code', Bool(true='Yes', false='No', default=True))
	return sb, lambda fg, type, *args: type.parse()(*map(lambda a: a.parse(), args))

###########################################################################
#	Gaussian minimum shift keying
###########################################################################	
	
def GMSKMod(sb):
	fcn = blks2.gmsk_mod
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Complex())	
	sb.add_param('Samples/Symbol', Int(2, min=2))
	sb.add_param('Filter BW', Float(0.35))
	return sb, lambda fg, *args: fcn(*map(lambda a: a.parse(), args))
	
def GMSKDemod(sb):
	fcn = blks2.gmsk_demod
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Byte())	
	sb.add_param('Samples/Symbol', Int(2, min=2))
	sb.add_param('Gain Mu', Float(0.05))
	sb.add_param('Mu', Float(0.5, min=0, max=1))
	sb.add_param('Omega Relative Limit', Float(0.005))
	sb.add_param('Frequency Error', Float(0))
	sb.set_docs('''Mu is the fractional delay between 0 and 1''')
	return sb, lambda fg, *args: fcn(*map(lambda a: a.parse(), args))
	
###########################################################################
#	Quadrature amplitude modulation
###########################################################################
	
def QAMMod(sb):
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Complex())	
	sb.add_param('Type', Enum([
		('QAM 8', blks2.qam8_mod), 
		('QAM 16', blks2.qam16_mod), 
		('QAM 64', blks2.qam64_mod),
		('QAM 256', blks2.qam256_mod),
	]), type=True)
	sb.add_param('Samples/Symbol', Int(2, min=2))
	sb.add_param('Excess BW', Float(0.35))
	sb.add_param('Use Gray Code', Bool(true='Yes', false='No', default=True))
	return sb, lambda fg, type, *args: type.parse()(*map(lambda a: a.parse(), args))
	
def QAMDemod(sb):
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Byte())	
	sb.add_param('Type', Enum([
		('QAM 8', blks2.qam8_demod), 
		('QAM 16', blks2.qam16_demod),
		('QAM 64', blks2.qam64_demod),
		('QAM 256', blks2.qam256_demod),
	]), type=True)
	sb.add_param('Samples/Symbol', Int(2, min=2))
	sb.add_param('Excess BW', Float(0.35))
	sb.add_param('Costas Alpha', Float(0.03))
	sb.add_param('Gain Mu', Float(0.05))
	sb.add_param('Mu', Float(0.005))
	sb.add_param('Omega Relative Limit', Float(0.05))
	sb.add_param('Use Gray Code', Bool(true='Yes', false='No', default=True))
	return sb, lambda fg, type, *args: type.parse()(*map(lambda a: a.parse(), args))
		
###########################################################################
#	Phase Locked Loops
###########################################################################	
	
def CostasLoop(sb):
	fcn = gr.costas_loop_cc
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('bout', Complex())
	sb.add_output_socket('nout', Complex(), optional=True)
	sb.add_param('Alpha', Float())
	sb.add_param('Beta', Float())
	sb.add_param('Max Freq', Float())
	sb.add_param('Min Freq', Float())
	sb.add_param('Order', Enum([('2', 2), ('4', 4)]))
	sb.set_docs('''\
Carrier tracking PLL for QPSK input. \
The Costas loop has two output streams: \
"bout" is the baseband I stream, "nout" is the normalized frequency of the loop. \
The "nout" may be left open. ''')
	return sb, lambda fg, *args: fcn(*map(lambda a: a.parse(), args))
	
def PLL(sb):
	type = Enum([						#fcn, output type
		('Carrier Tracking', (gr.pll_carriertracking_cc, Complex())),
		('Freq Det', (gr.pll_freqdet_cf, Float())),
		('Ref Out', (gr.pll_refout_cc, Complex())),
	])
	sb.add_input_socket('in', Complex())
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, type=True)	
	sb.add_param('Alpha', Float())
	sb.add_param('Beta', Float())
	sb.add_param('Max Freq', Float())
	sb.add_param('Min Freq', Float())
	sb.set_docs('''\
All settings max_freq and min_freq are in terms of radians per sample, NOT HERTZ. \
Alpha is the phase gain (first order, units of radians per radian), and \
beta is the frequency gain (second order, units of radians per sample per radian).

Carrier Tracking - This PLL locks onto a [possibly noisy] reference carrier on the input and outputs that signal, downconverted to DC.

Freq Det - This PLL locks onto a [possibly noisy] reference carrier on the input and outputs an estimate of that frequency in radians per sample.

Ref Out - This PLL locks onto a [possibly noisy] reference carrier on the input and outputs a clean version which is phase and frequency aligned to it.
''')
	return sb, lambda fg, type, *args: type.parse()[0](*map(lambda a: a.parse(), args))		
	