"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Operators
#These blocks implement basic mathematical operations + - * / log10...
#@author Josh Blum

from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import MAX_NUM_SOCKETS

variable_inputs_doc_string = '''2 <= Num Inputs <= %d.'''%MAX_NUM_SOCKETS

def Add(sb):
	type = Enum([
		('Complex', (gr.add_cc, Complex())), 
		('Float', (gr.add_ff, Float())), 
		('Int', (gr.add_ii, Int())), 
		('Short', (gr.add_ss, Short())),
	], 1)		
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Inputs', Int(2, min=2, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.set_docs(variable_inputs_doc_string)	
	return sb, lambda fg, type, num_inputs: type.parse()[0]()
	
def AddVector(sb):
	type = Enum([
		('Complex Vector', (gr.add_vcc, ComplexVector())), 
		('Float Vector', (gr.add_vff, FloatVector())), 
		('Int Vector', (gr.add_vii, IntVector())), 
		('Short Vector', (gr.add_vss, ShortVector())),
	], 1)		
	sb.add_input_socket('vin', Variable(type, index=1))
	sb.add_output_socket('vout', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Inputs', Int(2, min=2, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.add_param('Items / Block', Int(1, min=1))
	sb.set_docs(variable_inputs_doc_string)	
	return sb, lambda fg, type, num_inputs, items_per_block: type.parse()[0](items_per_block.parse())
	
def AddConstant(sb):
	type = Enum([
		('Complex', (gr.add_const_cc, Complex())), 
		('Float', (gr.add_const_ff, Float())), 
		('Int', (gr.add_const_ii, Int())), 
		('Short', (gr.add_const_ss, Short())),
	], 1)		
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Constant', Variable(type, 1, index=1), variable=True)
	def make(fg, type, constant): 
		block = type.parse()[0](constant.parse())
		fg.add_callback(block.set_k, constant)	
		return block
	return sb, make
	
def AddConstantVector(sb):
	type = Enum([
		('Complex Vector', (gr.add_const_vcc, ComplexVector())), 
		('Float Vector', (gr.add_const_vff, FloatVector())), 
		('Int Vector', (gr.add_const_vii, IntVector())), 
		('Short Vector', (gr.add_const_vss, ShortVector())),
	], 1)		
	sb.add_input_socket('vin', Variable(type, index=1))
	sb.add_output_socket('vout', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Constant', Variable(type, 1, index=1), variable=True)
	def make(fg, type, constant): 
		block = type.parse()[0](constant.parse())
		fg.add_callback(block.set_k, constant)	
		return block
	return sb, make
	
def Multiply(sb):
	type = Enum([
		('Complex', (gr.multiply_cc, Complex())), 
		('Float', (gr.multiply_ff, Float())), 
		('Int', (gr.multiply_ii, Int())), 
		('Short', (gr.multiply_ss, Short())),
	], 1)		
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Inputs', Int(2, min=2, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.set_docs(variable_inputs_doc_string)	
	return sb, lambda fg, type, num_inputs: type.parse()[0]()
	
def MultiplyVector(sb):
	type = Enum([
		('Complex Vector', (gr.multiply_vcc, ComplexVector())), 
		('Float Vector', (gr.multiply_vff, FloatVector())), 
		('Int Vector', (gr.multiply_vii, IntVector())), 
		('Short Vector', (gr.multiply_vss, ShortVector())),
	], 1)		
	sb.add_input_socket('vin', Variable(type, index=1))
	sb.add_output_socket('vout', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Inputs', Int(2, min=2, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.add_param('Items / Block', Int(1, min=1))
	sb.set_docs(variable_inputs_doc_string)	
	return sb, lambda fg, type, num_inputs, items_per_block: type.parse()[0](items_per_block.parse())
	
def MultiplyConstant(sb):
	type = Enum([
		('Complex', (gr.multiply_const_cc, Complex())), 
		('Float', (gr.multiply_const_ff, Float())), 
		('Int', (gr.multiply_const_ii, Int())), 
		('Short', (gr.multiply_const_ss, Short())),
	], 1)		
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Constant', Variable(type, 1, index=1), variable=True)
	def make(fg, type, constant): 
		block = type.parse()[0](constant.parse())
		fg.add_callback(block.set_k, constant)	
		return block
	return sb, make
	
def MultiplyConstantVector(sb):
	type = Enum([
		('Complex Vector', (gr.multiply_const_vcc, ComplexVector())), 
		('Float Vector', (gr.multiply_const_vff, FloatVector())), 
		('Int Vector', (gr.multiply_const_vii, IntVector())), 
		('Short Vector', (gr.multiply_const_vss, ShortVector())),
	], 1)		
	sb.add_input_socket('vin', Variable(type, index=1))
	sb.add_output_socket('vout', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Constant', Variable(type, 1, index=1), variable=True)
	def make(fg, type, constant): 
		block = type.parse()[0](constant.parse())
		fg.add_callback(block.set_k, constant)	
		return block
	return sb, make
	
def Subtract(sb):
	type = Enum([
		('Complex', (gr.sub_cc, Complex())), 
		('Float', (gr.sub_ff, Float())), 
		('Int', (gr.sub_ii, Int())), 
		('Short', (gr.sub_ss, Short())),
	], 1)		
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_input_socket('-in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Inputs', Int(2, min=2, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.set_docs(variable_inputs_doc_string)
	return sb, lambda fg, type, num_inputs: type.parse()[0]()
	
def Divide(sb):
	type = Enum([
		('Complex', (gr.divide_cc, Complex())), 
		('Float', (gr.divide_ff, Float())), 
		('Int', (gr.divide_ii, Int())), 
		('Short', (gr.divide_ss, Short())),
	], 1)		
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_input_socket('/in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Num Inputs', Int(2, min=2, max=MAX_NUM_SOCKETS), 
		show_label=False, input_sockets_controller=True)
	sb.set_docs(variable_inputs_doc_string)
	return sb, lambda fg, type, num_inputs: type.parse()[0]()
	
def nLog10(sb):
	vlen = Int(1, min=1)
	sb.add_input_socket('vin', Float(), vlen=vlen)
	sb.add_output_socket('vout', Float(), vlen=vlen)
	sb.add_param('n', Float(1))	
	sb.add_param('k', Float(0))
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''output = n*log10(input) + k''')	
	return sb, lambda fg, n, k, vlen: gr.nlog10_ff(n.parse(), vlen.parse(), k.parse())
	