"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Packet
#Support for the gr packet framework.
#@author Josh Blum

import struct
import os
from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import all_choices
from ext import packet

def PacketMod(sb):
	packet.mod_pkts, packet.packet_mod #uses
	type = Enum(all_choices, 1)
	sb.add_input_socket('in', Variable(type))
	sb.add_output_socket('out', Byte())		
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Samples/Symbol', Int(1))
	sb.add_param('Bits/Symbol', Int(1))
	sb.add_param('Payload Length', Int(packet.DEFAULT_PAYLOAD_LEN))
	sb.add_param('Access Code', String())
	sb.add_param('Pad for USRP', Bool(true='Yes', false='No', default=True))
	sb.set_docs("""\
Packet modulator block, for use with the gnuradio modulator blocks: gmsk, psk, qam.

Access Code: string of 1's and 0's, leave blank for default.

Bits/Symbol should be set accordingly: 
	gmsk -> 1
	dbpsk -> 1
	dqpsk -> 2
	d8psk -> 3
	qam8 -> 3
	qam16 -> 4
	qam64 -> 6
	qam256 -> 8
""")
	def make(fg, type, samples_per_symbol, bits_per_symbol, payload_length, access_code, pad_for_usrp):
		access_code = access_code.parse()
		if not access_code: access_code = None
		mod = packet.mod_pkts(
			samples_per_symbol=samples_per_symbol.parse(), 
			bits_per_symbol=bits_per_symbol.parse(), 
			access_code=access_code, 
			pad_for_usrp=pad_for_usrp.parse(),
		)
		helper = packet.packet_mod(			
			item_size_in=type.parse().get_num_bytes(),
			item_size_out=Byte().get_num_bytes(),
			modulator=mod,
			payload_length=payload_length.parse(), 
		)
		return helper
	return sb, make
	
def PacketDemod(sb):
	packet.demod_pkts, packet.packet_demod #uses
	type = Enum(all_choices, 1)
	sb.add_input_socket('in', Byte())
	sb.add_output_socket('out', Variable(type))		
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Access Code', String())
	sb.add_param('Threshold', Int(-1))
	sb.set_docs("""\
Packet demodulator block, for use with the gnuradio demodulator blocks: gmsk, psk, qam.

Access Code: string of 1's and 0's, leave blank for default.
""")
	def make(fg, type, access_code, threshold):
		access_code = access_code.parse()
		if not access_code: access_code = None
		demod = packet.demod_pkts(
			access_code=access_code, 
			threshold=threshold.parse(),
		)
		helper = packet.packet_demod(
			item_size_in=Byte().get_num_bytes(),
			item_size_out=type.parse().get_num_bytes(),
			demodulator=demod,
		)
		return helper
	return sb, make
		
#######################################################################################
##	TUN/TAP 
#######################################################################################

IFF_TUN		= 0x0001   # tunnel IP packets
IFF_TAP		= 0x0002   # tunnel ethernet frames
IFF_NO_PI	= 0x1000   # don't pass extra packet info
IFF_ONE_QUEUE	= 0x2000   # beats me ;)
DEFAULT_TUN_DEVICE = '/dev/net/tun'
DEFAULT_VIRTUAL_DEVICE = 'tun%d'
DEFAULT_IP_ADDR = '10.0.0.1'

def open_tun_interface(tun_device_filename=DEFAULT_TUN_DEVICE, virtual_device_filename=DEFAULT_VIRTUAL_DEVICE):
	"""!
	Open a virtual ethernet interface via the Tun/Tap framework.
	An alternative function can be found: "from eunuchs.tuntap import opentuntap"
	@param tun_device_filename the path to the tun device
	@param virtual_device_filename the name of the virtual device (to be created)
	@return a file descriptor to rw the device, name of the virtual device (created)
	"""
	from fcntl import ioctl    
	mode = IFF_TAP | IFF_NO_PI
	TUNSETIFF = 0x400454ca
	tun_fd = os.open(tun_device_filename, os.O_RDWR)
	ifs = ioctl(tun_fd, TUNSETIFF, struct.pack("16sH", virtual_device_filename, mode))
	ifname = ifs[:16].strip("\x00")
	return tun_fd, ifname

class TunTapHelper(gr.hier_block2):
	"""Make the tun tap hier2 block."""
	def __init__(self, item_size, tun_fd):
		"""!
		TunTapHelper constructor.
		@param item_size the size in bytes of the IO data stream
		@param tun_fd the file descriptor for the virtual device
		"""
		#create hier block
		gr.hier_block2.__init__(
			self, 'tun_tap', 
			gr.io_signature(1, 1, item_size), 
			gr.io_signature(1, 1, item_size)
		)
		#I/O blocks
		sink = gr.file_descriptor_sink(item_size, tun_fd)
		source = gr.file_descriptor_source(item_size, tun_fd)
		#connect 
		self.connect(self, sink)
		self.connect(source, self)		
    
def TunTap(sb):
	gr.file_descriptor_sink, gr.file_descriptor_source	#uses
	type = Enum(all_choices, 1)
	sb.add_input_socket('in', Variable(type))
	sb.add_output_socket('out', Variable(type))		
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Tun Device', String(DEFAULT_TUN_DEVICE))
	sb.add_param('Virtual Device', String(DEFAULT_VIRTUAL_DEVICE))
	sb.add_param('IP Address', String(DEFAULT_IP_ADDR))
	sb.set_docs('''\
Foward data between gnuradio and a virtual ethernet interface.
---
Tun Device: File path to the tun device.

Virtual Device: Desired name for the virtual ethernet device. \
"%d" will give the device a number starting at zero.

IP Address: IP address for the virtual device, leave blank and device will not be configured.
''')
	def make(fg, type, tun, virt, ip_addr):
		item_size = type.parse().get_num_bytes()
		tun_fd, ifname = open_tun_interface(tun.parse(), virt.parse())
		#try to set the ip address
		ip_addr = ip_addr.parse()
		if ip_addr: os.system('ifconfig %s %s'%(ifname, ip_addr))
		return TunTapHelper(item_size, tun_fd)
	return sb, make
	
	