"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Sinks
#Various data sinks.
#@author Josh Blum

from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import *
from Constants import DEFAULT_FILE_PATH,MUTEX

def NullSink(sb):
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''\
If the data stream is not throttled, the null sink will use 100% CPU. \
Manually connect a throttle to the null sink to save the CPU.''')
	return sb, lambda fg, type, vlen: gr.null_sink(type.parse().get_num_bytes()*vlen.parse())

def FileSink(sb):
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileSave(DEFAULT_FILE_PATH))
	sb.add_param('Vector Length', vlen)
	sb.set_docs('''\
If the data stream is not throttled, the file sink will use 100% CPU. \
Manually connect a throttle to the file sink to save the CPU.''')
	return sb, lambda fg, type, file_path, vlen: gr.file_sink(type.parse().get_num_bytes()*vlen.parse(), file_path.parse())	
	
def UDPSink(sb):
	fcn = gr.udp_sink	
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_input_socket('in', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Source Addr', String(''))
	sb.add_param('Source Port', Short(default_udp_port, min=1))
	sb.add_param('Destination Addr', String(''))
	sb.add_param('Destination Port', Short(default_udp_port, min=1))
	sb.add_param('Payload Size', Int(512, min=1))
	sb.add_param('Vector Length', vlen)
	return sb, lambda fg, type, s_addr, s_port, d_addr, d_port, size, vlen: fcn(
		type.parse().get_num_bytes()*vlen.parse(), 
		s_addr.parse(), s_port.parse(),
		d_addr.parse(), d_port.parse(), 
		size.parse())
			
#######################################################################################
##	Variable Sink definition and thread class
#######################################################################################
	
import threading
import Variables
import numpy
	
class VariableSinkThread(threading.Thread):
	"""
	This thread will read the vector sink,
	write the numeric value back to the variable,
	and re-parse the flow graph's callbacks.
	"""
	def __init__(self, fg, var_key, msgq):
		"""!
		VariableSinkThread constructor.
		@param fg the gr flow graph
		@param var_key the variable key
		@param msgq the message queue
		"""
		self.fg = fg
		self.var_key = var_key
		self.msgq = msgq
		threading.Thread.__init__(self)
		self.setDaemon(1)
		self.keep_running = True
		self.start()
		print 'Created variable sink thread for variable "%s".'%var_key
				
	def run(self):
		"""
		In an endless while loop: read the vector sink,
		write to the variable, and parse the callbacks.
		"""
		while self.keep_running:
			msg = self.msgq.delete_head()  # blocking read of message queue
			itemsize = int(msg.arg1())
			nitems = int(msg.arg2())				
			s = msg.to_string()            # get the body of the msg as a string				
			# There may be more than one number in the message.
			# If so, we take only the last one
			if nitems > 1:
				start = itemsize * (nitems - 1)
				s = s[start:start+itemsize]				
			#	parse the data to a complex or float/int string	#
			complex_data = numpy.fromstring (s, numpy.float32)
			if len(complex_data) == 2: new_value = "%f + %fj"%(complex_data[0], complex_data[1])
			else: new_value = "%f"%(complex_data[0],)
			# write the new value #		
			Variables.reregister(self.var_key, new_value)				
			#	parse the call backs	#				
			self.fg.parse_callbacks()	
		
def VariableSink(sb):
	type = Enum(all_choices, 1)
	sb.add_input_socket('in', Variable(type))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Variable', VariableKeySelector())	
	sb.add_param('Samp Rate', Float(default_samp_rate))
	sb.set_docs('''\
A thread reads a message sink and writes to the variable. \
The variable cannot have a range. \
The variable must be specified using its name without the "$" symbol. \
The variable sink has a throttle automatically attatched to it at runtime to save the CPU.''')	
	def make(fg, type, var_key, samp_rate):
		msgq = gr.msg_queue(DEFAULT_QUEUE_LIMIT)
		item_size = type.parse().get_num_bytes()
		block = gr.message_sink(item_size, msgq, True)
		var_sink_thread = VariableSinkThread(fg, var_key.parse(), msgq)
		throttle = gr.throttle(item_size, samp_rate.parse())
		fg.connect(throttle, block)		
		return throttle
	return sb, make	
	
def RS232Sink(sb):
	import serial, os, fcntl
	type = Enum(all_choices, 1)
	sb.add_input_socket('in', Variable(type))	
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Port', String('/dev/ttyS0'))
	sb.add_param('Baud Rate', Int(9600, min=0))
	sb.add_param('Byte Size', Enum([
				('8 bits', serial.EIGHTBITS),	
				('7 bits', serial.SEVENBITS),	
				('6 bits', serial.SIXBITS),	
				('5 bits', serial.FIVEBITS),	
			]
		)
	)
	sb.add_param('Parity', Enum([
				('None', serial.PARITY_NONE),	
				('Even', serial.PARITY_EVEN),	
				('Odd', serial.PARITY_ODD),		
			]
		)
	)
	sb.add_param('Stop Bits', Enum([
				('1 bit', serial.STOPBITS_ONE),	
				('2 bits', serial.STOPBITS_TWO),		
			]
		)
	)
	sb.add_param('Soft Flow Control', Enum([
				('Off', serial.XOFF),	
				('On', serial.XON),		
			]
		)
	)
	def make(fg, type, port, baudrate, bytesize, parity, stopbits, xonxoff):
		ser = serial.Serial(
			port = port.parse(),
			baudrate = baudrate.parse(),
			bytesize = bytesize.parse(),
			parity = parity.parse(),
			stopbits = stopbits.parse(),
			timeout = None,
			xonxoff = xonxoff.parse(),
		)
		fd = ser.fileno()
		fcntl.fcntl(fd, fcntl.F_SETFL, os.O_SYNC | os.O_WRONLY)  #set blocking, and write only
		return gr.file_descriptor_sink(type.parse().get_num_bytes(), fd)
	return sb, make
	
	