"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Sources
#Various data sources.
#@author Josh Blum

from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import *
from Constants import DEFAULT_FILE_PATH
import random

MAX_SEED = 1000000	#constraint for seed generation in noise and random source 

def NullSource(sb):
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Vector Length', vlen)
	return sb, lambda fg, type, vlen: gr.null_source(type.parse().get_num_bytes()*vlen.parse())
	
def SignalSource(sb):
	type = Enum([
		('Complex', (gr.sig_source_c, Complex())), 
		('Float', (gr.sig_source_f, Float())), 
		('Int', (gr.sig_source_i, Int())), 
		('Short', (gr.sig_source_s, Short())),
	], 1)
	wfm_types = [
		('Constant', gr.GR_CONST_WAVE), 
		('Cosine', gr.GR_COS_WAVE), 
		('Sine', gr.GR_SIN_WAVE),
		('Square', gr.GR_SQR_WAVE),
		('Triangle', gr.GR_TRI_WAVE),
		('Saw Tooth', gr.GR_SAW_WAVE),
	]
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Sampling Rate', Float(default_samp_rate), variable=True)
	sb.add_param('Wave Form', Enum(wfm_types, 1))
	sb.add_param('Frequency', Float(1000), variable=True)
	sb.add_param('Amplitude', Float(1), variable=True)
	sb.add_param('Offset', Variable(type, index=1), variable=True)
	def make(fg, type, samp_freq, waveform, freq, amplitude, offset):
		block = type.parse()[0](samp_freq.parse(), waveform.parse(), freq.parse(), amplitude.parse(), offset.parse())		
		fg.add_callback(block.set_sampling_freq, samp_freq)
		fg.add_callback(block.set_frequency, freq)
		fg.add_callback(block.set_amplitude, amplitude)
		fg.add_callback(block.set_offset, offset)
		return block
	return sb, make
	
def NoiseSource(sb):
	type = Enum([
		('Complex', (gr.noise_source_c, Complex())), 
		('Float', (gr.noise_source_f, Float())), 
		('Int', (gr.noise_source_i, Int())), 
		('Short', (gr.noise_source_s, Short())),
	], 1)
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Noise Type', Enum([('Uniform', gr.GR_UNIFORM), 
												('Gaussian', gr.GR_GAUSSIAN), 
												('Laplacian', gr.GR_LAPLACIAN),
												('Impulse', gr.GR_IMPULSE)],1))
	sb.add_param('Ampliude', Float(.1), variable=True)
	sb.add_param('Seed', Int(random.randint(0, MAX_SEED), min=0, max=MAX_SEED))	
	def make(fg, type, noise_type, amplitude, seed):
		block = type.parse()[0](noise_type.parse(), amplitude.parse(), seed.parse())	
		fg.add_callback(block.set_amplitude, amplitude)
		return block
	return sb, make	
	
def FileSource(sb):
	type = Enum(all_choices, 1)	
	vlen = Int(1, min=1)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Repeat', Bool(true='Yes', false='No', default=True))
	sb.add_param('Vector Length', vlen)
	return sb, lambda fg, type, file_path, repeat, vlen: gr.file_source(type.parse().get_num_bytes()*vlen.parse(), file_path.parse(), repeat.parse())	
	
def VectorSource(sb):
	enum_all_vector_data_types = Enum([
		('Complex Vector', (gr.vector_source_c, ComplexVector(), Complex())),
		('Float Vector', (gr.vector_source_f, FloatVector(), Float())),
		('Int Vector', (gr.vector_source_i, IntVector(), Int())),
		('Short Vector', (gr.vector_source_s, ShortVector(), Short())),
		('Byte Vector', (gr.vector_source_b, ByteVector(), Byte()))
	], 1)#index of 1 makes the default float
	sb.add_output_socket('out', Variable(enum_all_vector_data_types, index=2))
	sb.add_param('Type', enum_all_vector_data_types, False, type=True)
	sb.add_param('Vector', Variable(enum_all_vector_data_types, "0, 1", index=1))
	sb.add_param('Repeat', Bool(true='Yes', false='No', default=True))
	sb.set_docs('''\
This vector source takes a vector and outputs a stream of repeating elements. \
The vector paramater is a set of comma-separted numerical expressions. ''')	
	return sb, lambda fg, type, vector, repeat: type.parse()[0](vector.parse(), repeat.parse())	
	
def RandomVector(sb):
	enum_all_vector_data_types = Enum([
		('Float', (gr.vector_source_f, Float())),
		('Int', (gr.vector_source_i, Int())),
		('Short', (gr.vector_source_s, Short())),
		('Byte', (gr.vector_source_b, Byte()))
	])
	sb.add_output_socket('out', Variable(enum_all_vector_data_types, index=1))
	sb.add_param('Type', enum_all_vector_data_types, False, type=True)
	sb.add_param('Vector Length', Int(1000, min=1))
	sb.add_param('Min', Variable(enum_all_vector_data_types, -100, index=1))
	sb.add_param('Max', Variable(enum_all_vector_data_types, 100, index=1))
	sb.add_param('Seed', Int(random.randint(0, MAX_SEED), min=0, max=MAX_SEED))
	sb.add_param('Repeat', Bool(true='Yes', false='No', default=True))
	sb.set_docs('''\
This random source takes a random vector and outputs a stream of repeating elements. \
Float outputs will be random decimal values between min and max. \
Other output types will be random integer values between min and max. 
>>> Creates a gr.vector_source_* with a random vector.''')
	def make(fg, type, vlen, min, max, seed, repeat):
		fcn = type.parse()[0]
		type = type.parse()[1]
		min = min.parse()
		max = max.parse()
		vector = list()
		for i in range(vlen.parse()):
			if type.get_type() == Float().get_type(): vector.append(random.uniform(min, max))
			else: vector.append(random.randint(min, max))
		return fcn(vector, repeat.parse())
	return sb, make
		
def UDPSource(sb):
	fcn = gr.udp_source	
	type = Enum(all_choices, 1)
	vlen = Int(1, min=1)
	sb.add_output_socket('out', Variable(type), vlen=vlen)
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Source Addr', String(''))
	sb.add_param('Source Port', Short(default_udp_port, min=1))	
	sb.add_param('Payload Size', Int(512, min=1))
	sb.add_param('Vector Length', vlen)
	return sb, lambda fg, type, addr, port, size, vlen: fcn(type.parse().get_num_bytes()*vlen.parse(), addr.parse(), port.parse(), size.parse())	

def RS232Source(sb):
	import serial, os, fcntl
	type = Enum(all_choices, 1)
	sb.add_output_socket('out', Variable(type))	
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Port', String('/dev/ttyS0'))
	sb.add_param('Baud Rate', Int(9600, min=0))
	sb.add_param('Byte Size', Enum([
				('8 bits', serial.EIGHTBITS),	
				('7 bits', serial.SEVENBITS),	
				('6 bits', serial.SIXBITS),	
				('5 bits', serial.FIVEBITS),	
			]
		)
	)
	sb.add_param('Parity', Enum([
				('None', serial.PARITY_NONE),	
				('Even', serial.PARITY_EVEN),	
				('Odd', serial.PARITY_ODD),		
			]
		)
	)
	sb.add_param('Stop Bits', Enum([
				('1 bit', serial.STOPBITS_ONE),	
				('2 bits', serial.STOPBITS_TWO),		
			]
		)
	)
	sb.add_param('Soft Flow Control', Enum([
				('Off', serial.XOFF),	
				('On', serial.XON),		
			]
		)
	)
	def make(fg, type, port, baudrate, bytesize, parity, stopbits, xonxoff):
		ser = serial.Serial(
			port = port.parse(),
			baudrate = baudrate.parse(),
			bytesize = bytesize.parse(),
			parity = parity.parse(),
			stopbits = stopbits.parse(),
			timeout = None,
			xonxoff = xonxoff.parse(),
		)
		fd = ser.fileno()
		fcntl.fcntl(fd, fcntl.F_SETFL, os.O_SYNC | os.O_RDONLY)  #set blocking, and read only
		return gr.file_descriptor_source(type.parse().get_num_bytes(), fd)
	return sb, make

	