"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.Trellis
#Trellis blocks.
#@author Achilleas Anastasopoulos

from DataTypes import *
from gnuradio import gr
from SignalBlockConstants import all_choices
from Constants import DEFAULT_FILE_PATH
	
def Encoder(sb):
	from gnuradio import trellis
	fsm = trellis.fsm
	type = Enum([('Byte->Byte', (trellis.encoder_bb, Byte(),Byte())),
		      ('Byte->Short', (trellis.encoder_bs, Byte(),Short())),
		      ('Byte->Int', (trellis.encoder_bi, Byte(),Int())),
		      ('Short->Short', (trellis.encoder_ss, Short(),Short())),
		      ('Short->Int', (trellis.encoder_si, Short(),Int())),
		      ('Int->Int', (trellis.encoder_ii, Int(),Int())),
		      ],3)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=2))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Initial State', Int())
	sb.set_docs('''Trellis Encoder. For the time being the underlying FSM can only be read from a file...''')
	def make(fg, type, filename, s0):
		block = type.parse()[0](fsm(filename.parse()), s0.parse())
		return block
	return sb, make

def Metrics(sb):
	from gnuradio import trellis
	type = Enum([('Complex', (trellis.metrics_c, Complex(),ComplexVector())),
		      ('Float', (trellis.metrics_f, Float(),FloatVector())),
		      ('Int', (trellis.metrics_i, Int(),IntVector())),
		      ('Short', (trellis.metrics_s, Short(),ShortVector())),],1)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Float())
	sb.add_param('Type', type, False, type=True)
	sb.add_param('Output cardinality', Int())
	sb.add_param('Dimensionality', Int())
	sb.add_param('Constellation', Variable(type, index=2))
	sb.add_param('Metric Type',
	      Enum([('Euclidean', trellis.TRELLIS_EUCLIDEAN),
		     ('Hard Symbol', trellis.TRELLIS_HARD_SYMBOL),
		     ('Hard Bit', trellis.TRELLIS_HARD_BIT)],0))
	sb.set_docs('''Generate metrics required for Viterbi or SISO algorithms.''')
	def make(fg, type, O, D, Con, MetrType):
		block = type.parse()[0](O.parse(), D.parse(), Con.parse(), MetrType.parse())
		return block
	return sb, make

def Viterbi(sb):
	from gnuradio import trellis
	fsm = trellis.fsm
	type = Enum([('Byte', (trellis.viterbi_b, Byte())),
		      ('Short', (trellis.viterbi_s, Short())),
		      ('Int', (trellis.viterbi_i, Int())),
		      ],1)
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Length', Int())
	sb.add_param('Initial State', Int())
	sb.add_param('Final State', Int())
	sb.set_docs('''Viterbi Decoder. FSM is specified through its filename only...''')
	def make(fg, type, filename, K, s0, sK):
		block = type.parse()[0](fsm(filename.parse()), K.parse(), s0.parse(), sK.parse())
		return block
	return sb, make

def Viterbi_Combined(sb):
	from gnuradio import trellis
	fsm = trellis.fsm
	type = Enum([
	     ('Complex->Byte', (trellis.viterbi_combined_cb, Complex(), ComplexVector(), Byte())),
	     ('Complex->Short', (trellis.viterbi_combined_cs, Complex(), ComplexVector(), Short())),
	     ('Complex->Int', (trellis.viterbi_combined_ci, Complex(), ComplexVector(), Int())),
	     ('Float->Byte', (trellis.viterbi_combined_fb, Float(), FloatVector(), Byte())),
	     ('Float->Short', (trellis.viterbi_combined_fs, Float(), FloatVector(), Short())),
	     ('Float->Int', (trellis.viterbi_combined_fi, Float(), FloatVector(), Int())),
	     ('Int->Byte', (trellis.viterbi_combined_ib, Float(), FloatVector(), Byte())),
	     ('Int->Short', (trellis.viterbi_combined_is, Float(), FloatVector(), Short())),
	     ('Int->Int', (trellis.viterbi_combined_ii, Float(), FloatVector(), Int())),
	     ('Short->Byte', (trellis.viterbi_combined_sb, Short(), ShortVector(), Byte())),
	     ('Short->Short', (trellis.viterbi_combined_ss, Short(), ShortVector(), Short())),
	     ('Short->Int', (trellis.viterbi_combined_si, Short(), ShortVector(), Int())),
		      ],4)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Variable(type, index=3))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Length', Int())
	sb.add_param('Initial State', Int())
	sb.add_param('Final State', Int())
	sb.add_param('Dimensionality', Int())
	sb.add_param('Constellation', Variable(type, index=2))
	sb.add_param('Metric Type',
	      Enum([('Euclidean', trellis.TRELLIS_EUCLIDEAN),
		     ('Hard Symbol', trellis.TRELLIS_HARD_SYMBOL),
		     ('Hard Bit', trellis.TRELLIS_HARD_BIT)],0))
	sb.set_docs('''Viterbi Decoder combined with metric calculation. FSM is specified through its filename only...''')
	def make(fg, type, filename, K, s0, sK, D, Con, MetrType):
		block = type.parse()[0](fsm(filename.parse()), K.parse(), s0.parse(), sK.parse(), D.parse(), Con.parse(), MetrType.parse())
		return block
	return sb, make

def Soft_In_Soft_Out(sb):
	from gnuradio import trellis
	fcn = trellis.siso_f
	fsm = trellis.fsm
	sb.add_input_socket('in', Float())
	sb.add_output_socket('out', Float())
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Length', Int())
	sb.add_param('Initial State', Int())
	sb.add_param('Final State', Int())
	sb.add_param('Generate a-posteriori info on Input',
			  Enum([('Yes', True), ('No', False)]))
	sb.add_param('Generate a-posteriori info on Output',
			  Enum([('Yes', True), ('No', False)]))
	sb.add_param('min-sum or sum-product',
			  Enum([('Min-Sum', trellis.TRELLIS_MIN_SUM),
				 ('Sum-Product', trellis.TRELLIS_SUM_PRODUCT)],0))
	sb.set_docs('''BCJR Algorithm. FSM is specified through its filename only...''')
	def make(fg, filename, K, s0, sK,posti, posto, siso_type):
		block = fcn(fsm(filename.parse()), K.parse(), s0.parse(), sK.parse(), posti.parse(), posto.parse(), siso_type.parse())
		return block
	return sb, make

def Soft_In_Soft_Out_Combined(sb):
	from gnuradio import trellis
	fsm = trellis.fsm
	type = Enum([
	     #('Complex', (trellis.siso_combined_c, Complex(), ComplexVector())),
	     ('Float', (trellis.siso_combined_f, Float(), FloatVector())),
	     #('Int', (trellis.siso_combined_i, Int(), IntVector())),
	     #('Short', (trellis.siso_combined_s, Short(), ShortVector())),
		      ],0)
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_output_socket('out', Float())
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Length', Int())
	sb.add_param('Initial State', Int())
	sb.add_param('Final State', Int())
	sb.add_param('Generate a-posteriori info on Input',
			  Enum([('Yes', True), ('No', False)]))
	sb.add_param('Generate a-posteriori info on Output',
			  Enum([('Yes', True), ('No', False)]))
	sb.add_param('min-sum or sum-product',
			  Enum([('Min-Sum', trellis.TRELLIS_MIN_SUM),
				 ('Sum-Product', trellis.TRELLIS_SUM_PRODUCT)],0))
	sb.add_param('Dimensionality', Int())
	sb.add_param('Constellation', Variable(type, index=2))
	sb.add_param('Metric Type',
	      Enum([('Euclidean', trellis.TRELLIS_EUCLIDEAN),
		     ('Hard Symbol', trellis.TRELLIS_HARD_SYMBOL),
		     ('Hard Bit', trellis.TRELLIS_HARD_BIT)],0))
	sb.set_docs('''BCJR Algorithm combined with metric calculation. FSM is specified through its filename only...''')
	def make(fg, type, filename, K, s0, sK, posti, posto, siso_type, dim, constel, metr_type):
		block = type.parse()[0](fsm(filename.parse()), K.parse(), s0.parse(), sK.parse(), posti.parse(), posto.parse(), siso_type.parse(), dim.parse(), constel.parse(), metr_type.parse())
		return block
	return sb, make

def Interleaver(sb):
	from gnuradio import trellis
	fcn = trellis.permutation
	interleaver = trellis.interleaver
	type = Enum(all_choices, 3)
	sb.add_input_socket('in', Variable(type))
	sb.add_output_socket('out', Variable(type))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Symbols per Block', Int(1))
	sb.set_docs('''Interleave chunks of symbols. Interleaver specification is given by specifying a file...''')
	def make(fg, type, filename, SpB):
		inter = interleaver(filename.parse())
		block = fcn(inter.K(), inter.INTER(), SpB.parse(), type.parse().get_num_bytes())
		return block
	return sb, make

def Deinterleaver(sb):
	from gnuradio import trellis
	fcn = trellis.permutation
	interleaver = trellis.interleaver
	type = Enum(all_choices, 3)
	sb.add_input_socket('in', Variable(type))
	sb.add_output_socket('out', Variable(type))
	sb.add_param('Type', type, False, type=True)
	sb.add_param('File Path', FileOpen(DEFAULT_FILE_PATH))
	sb.add_param('Symbols per Block', Int(1))
	sb.set_docs('''Deinterleave chunks of symbols. Interleaver specification is given by specifying a file...''')
	def make(fg, type, filename, SpB):
		inter = interleaver(filename.parse())
		block = fcn(inter.K(), inter.DEINTER(), SpB.parse(), type.parse().get_num_bytes())
		return block
	return sb, make

