"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package SignalBlockDefs.USRP
#The source and sink blocks for the USRP.
#@author Josh Blum

from DataTypes import *
from gnuradio import gr
import sys

####################################################################
#	Constants
####################################################################

##default interpolation for a new sink
DEFAULT_INTERP = 512

##default decimation for a new source
DEFAULT_DECIM = 256

##auto tr options
_auto_tr_options = [
	('Ignore', None),
	('Enable', True),
	('Disable', False),
]

##transmit enable options
_tx_enb_options = [
	('Ignore', None),
	('Enable', True),
	('Disable', False),
]

##receive antenna options
_rx_ant_options = [
	('Ignore', None),
	('TX/RX', 'TX/RX'),
	('RX2', 'RX2'),
]

##helpful usrp source message
_usrp_source_msg = '''
The USRP source outputs 64 Megasamples per second / decimation.

--- Flex RF specific ---
The "Auto TR" and "RX Antenna" settings are flex rf specific and should be left at "Ignore" unless this is a flex rf board.

If enabled, "Auto Transmit/Receive Switching" handles the preference for transmit packets vs receive packets. \
By default, "Auto TR" is disabled.

The "Receive Antenna Setting" selects one of the SMA connectors as the data source. \
By default, 'TX/RX' is selected.
'''

##helpful usrp sink message
_usrp_sink_msg = '''
The USRP sink inputs 128 Megasamples per second / interpolation.

--- Flex RF specific ---
The "Auto TR" and "TX Enable" settings are flex rf specific and should be left at "Ignore" unless this is a flex rf board.

If enabled, "Auto Transmit/Receive Switching" handles the preference for transmit packets vs receive packets. \
By default, "Auto TR" is disabled.

The "Transmit Enable" configures the transmitter to be on or off. \
Do not leave this unconfigured.
'''

####################################################################
#	Helper Functions
####################################################################

def _set_freq(u, which, subdev, freq, verbose=False):
	"""!
	Set the carrier frequency for the given subdevice.	
	@param u the usrp source/sink
	@param which specifies the DDC/DUC number
	@param freq the carrier frequency in Hz
	@param verbose if true, print usrp tuning information
	"""
	r = u.tune(which, subdev, freq)
	if verbose and r:	
			print subdev.side_and_name()
			print "  r.baseband_freq =", r.baseband_freq
			print "  r.dxc_freq      =", r.dxc_freq
			print "  r.residual_freq =", r.residual_freq
			print "  r.inverted      =", r.inverted
	elif not r: raise ValueError, '"%s" is not a valid frequency for this sub-device\n'%freq
		
def _setup_rx_subdev(fg, u, subdev_spec, ddc, gain, freq, auto_tr=None, rx_ant=None):
	"""!
	Setup a usrp receive subdevice by setting gain and freq.
	Add the gain and freq callbacks to the flow graph.
	FlexRF: Handle auto transmit/receive and set the receive antenna. 
	@param fg the flow graph
	@param u the usrp object
	@param subdev_spec the sub-device specification
	@param ddc which ddc to use: 0 or 1
	@param gain the gain DataType
	@param freq the freq DataType
	@param auto_tr auto transmit/receive True, False, or None
	@param rx_ant the receive antenna: 'TX/RX', 'RX2', or None
	"""
	from gnuradio import usrp
	subdev = usrp.selected_subdev(u, subdev_spec)#get the subdev
	subdev.set_gain(gain.parse())	
	_set_freq(u, ddc, subdev, freq.parse(), verbose=True)
	fg.add_callback(lambda f: _set_freq(u, ddc, subdev, f), freq)
	fg.add_callback(subdev.set_gain, gain) #callback for gain setting 
	if auto_tr != None: subdev.set_auto_tr(auto_tr)
	if rx_ant != None: subdev.select_rx_antenna(rx_ant)

def _setup_tx_subdev(fg, u, subdev_spec, gain, freq, auto_tr=None, tx_enb=None):
	"""!
	Setup a usrp receive subdevice by setting gain and freq.
	Add the gain and freq callbacks to the flow graph.
	FlexRF: Handle auto transmit/receive and enable the transmitter. 
	@param fg the flow graph
	@param u the usrp object
	@param subdev_spec the sub-device specification
	@param gain the gain DataType
	@param freq the freq DataType
	@param auto_tr auto transmit/receive True, False, or None
	@param tr_enb the transmit enable: True, False, or None
	"""
	from gnuradio import usrp
	subdev = usrp.selected_subdev(u, subdev_spec)#get the subdev
	subdev.set_gain(gain.parse())	
	_set_freq(u, subdev._which, subdev, freq.parse(), verbose=True)
	fg.add_callback(lambda f: _set_freq(u, subdev._which, subdev, f), freq)
	fg.add_callback(subdev.set_gain, gain) #callback for gain setting 
	if auto_tr != None: subdev.set_auto_tr(auto_tr)
	if tx_enb != None: subdev.set_enable(tx_enb)
		
####################################################################
#	Reguar USRP Source
####################################################################

def USRPSource(sb):
	from gnuradio import usrp
	type = Enum([('Complex', (usrp.source_c, Complex())), ('IShort', (usrp.source_s, Short()))])
	sb.add_output_socket('out', Variable(type, index=1))
	sb.add_param('Output Type', type, False, type=True)
	sb.add_param('Unit Number', Int(0, min=0))
	sb.add_param('Side:Subdevice', Enum([
		('Auto', None),
		('Side A:0', (0, 0)), 
		('Side B:0', (1, 0)),
		('Side A:1', (0, 1)), 
		('Side B:1', (1, 1)),
	]))
	sb.add_param('Frequency', Float(), variable=True)
	sb.add_param('Decimation', Int(DEFAULT_DECIM, min=1))
	sb.add_param('Gain', Float(0), variable=True)
	sb.add_param('Mux', Hex())
	sb.add_param('Auto T/R', Enum(_auto_tr_options))
	sb.add_param('RX Antenna', Enum(_rx_ant_options))
	sb.set_docs('''\
If left at 0, the Mux value will be automatically determined. \
Otherwise, the Mux value must be in hexidecimal format.
''' + _usrp_source_msg)
	def make(fg, type, number, subdev_spec, freq, decim, gain, mux, auto_tr, rx_ant):
		number = number.parse()
		subdev_spec = subdev_spec.parse()	
		u = type.parse()[0](number, nchan=1)
		if subdev_spec == None: subdev_spec = usrp.pick_rx_subdevice(u)
		u.set_decim_rate(decim.parse())
		if mux.parse() == 0: mux = usrp.determine_rx_mux_value(u, subdev_spec)
		else: mux = mux.parse()
		u.set_mux(mux)			
		_setup_rx_subdev(fg, u, subdev_spec, 0, gain, freq, auto_tr.parse(), rx_ant.parse())			
		return u
	return sb, make
	
####################################################################
#	Dual USRP Source
####################################################################	
	
def USRPDualSource(sb):
	from gnuradio import usrp
	default_mux = '0x3210'
	type = Enum([('Complex', (usrp.source_c, Complex())), ('IShort', (usrp.source_s, Short()))])
	sb.add_output_socket('outA', Variable(type, index=1))
	sb.add_output_socket('outB', Variable(type, index=1))
	sb.add_param('Output Type', type, False, type=True)
	sb.add_param('Unit Number', Int(0, min=0))	
	sb.add_param('FrequencyA', Float(), variable=True)
	sb.add_param('FrequencyB', Float(), variable=True)
	sb.add_param('Decimation', Int(DEFAULT_DECIM, min=1))
	sb.add_param('GainA', Float(0), variable=True)	
	sb.add_param('GainB', Float(0), variable=True)
	sb.add_param('Mux', Hex(default_mux))		
	sb.add_param('Auto T/R', Enum(_auto_tr_options))
	sb.add_param('RX AntennaA', Enum(_rx_ant_options))
	sb.add_param('RX AntennaB', Enum(_rx_ant_options))
	sb.set_docs('''\
Default mux is %s.
Channel A (I & Q) will get ADC0 and ADC1
Channel B (I & Q) will get ADC2 and ADC3
'''%default_mux + _usrp_source_msg)
	def make(fg, type, number, freq_a, freq_b, decim, gain_a, gain_b, mux, auto_tr, rx_ant_a, rx_ant_b):
		number = number.parse()
		u = type.parse()[0](number, nchan=2)		
		u.set_decim_rate(decim.parse())
		u.set_mux(mux.parse())
		# setup A	#		
		_setup_rx_subdev(fg, u, (0, 0), 0, gain_a, freq_a, auto_tr.parse(), rx_ant_a.parse())				
		#	setup B	#
		_setup_rx_subdev(fg, u, (1, 0), 1, gain_b, freq_b, auto_tr.parse(), rx_ant_b.parse())	
		###	Create an interleaver for the 2 data streams	###
		deinter = gr.deinterleave(type.parse()[1].get_num_bytes())
		fg.connect(u, deinter)
		return deinter
	return sb, make	

####################################################################
#	Reguar USRP Sink
####################################################################

def USRPSink(sb):
	from gnuradio import usrp
	type = Enum([('Complex', (usrp.sink_c, Complex())), ('IShort', (usrp.sink_s, Short()))])
	sb.add_input_socket('in', Variable(type, index=1))
	sb.add_param('Input Type', type, False, type=True)
	sb.add_param('Unit Number', Int(0, min=0))
	sb.add_param('Side', Enum([
		('Auto', None),
		('Side A', (0, 0)), 
		('Side B', (1, 0)),
	]))
	sb.add_param('Frequency', Float(), variable=True)
	sb.add_param('Interpolation', Int(DEFAULT_INTERP, min=1))
	sb.add_param('Gain', Float(0), variable=True)	
	sb.add_param('Mux', Hex())
	sb.add_param('Auto T/R', Enum(_auto_tr_options))
	sb.add_param('TX Enable', Enum(_tx_enb_options))
	sb.set_docs('''\
If left at 0, the Mux value will be automatically determined. \
Otherwise, the Mux value must be in hexidecimal format.
''' + _usrp_sink_msg)
	def make(fg, type, number, subdev_spec, freq, interp, gain, mux, auto_tr, tx_enb):
		number = number.parse()
		subdev_spec = subdev_spec.parse()	
		u = type.parse()[0](number, nchan=1)
		if subdev_spec == None: subdev_spec = usrp.pick_tx_subdevice(u)
		u.set_interp_rate(interp.parse())
		if mux.parse() == 0: mux = usrp.determine_tx_mux_value(u, subdev_spec)
		else: mux = mux.parse()
		u.set_mux(mux)
		_setup_tx_subdev(fg, u, subdev_spec, gain, freq, auto_tr.parse(), tx_enb.parse())
		return u
	return sb, make
	
####################################################################
#	Dual USRP Sink
####################################################################	
	
def USRPDualSink(sb):
	from gnuradio import usrp
	default_mux = '0xba98'
	type = Enum([('Complex', (usrp.sink_c, Complex())), ('IShort', (usrp.sink_s, Short()))])
	sb.add_input_socket('Ain', Variable(type, index=1))
	sb.add_input_socket('Bin', Variable(type, index=1))
	sb.add_param('Input Type', type, False, type=True)
	sb.add_param('Unit Number', Int(0, min=0))
	sb.add_param('FrequencyA', Float(), variable=True)
	sb.add_param('FrequencyB', Float(), variable=True)
	sb.add_param('Interpolation', Int(DEFAULT_INTERP, min=1))
	sb.add_param('GainA', Float(0), variable=True)	
	sb.add_param('GainB', Float(0), variable=True)
	sb.add_param('Mux', Hex(default_mux))	
	sb.add_param('Auto T/R', Enum(_auto_tr_options))
	sb.add_param('TX EnableA', Enum(_tx_enb_options))
	sb.add_param('TX EnableB', Enum(_tx_enb_options))
	sb.set_docs('''\
Default mux is %s. 
This will use all 4 DACs: 
DAC0 and DAC1 will get channel A (I & Q).
DAC3 and DAC2 will get channel B (I & Q).
'''%default_mux + _usrp_sink_msg)
	def make(fg, type, number, freq_a, freq_b, interp, gain_a, gain_b, mux, auto_tr, tx_enb_a, tx_enb_b):
		number = number.parse()
		u = type.parse()[0](number, nchan=2)		
		u.set_interp_rate(interp.parse())
		u.set_mux(mux.parse())
		# setup side A	#	
		_setup_tx_subdev(fg, u, (0, 0), gain_a, freq_a, auto_tr.parse(), tx_enb_a.parse())		
		#	setup side B	#
		_setup_tx_subdev(fg, u, (1, 0), gain_b, freq_b, auto_tr.parse(), tx_enb_b.parse())	
		###	Create an interleaver for the 2 data streams	###
		inter = gr.interleave(type.parse()[1].get_num_bytes())
		fg.connect(inter, u)
		return inter
	return sb, make
	
	