#!/usr/bin/env python
#
# Copyright 2003,2004,2006,2007 Free Software Foundation, Inc.
# 
# This file is part of GNU Radio
# 
# GNU Radio is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# GNU Radio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GNU Radio; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
# 

from gnuradio import gr, packet_utils
import gnuradio.gr.gr_threading as _threading

##payload length in bytes
DEFAULT_PAYLOAD_LEN = 512

##how many messages in a queue
DEFAULT_MSGQ_LIMIT = 2

##threshold for unmaking packets
DEFAULT_THRESHOLD = 12

#######################################################################################
##	Packet Modulation
#######################################################################################

class packet_mod_thread(_threading.Thread):
	"""
	Sample the input message queue and call send.		
	"""
	
	def __init__(self, msgq, payload_length, send):
		"""!
		packet_mod_thread contructor.
		@param msgq the message queue to sample
		@param payload_length number of bytes in a data-stream slice
		@param send a function that takes a payload
		"""		
		self._msgq = msgq
		self._payload_length = payload_length
		self._send = send
		_threading.Thread.__init__(self)
		self.setDaemon(1)
		self.keep_running = True
		self.start()
		
	def run(self):
		"""
		Run this thread by sampling the message queue in and cutting out chunks of size payload.
		"""
		sample = '' #residual sample
		while self.keep_running:
			msg = self._msgq.delete_head() #blocking read of message queue
			sample = sample + msg.to_string() #get the body of the msg as a string	
			while len(sample) >= self._payload_length:		
				payload = sample[0:self._payload_length]
				sample = sample[self._payload_length:]			
				self._send(payload)	

class packet_mod(gr.hier_block2):
	"""
	Hierarchical block for wrapping packet-based modulators.
	"""
	
	def __init__(self, item_size_in, item_size_out, modulator, payload_length=None):
		"""!
		packet_mod constructor.
		@param item_size_in the size of the input data stream in bytes
		@param item_size_out the size of the output data stream in bytes
		@param modulator the packet-based modulator
		@param payload_length number of bytes in a data-stream slice
		"""
		#setup
		self._item_size_in = item_size_in
		self._item_size_out = item_size_out
		if payload_length is None: #get packet length
			payload_length = DEFAULT_PAYLOAD_LEN
		if payload_length%self._item_size_in != 0:	#verify that packet length is a multiple of the stream size
			raise ValueError, 'The packet length: "%d" is not a mutiple of the stream size: "%d".'%(payload_length, self._item_size_in)
		self._payload_length = payload_length
		self._modulator = modulator		
		#create blocks
		self._msgq_in = gr.msg_queue(DEFAULT_MSGQ_LIMIT)
		msg_sink	= gr.message_sink(self._item_size_in, self._msgq_in, False) #False -> blocking		
		#initialize hier2
		gr.hier_block2.__init__(
			self, 
			"packet mod",
			gr.io_signature(1, 1, self._item_size_in), # Input signature
			gr.io_signature(1, 1, self._item_size_out) # Output signature
		)
		#connect
		self.connect(self, msg_sink)
		self.connect(self._modulator, self)
		#start thread
		packet_mod_thread(self._msgq_in, self._payload_length, self._modulator.send_pkt)		

class mod_pkts(gr.hier_block2):
	"""
	Hierarchical block for sending packet based data.		
	"""
	
	def __init__(self, samples_per_symbol, bits_per_symbol, access_code=None, pad_for_usrp=True):
		"""!
		mod_pkts constructor.
		@param samples_per_symbol number of samples per symbol
		@param bits_per_symbol number of bits per symbol
		@param access_code AKA sync vector
		@param pad_for_usrp If true, packets are padded such that they end up a multiple of 128 samples
		"""
		#setup parameters
		self._samples_per_symbol = samples_per_symbol
		self._bits_per_symbol = bits_per_symbol
		self._pad_for_usrp = pad_for_usrp
		if access_code is None: #get access code
			access_code = packet_utils.default_access_code
		if not packet_utils.is_1_0_string(access_code):
			raise ValueError, "Invalid access_code %r. Must be string of 1's and 0's" % (access_code,)
		self._access_code = access_code		
		self._pad_for_usrp = pad_for_usrp		
		#create blocks
		msg_source = gr.message_source(gr.sizeof_char, DEFAULT_MSGQ_LIMIT)
		self._msgq_out = msg_source.msgq()		
		#initialize hier2
		gr.hier_block2.__init__(
			self, 
			"mod pkts",
			gr.io_signature(0, 0, 0), # Input signature
			gr.io_signature(1, 1, gr.sizeof_char) # Output signature
		)
		#connect
		self.connect(msg_source, self)
		
	def send_pkt(self, payload='', eof=False):
		"""!
		Wrap the payload in a packet and push onto the message queue.
		@param payload string, data to send
		@param eof boolean, end of file
		"""
		if eof: 
			msg = gr.message(1) # tell self._pkt_input we're not sending any more packets
		else:			
			packet = packet_utils.make_packet(
				payload,
				self._samples_per_symbol,
				self._bits_per_symbol,
				self._access_code,
				self._pad_for_usrp
			)
			msg = gr.message_from_string(packet)
		self._msgq_out.insert_tail(msg)
	
#######################################################################################
##	Packet Demodulation
#######################################################################################

class packet_demod(gr.hier_block2):
	"""
	Hierarchical block for wrapping packet-based demodulators.
	"""
	
	def __init__(self, item_size_in, item_size_out, demodulator):
		"""!
		packet_demod constructor.
		@param item_size_in the size of the input data stream in bytes
		@param item_size_out the size of the output data stream in bytes
		@param demodulator the packet-based demodulator
		"""
		#setup
		self._item_size_in = item_size_in
		self._item_size_out = item_size_out
		self._demodulator = demodulator	
		#blocks
		msg_source = gr.message_source(self._item_size_out, DEFAULT_MSGQ_LIMIT)
		self._msgq_out = msg_source.msgq()		
		#initialize hier2
		gr.hier_block2.__init__(
			self, 
			"packet demod",
			gr.io_signature(1, 1, self._item_size_in), # Input signature
			gr.io_signature(1, 1, self._item_size_out) # Output signature
		)
		#connect
		self.connect(self, gr.kludge_copy(self._item_size_in), self._demodulator)
		self.connect(msg_source, self)
		#set callback
		self._demodulator._watcher.callback = self._recv_packet

	def _recv_packet(self, ok, payload):
		"""!
		Extract the payload from the packet and push onto message queue.
		@param ok boolean ok
		@param payload data received
		"""
		msg = gr.message_from_string(payload, 0, self._item_size_out, len(payload)/self._item_size_out)
		if ok: self._msgq_out.insert_tail(msg)
	
class demod_pkts(gr.hier_block2):
	"""
	Hierarchical block for demodulating and deframing packets.
	"""

	def __init__(self, access_code=None, callback=None, threshold=-1):
		"""!
		demod_pkts contructor.   
		@param access_code AKA sync vector
		@param callback function of two args: ok, payload
		@param threshold detect access_code with up to threshold bits wrong (-1 -> use default)
		"""
		#access code
		if access_code is None: #get access code
			access_code = packet_utils.default_access_code
		if not packet_utils.is_1_0_string(access_code):
			raise ValueError, "Invalid access_code %r. Must be string of 1's and 0's" % (access_code,)
		self._access_code = access_code		
      #threshold
		if threshold == -1: threshold = DEFAULT_THRESHOLD 
		self._threshold = threshold  
		self._callback = callback
		#initialize hier2
		gr.hier_block2.__init__(
			self,
			"demod pkts",
			gr.io_signature(1, 1, gr.sizeof_char), # Input signature
			gr.io_signature(0, 0, 0), # Output signature
		) 
		#blocks
		self._msgq_in = gr.msg_queue(DEFAULT_MSGQ_LIMIT)          # holds packets from the PHY
		correlator = gr.correlate_access_code_bb(self._access_code, self._threshold)
		framer_sink = gr.framer_sink_1(self._msgq_in)		
		#connect
		self.connect(self, correlator, framer_sink)
		#start thread
		self._watcher = _queue_watcher_thread(self._msgq_in, self._callback)
			
class _queue_watcher_thread(_threading.Thread):
    def __init__(self, rcvd_pktq, callback):
        _threading.Thread.__init__(self)
        self.setDaemon(1)
        self.rcvd_pktq = rcvd_pktq
        self.callback = callback
        self.keep_running = True
        self.start()

    def run(self):
        while self.keep_running:
            msg = self.rcvd_pktq.delete_head()
            ok, payload = packet_utils.unmake_packet(msg.to_string(), int(msg.arg1()))
            if self.callback:
                self.callback(ok, payload)

