"""
Copyright 2007 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""
##@package Variables
#Holds variable names, their values, and ranges.
#@author Josh Blum

from Constants import DEFAULT_SLIDER_STEPS

##Hold the variables for the executing program. This dictionary maps keys to values,ranges.
VARS_DICT = dict()

##place this character before a variable
VAR_MARKER = '$' 

##get a list of variable keys
get_keys = VARS_DICT.keys	

def clear_vars():
	"""Call unregister on every var key."""
	for key in VARS_DICT.keys(): unregister(key)

def register(key, value, min='', max='', step=''):
	"""!
	Register the value and a [range] for a given key.	
	The range is a min, max, and a step size.
	If no range is specified, value can be any string.
	If a range is specified min, max, and step numbers have to be parsable as floats.
	Partially completed ranges will be intelligently filled in.
	Partially blanked ranges will be cleared.	
	All parameters must be strings!	
	@param key the variable key
	@param value the value
	@param min the mimimum of the variable's range
	@param max the maximum of the variable's range
	@param step the step size for the variable's range
	@throw Exception invalid arguments	
	"""
	#	make sure that all parameters are strings #
	for var in (key, value, min, max, step):
		if type(var) != type(str()): raise TypeError, '"%s" var is not of type string'%var
	#	make sure that the key does not already exists #	
	if has_key(key): raise KeyError, '"%s" already exists!'%key
	#	make sure that the key name is valid #
	if not is_key_name_valid(key): raise KeyError, '"%s" is a invalid key name!'%key	
	#	count the number of blank strings in min, max, and step #
	num_blanks = [min, max, step].count('')
	if num_blanks == 3: pass #allow the vars to be registered normally
	else:	#num_blanks is less than 3
		try: float(value)	#make sure that value is a float
		except ValueError: raise TypeError, '"%s" must be parseable to type float'%value
		if num_blanks == 2:	#do something intelligent to fill the two blanks
			#	make sure that the non-blank value is a float	#
			for var in (min, max, step):
				if var != '':	#only try to parse the non-blank variable
					try: float(var)	#make sure that value is a float
					except ValueError: raise TypeError, '"%s" must be parseable to type float'%var
			#	determine the non-blank var and generate the other two vars	#
			if min != '': max = value
			elif max != '': min = value
			elif step != '': raise ValueError, 'step cannot have a value when min and max are blank'
		elif num_blanks == 1:	#clear all vars for min, max, and step
			if step != '': min = max = step = '' #only if step was not the blank one
		elif num_blanks == 0: pass #none were blank, move onto the validity check
	# check the values for validity #	
	if min != '' and max != '': #only check for validity if min and max are non blank
		if float(value) < float(min) or float(value) > float(max) or float(min) >= float(max):
			raise ValueError, '%s is not greater than or equal to %s and less than and equal to %s'%(value, min, max)	
		if step == '': step = str((float(max) - float(min))/float(DEFAULT_SLIDER_STEPS))
		if (float(max) - float(min))/float(step) < 1: raise ValueError, '%s step size is too large'%step
	###	register	the values ###
	VARS_DICT[key] = (value, min, max, step)
	
def reregister(key, new_value):
	"""!
	Register a new value for variable: 
	Unregister the variable and try to register it with the new values.
	If an exception is raised, register the old values.
	@param key the variable key
	@param new_value the value	
	@throw Exception invalid arguments	
	@return true if the variable was registered
	"""
	value, min, max, step = get_values(key)		
	if new_value != value:
		unregister(key)
		try: 
			register(key, new_value, min, max, step)
			return True
		except Exception, e: 
			register(key, value, min, max, step)
			raise e
	else: return False
	
def is_key_name_valid(key_name):
	"""!
	Is the key valid?
	Valid keys start with a letter and may contain only letters, numbers, underscrores(_).
	@param key_name the potential key name (string)
	@return true if the key name is valid
	"""
	if len(key_name) == 0 or len(key_name) > 20: return False
	for i,char in enumerate(key_name):
		if i == 0 and not char.isalpha(): return False
		elif not char.isalpha() and not char.isalnum() and char not in ('_'): return False
	return True

def unregister(key):
	"""!
	Remove a key if it exisits.
	@param key the variable key to remove
	"""	
	if has_key(key): del VARS_DICT[key]
							
def has_key(key):
	"""!
	Is the key in this registry?
	@param key the variable key potentially in the registry
	@return true if key is in the registry
	"""	
	return key in VARS_DICT

def get_values(key):
	"""!
	Get the valuesfor this key.
	@param key the variable key
	@return the (value,min,max,step) tuple
	"""
	return VARS_DICT[key]
	
def is_ranged(key):
	"""!
	Does this variable have a range?
	The variable has a range if min, max, and step are non-blank.
	Only check the minimum for non-blankness.
	@param key the variable key
	@return true if the variable has a range
	"""
	return get_values(key)[1] != ''
		
def get_value(key):
	"""!
	Get the value for this key.
	@param key the variable key
	@return the value
	"""
	return get_values(key)[0]
	
def replace_var_instances(string):
	"""!
	Replace every instance of a marker + variable key with the string representation of the variable.
	@param string the input string
	@return the string with variables substituted out
	"""
	for key in get_keys(): string = string.replace(VAR_MARKER + key, get_value(key))
	return string		

##########################################################################
##	Import the Variables
##########################################################################	
def from_nested_data(nested_data):
	"""!
	Load the variables from nested data.	
	Clear all of the registered variables,
	and load all variables from nested data.	
	@param nested_data the nested data for variables	
	@return a list of keys, in the order
	"""
	import ParseXML,Messages
	find_data = ParseXML.find_data
	clear_vars()
	keys = list()
	for var in nested_data:
		var = find_data([var], 'var')				
		key = find_data(var, 'key')
		value = find_data(var, 'value')
		# min, max, and step are optional	#
		min = find_data(var, 'min')
		max = find_data(var, 'max')
		step = find_data(var, 'step')		
		#	if min, max, and step are all None, make them blank strings #
		if step == max == min == None: step = min = max = ''	
		#	if step is None, make it a blank string, for backwards compadibility	#
		if step == None: step = ''		
		try: 
			register(key, value, min, max, step)
			keys.append(key)
		except Exception, e:  
			Messages.send_error_load('Could not register variable with: "%s, %s, %s, %s, %s":\n\t%s'%(key, value, min, max, step, e))
			try: 
				register(key, value)	#error!, register only the value
				keys.append(key)
			except Exception, e: 
				Messages.send_error_load('Could not register variable with: "%s, %s":\n\t%s'%(key, value, e))
				Messages.send_error_load('Variable "%s" not registered!'%key)
				pass	#dont register!		
	return keys
	
if __name__ == '__main__':	
	"""Use the main method to test variable's functions."""
	pass

		